@@ -1,6 +1,6 @@
 /* cygerrno.h: main Cygwin header file.
 
-   Copyright 2000, 2001, 2002, 2003 Red Hat, Inc.
+   Copyright 2000, 2001, 2002, 2003, 2004 Red Hat, Inc.
 
 This file is part of Cygwin.
 
@@ -18,7 +18,7 @@ int __stdcall geterrno_from_win_error (DWORD code, int deferrno) __attribute__ (
 #define __seterrno_from_win_error(val) seterrno_from_win_error (__FILE__, __LINE__, val)
 
 #ifndef DEBUGGING
-#define set_errno(val) (errno = (val))
+#define set_errno(val) (errno = (val), _impure_ptr->_errno = (val))
 #else
 int __stdcall __set_errno (const char *ln, int ln, int val) __attribute ((regparm(3)));
 #define set_errno(val) __set_errno (__PRETTY_FUNCTION__, __LINE__, (val))
