@@ -1,6 +1,6 @@
 /* dcrt0.cc -- essentially the main() for the Cygwin dll
 
-   Copyright 1996, 1997, 1998, 1999, 2000, 2001, 2002, 2003 Red Hat, Inc.
+   Copyright 1996, 1997, 1998, 1999, 2000, 2001, 2002, 2003, 2004 Red Hat, Inc.
 
 This file is part of Cygwin.
 
@@ -10,6 +10,7 @@ details. */
 
 #include "winsup.h"
 #include <unistd.h>
+#include <stdio.h>
 #include <stdlib.h>
 #include "glob.h"
 #include "exceptions.h"
@@ -23,8 +24,8 @@ details. */
 #define NEED_VFORK
 #include "perprocess.h"
 #include "security.h"
-#include "fhandler.h"
 #include "path.h"
+#include "fhandler.h"
 #include "dtable.h"
 #include "cygheap.h"
 #include "child_info_magic.h"
@@ -34,31 +35,45 @@ details. */
 #include "dll_init.h"
 #include "cygthread.h"
 #include "sync.h"
+#include "heap.h"
 
 #define MAX_AT_FILE_LEVEL 10
 
 #define PREMAIN_LEN (sizeof (user_data->premain) / sizeof (user_data->premain[0]))
 
-HANDLE NO_COPY hMainProc;
+HANDLE NO_COPY hMainProc = (HANDLE) -1;
 HANDLE NO_COPY hMainThread;
 
-sigthread NO_COPY mainthread;		// ID of the main thread
-
-per_thread_waitq NO_COPY waitq_storage;
+#ifdef NEWVFORK
 per_thread_vfork NO_COPY vfork_storage;
-per_thread_signal_dispatch NO_COPY signal_dispatch_storage;
+#endif
 
-per_thread NO_COPY *threadstuff[] = {&waitq_storage,
+per_thread NO_COPY *threadstuff[] = {
+#ifdef NEWVFORK
 				     &vfork_storage,
-				     &signal_dispatch_storage,
+#endif
 				     NULL};
 
 bool display_title;
 bool strip_title_path;
-bool allow_glob = TRUE;
+bool allow_glob = true;
 codepage_type current_codepage = ansi_cp;
 
-int cygwin_finished_initializing;
+int __argc_safe;
+int _declspec(dllexport) __argc;
+char _declspec(dllexport) **__argv;
+#ifdef NEWVFORK
+vfork_save NO_COPY *main_vfork;
+#endif
+
+static int NO_COPY envc;
+char NO_COPY **envp;
+
+extern "C" void __sinit (_reent *);
+
+_cygtls NO_COPY *_main_tls;
+
+bool NO_COPY cygwin_finished_initializing;
 
 /* Used in SIGTOMASK for generating a bit for insertion into a sigset_t.
    This is subtracted from the signal number prior to shifting the bit.
@@ -72,7 +87,6 @@ ResourceLocks _reslock NO_COPY;
 MTinterface _mtinterf;
 
 bool NO_COPY _cygwin_testing;
-unsigned NO_COPY _cygwin_testing_magic;
 
 char NO_COPY almost_null[1];
 
@@ -84,7 +98,7 @@ extern "C"
   char ***main_environ;
   /* __progname used in getopt error message */
   char *__progname;
-  struct _reent reent_data = _REENT_INIT(reent_data);
+  struct _reent reent_data;
   struct per_process __cygwin_user_data =
   {/* initial_sp */ 0, /* magic_biscuit */ 0,
    /* dll_major */ CYGWIN_VERSION_DLL_MAJOR,
@@ -108,7 +122,7 @@ extern "C"
    /* impure_ptr */ &reent_data,
   };
   bool ignore_case_with_glob;
-  int __declspec (dllexport) _check_for_executable = TRUE;
+  int __declspec (dllexport) _check_for_executable = true;
 #ifdef DEBUGGING
   int pinger;
 #endif
@@ -131,25 +145,18 @@ do_global_dtors (void)
 static void __stdcall
 do_global_ctors (void (**in_pfunc)(), int force)
 {
-  if (!force)
-    {
-      if (user_data->forkee || user_data->run_ctors_p)
-	return;		// inherit constructed stuff from parent pid
-      user_data->run_ctors_p = 1;
-    }
+  if (!force && user_data->forkee)
+    return;		// inherit constructed stuff from parent pid
 
   /* Run ctors backwards, so skip the first entry and find how many
      there are, then run them. */
 
-  void (**pfunc)() = in_pfunc;
+  void (**pfunc) () = in_pfunc;
 
   while (*++pfunc)
     ;
   while (--pfunc > in_pfunc)
     (*pfunc) ();
-
-  if (user_data->magic_biscuit == SIZEOF_PER_PROCESS)
-    atexit (do_global_dtors);
 }
 
 /*
@@ -158,7 +165,7 @@ do_global_ctors (void (**in_pfunc)(), int force)
  * A \@file is replaced with @file so that echo \@foo would print
  * @foo and not the contents of foo.
  */
-static int __stdcall
+static bool __stdcall
 insert_file (char *name, char *&cmd)
 {
   HANDLE f;
@@ -175,7 +182,7 @@ insert_file (char *name, char *&cmd)
   if (f == INVALID_HANDLE_VALUE)
     {
       debug_printf ("couldn't open file '%s', %E", name);
-      return FALSE;
+      return false;
     }
 
   /* This only supports files up to about 4 billion bytes in
@@ -185,7 +192,7 @@ insert_file (char *name, char *&cmd)
   if (size == 0xFFFFFFFF)
     {
       debug_printf ("couldn't get file size for '%s', %E", name);
-      return FALSE;
+      return false;
     }
 
   int new_size = strlen (cmd) + size + 2;
@@ -193,7 +200,7 @@ insert_file (char *name, char *&cmd)
   if (!tmp)
     {
       debug_printf ("malloc failed, %E");
-      return FALSE;
+      return false;
     }
 
   /* realloc passed as it should */
@@ -204,13 +211,13 @@ insert_file (char *name, char *&cmd)
   if (!rf_result || (rf_read != size))
     {
       debug_printf ("ReadFile failed, %E");
-      return FALSE;
+      return false;
     }
 
   tmp[size++] = ' ';
   strcpy (tmp + size, cmd);
   cmd = tmp;
-  return TRUE;
+  return true;
 }
 
 static inline int
@@ -436,8 +443,8 @@ check_sanity_and_sync (per_process *p)
 	       p->magic_biscuit, SIZEOF_PER_PROCESS);
 
   /* Complain if incompatible API changes made */
-  if (p->api_major != cygwin_version.api_major)
-    api_fatal ("cygwin DLL and APP are out of sync -- API version mismatch %d < %d",
+  if (p->api_major > cygwin_version.api_major)
+    api_fatal ("cygwin DLL and APP are out of sync -- API version mismatch %d > %d",
 	       p->api_major, cygwin_version.api_major);
 
   if (CYGWIN_VERSION_DLL_MAKE_COMBINED (p->dll_major, p->dll_minor) <=
@@ -451,8 +458,7 @@ static MEMORY_BASIC_INFORMATION NO_COPY sm;
 #define CYGWIN_GUARD ((wincap.has_page_guard ()) ? \
 		     PAGE_EXECUTE_READWRITE|PAGE_GUARD : PAGE_NOACCESS)
 
-// __inline__ void
-extern void
+static void
 alloc_stack_hard_way (child_info_fork *ci, volatile char *b)
 {
   void *new_stack_pointer;
@@ -488,7 +494,7 @@ alloc_stack_hard_way (child_info_fork *ci, volatile char *b)
     api_fatal ("fork: couldn't get new stack info, %E");
   if (!noguard)
     {
-      m.BaseAddress = (LPVOID)((DWORD)m.BaseAddress - 1);
+      m.BaseAddress = (LPVOID) ((DWORD) m.BaseAddress - 1);
       if (!VirtualAlloc ((LPVOID) m.BaseAddress, 1, MEM_COMMIT,
 			 CYGWIN_GUARD))
 	api_fatal ("fork: couldn't allocate new stack guard page %p, %E",
@@ -497,7 +503,7 @@ alloc_stack_hard_way (child_info_fork *ci, volatile char *b)
   if (!VirtualQuery ((LPCVOID) m.BaseAddress, &m, sizeof m))
     api_fatal ("fork: couldn't get new stack info, %E");
   ci->stacktop = m.BaseAddress;
-  *b = 0;
+  b[0] = '\0';
 }
 
 /* extend the stack prior to fork longjmp */
@@ -512,66 +518,142 @@ alloc_stack (child_info_fork *ci)
   if (!VirtualQuery ((LPCVOID) &b, &sm, sizeof sm))
     api_fatal ("fork: couldn't get stack info, %E");
 
-  if (sm.AllocationBase != ci->stacktop)
-    alloc_stack_hard_way (ci, b + sizeof (b) - 1);
-  else
+  if (sm.AllocationBase == ci->stacktop)
     ci->stacksize = 0;
+  else
+    alloc_stack_hard_way (ci, b + sizeof (b) - 1);
 
   return;
 }
 
-static NO_COPY int mypid = 0;
-int __argc_safe;
-int _declspec(dllexport) __argc;
-char _declspec(dllexport) **__argv;
-vfork_save NO_COPY *main_vfork = NULL;
-
+#ifdef DEBUGGING
 void
-sigthread::init (const char *s)
+break_here ()
 {
-  InitializeCriticalSection (&lock);
-  id = GetCurrentThreadId ();
+  debug_printf ("break here");
+}
+#endif
+
+static void
+initial_env ()
+{
+  char buf[CYG_MAX_PATH + 1];
+  if (GetEnvironmentVariable ("CYGWIN_TESTING", buf, sizeof (buf) - 1))
+    _cygwin_testing = 1;
+
+#ifdef DEBUGGING
+  DWORD len;
+
+  if (GetEnvironmentVariable ("CYGWIN_SLEEP", buf, sizeof (buf) - 1))
+    {
+      DWORD ms = atoi (buf);
+      buf[0] = '\0';
+      len = GetModuleFileName (NULL, buf, CYG_MAX_PATH);
+      console_printf ("Sleeping %d, pid %u %s\n", ms, GetCurrentProcessId (), buf);
+      Sleep (ms);
+      if (!strace.active)
+	{
+	  strace.inited = 0;
+	  strace.hello ();
+	}
+    }
+  if (GetEnvironmentVariable ("CYGWIN_DEBUG", buf, sizeof (buf) - 1))
+    {
+      char buf1[CYG_MAX_PATH + 1];
+      len = GetModuleFileName (NULL, buf1, CYG_MAX_PATH);
+      strlwr (buf1);
+      strlwr (buf);
+      char *p = strchr (buf, ':');
+      if (!p)
+	p = (char *) "gdb.exe -nw";
+      else
+	*p++ = '\0';
+      if (strstr (buf1, buf))
+	{
+	  error_start_init (p);
+	  try_to_debug ();
+	  console_printf ("*** Sending Break.  gdb may issue spurious SIGTRAP message.\n");
+	  DebugBreak ();
+	  break_here ();
+	}
+    }
+#endif
+
 }
 
-/* Take over from libc's crt0.o and start the application. Note the
-   various special cases when Cygwin DLL is being runtime loaded (as
-   opposed to being link-time loaded by Cygwin apps) from a non
-   cygwin app via LoadLibrary.  */
-static void
-dll_crt0_1 ()
+void __stdcall
+dll_crt0_0 ()
 {
-  /* According to onno@stack.urc.tue.nl, the exception handler record must
-     be on the stack.  */
-  /* FIXME: Verify forked children get their exception handler set up ok. */
-  exception_list cygwin_except_entry;
-
-  /* Initialize SIGSEGV handling, etc. */
-  init_exceptions (&cygwin_except_entry);
-
-  /* Set the os_being_run global. */
   wincap.init ();
-  check_sanity_and_sync (user_data);
+  initial_env ();
 
+  char zeros[sizeof (child_proc_info->zero)] = {0};
+
+  init_console_handler ();
+  init_global_security ();
+  if (!DuplicateHandle (GetCurrentProcess (), GetCurrentProcess (),
+		       GetCurrentProcess (), &hMainProc, 0, FALSE,
+			DUPLICATE_SAME_ACCESS))
+    hMainProc = GetCurrentProcess ();
+
+  DuplicateHandle (hMainProc, GetCurrentThread (), hMainProc,
+		   &hMainThread, 0, false, DUPLICATE_SAME_ACCESS);
+
+  (void) SetErrorMode (SEM_FAILCRITICALERRORS);
+
+  STARTUPINFO si;
+  GetStartupInfo (&si);
+  child_proc_info = (child_info *) si.lpReserved2;
+
+  int mypid = 0;
+  if (si.cbReserved2 < EXEC_MAGIC_SIZE || !child_proc_info
+      || memcmp (child_proc_info->zero, zeros,
+		 sizeof (child_proc_info->zero)) != 0)
+    child_proc_info = NULL;
+  else
+    {
+      if ((child_proc_info->intro & OPROC_MAGIC_MASK) == OPROC_MAGIC_GENERIC)
+	multiple_cygwin_problem ("proc", child_proc_info->intro, 0);
+      else if (child_proc_info->intro == PROC_MAGIC_GENERIC
+	       && child_proc_info->magic != CHILD_INFO_MAGIC)
+	multiple_cygwin_problem ("proc", child_proc_info->magic,
+				 CHILD_INFO_MAGIC);
+      else if (child_proc_info->cygheap != (void *) &_cygheap_start)
+	multiple_cygwin_problem ("cygheap", (DWORD) child_proc_info->cygheap,
+				 (DWORD) &_cygheap_start);
+      unsigned should_be_cb = 0;
+      switch (child_proc_info->type)
+	{
+	  case _PROC_FORK:
+	    user_data->forkee = child_proc_info->cygpid;
+	    should_be_cb = sizeof (child_info_fork);
+	    /* fall through */;
+	  case _PROC_SPAWN:
+	  case _PROC_EXEC:
+	    if (!should_be_cb)
+	      should_be_cb = sizeof (child_info);
+	    if (should_be_cb != child_proc_info->cb)
+	      multiple_cygwin_problem ("proc size", child_proc_info->cb, should_be_cb);
+	    else if (sizeof (fhandler_union) != child_proc_info->fhandler_union_cb)
+	      multiple_cygwin_problem ("fhandler size", child_proc_info->fhandler_union_cb, sizeof (fhandler_union));
+	    else
+	      {
+		cygwin_user_h = child_proc_info->user_h;
+		mypid = child_proc_info->cygpid;
+		break;
+	      }
+	  default:
+	    system_printf ("unknown exec type %d", child_proc_info->type);
+	    /* intentionally fall through */
+	  case _PROC_WHOOPS:
+	    child_proc_info = NULL;
+	    break;
+	}
+    }
+
+  device::init ();
   do_global_ctors (&__CTOR_LIST__, 1);
-
-  /* Nasty static stuff needed by newlib -- point to a local copy of
-     the reent stuff.
-     Note: this MUST be done here (before the forkee code) as the
-     fork copy code doesn't copy the data in libccrt0.cc (that's why we
-     pass in the per_process struct into the .dll from libccrt0). */
-
-  _impure_ptr = &reent_data;
-
-  user_data->resourcelocks->Init ();
-  user_data->threadinterface->Init ();
-
-  mainthread.init ("mainthread"); // For use in determining if signals
-				  //  should be blocked.
-
-  winpids::init ();
-
-  int envc = 0;
-  char **envp = NULL;
+  cygthread::init ();
 
   if (!child_proc_info)
     memory_init ();
@@ -609,9 +691,9 @@ dll_crt0_1 ()
 	    __argv = spawn_info->moreinfo->argv;
 	    envp = spawn_info->moreinfo->envp;
 	    envc = spawn_info->moreinfo->envc;
-	    cygheap->fdtab.fixup_after_exec (spawn_info->parent);
+	    envp = spawn_info->moreinfo->envp;
+	    cygheap->fdtab.fixup_after_exec ();
 	    signal_fixup_after_exec ();
-	    CloseHandle (spawn_info->parent);
 	    if (spawn_info->moreinfo->old_title)
 	      {
 		old_title = strcpy (title_buf, spawn_info->moreinfo->old_title);
@@ -625,11 +707,38 @@ dll_crt0_1 ()
 	CloseHandle (child_proc_info->pppid_handle);
     }
 
+  _cygtls::init ();
+
+  /* Initialize events */
+  events_init ();
+
+  cygheap->cwd.init ();
+}
+
+/* Take over from libc's crt0.o and start the application. Note the
+   various special cases when Cygwin DLL is being runtime loaded (as
+   opposed to being link-time loaded by Cygwin apps) from a non
+   cygwin app via LoadLibrary.  */
+static void
+dll_crt0_1 (char *)
+{
+  /* According to onno@stack.urc.tue.nl, the exception handler record must
+     be on the stack.  */
+  /* FIXME: Verify forked children get their exception handler set up ok. */
+  exception_list cygwin_except_entry;
+
+  check_sanity_and_sync (user_data);
+  malloc_init ();
+
+  /* Initialize SIGSEGV handling, etc. */
+  init_exceptions (&cygwin_except_entry);
+
+  user_data->resourcelocks->Init ();
+  user_data->threadinterface->Init ();
   ProtectHandle (hMainProc);
   ProtectHandle (hMainThread);
-  cygthread::init ();
 
-  /* Initialize pthread mainthread when not forked and it is save to call new,
+  /* Initialize pthread mainthread when not forked and it is safe to call new,
      otherwise it is reinitalized in fixup_after_fork */
   if (!user_data->forkee)
     pthread::init_mainthread ();
@@ -642,15 +751,10 @@ dll_crt0_1 ()
      Need to do this before any helper threads start. */
   debug_init ();
 
+#ifdef NEWVFORK
   cygheap->fdtab.vfork_child_fixup ();
-
-  (void) SetErrorMode (SEM_FAILCRITICALERRORS);
-
-  /* Initialize events. */
-  events_init ();
-
-  cygheap->cwd.init ();
   main_vfork = vfork_storage.create ();
+#endif
 
   cygbench ("pre-forkee");
   if (user_data->forkee)
@@ -664,10 +768,9 @@ dll_crt0_1 ()
 	 this step. */
       if (fork_info->stacksize)
 	{
-	  asm ("movl %0,%%fs:4" : : "r" (fork_info->stackbottom));
-	  asm ("movl %0,%%fs:8" : : "r" (fork_info->stacktop));
+	  _tlsbase = (char *) fork_info->stackbottom;
+	  _tlstop = (char *) fork_info->stacktop;
 	}
-
       longjmp (fork_info->jmp, fork_info->cygpid);
     }
 
@@ -687,9 +790,6 @@ dll_crt0_1 ()
   /* Allocate cygheap->fdtab */
   dtable_init ();
 
-  /* Init global well known SID objects */
-  cygsid::init ();
-
   /* Initialize user info. */
   uinfo_init ();
 
@@ -716,9 +816,9 @@ dll_crt0_1 ()
 	 win32 style. */
       if ((strchr (__argv[0], ':')) || (strchr (__argv[0], '\\')))
 	{
-	  char *new_argv0 = (char *) alloca (MAX_PATH);
+	  char *new_argv0 = (char *) malloc (CYG_MAX_PATH);
 	  cygwin_conv_to_posix_path (__argv[0], new_argv0);
-	  __argv[0] = new_argv0;
+	  __argv[0] = (char *) realloc (new_argv0, strlen (new_argv0) + 1);
 	}
     }
 
@@ -754,7 +854,7 @@ dll_crt0_1 ()
       set_console_title (cp);
     }
 
-  cygwin_finished_initializing = 1;
+  cygwin_finished_initializing = true;
   /* Call init of loaded dlls. */
   dlls.init ();
 
@@ -772,72 +872,31 @@ dll_crt0_1 ()
     }
 
   /* Disable case-insensitive globbing */
-  ignore_case_with_glob = FALSE;
+  ignore_case_with_glob = false;
 
-  /* Flush signals and ensure that signal thread is up and running. Can't
-     do this for noncygwin case since the signal thread is blocked due to
-     LoadLibrary serialization. */
-  wait_for_sigthread ();
 
   set_errno (0);
 
   MALLOC_CHECK;
   cygbench (__progname);
+
+  /* Flush signals and ensure that signal thread is up and running. Can't
+     do this for noncygwin case since the signal thread is blocked due to
+     LoadLibrary serialization. */
+  wait_for_sigthread ();
   if (user_data->main)
     exit (user_data->main (__argc, __argv, *user_data->envptr));
 }
 
-#ifdef DEBUGGING
-void
-break_here ()
+struct _reent *
+initialize_main_tls (char *padding)
 {
-  debug_printf ("break here");
-}
-#endif
-
-void
-initial_env ()
-{
-  DWORD len;
-  char buf[MAX_PATH + 1];
-#ifdef DEBUGGING
-  if (GetEnvironmentVariable ("CYGWIN_SLEEP", buf, sizeof (buf) - 1))
+  if (!_main_tls)
     {
-      DWORD ms = atoi (buf);
-      buf[0] = '\0';
-      len = GetModuleFileName (NULL, buf, MAX_PATH);
-      console_printf ("Sleeping %d, pid %u %s\n", ms, GetCurrentProcessId (), buf);
-      Sleep (ms);
-    }
-  if (GetEnvironmentVariable ("CYGWIN_DEBUG", buf, sizeof (buf) - 1))
-    {
-      char buf1[MAX_PATH + 1];
-      len = GetModuleFileName (NULL, buf1, MAX_PATH);
-      strlwr (buf1);
-      strlwr (buf);
-      char *p = strchr (buf, ':');
-      if (!p)
-	p = (char *) "gdb.exe -nw";
-      else
-	*p++ = '\0';
-      if (strstr (buf1, buf))
-	{
-	  error_start_init (p);
-	  try_to_debug ();
-	  break_here ();
-	}
-    }
-#endif
-
-  if (GetEnvironmentVariable ("CYGWIN_TESTING", buf, sizeof (buf) - 1))
-    {
-      _cygwin_testing = 1;
-      if ((len = GetModuleFileName (cygwin_hmodule, buf, MAX_PATH))
-	  && len > sizeof ("new-cygwin1.dll")
-	  && strcasematch (buf + len - sizeof ("new-cygwin1.dll"),
-			   "\\new-cygwin1.dll"))
-	_cygwin_testing_magic = 0x10;
+      _main_tls = &_my_tls;
+      _main_tls->init_thread (padding, NULL);
     }
+  return &_main_tls->local_clib;
 }
 
 /* Wrap the real one, otherwise gdb gets confused about
@@ -845,80 +904,47 @@ initial_env ()
 
    UPTR is a pointer to global data that lives on the libc side of the
    line [if one distinguishes the application from the dll].  */
-
+  
 extern "C" void __stdcall
 _dll_crt0 ()
 {
-  DECLARE_TLS_STORAGE;
-  initial_env ();
-  char zeros[sizeof (fork_info->zero)] = {0};
-  static NO_COPY STARTUPINFO si;
+  extern HANDLE sync_startup;
+  extern unsigned threadfunc_ix;
+  if (threadfunc_ix)
+    /* nothing to do */;
+  else if (!sync_startup)
+    system_printf ("internal error: sync_startup not called at start.  Expect signal problems.");
+  else
+    {
+      (void) WaitForSingleObject (sync_startup, INFINITE);
+      CloseHandle (sync_startup);
+    }
+
+  if (!threadfunc_ix)
+    system_printf ("internal error: couldn't determine location of thread function on stack.  Expect signal problems.");
 
   main_environ = user_data->envptr;
   *main_environ = NULL;
 
-  init_console_handler ();
-  init_global_security ();
-  if (!DuplicateHandle (GetCurrentProcess (), GetCurrentProcess (),
-		       GetCurrentProcess (), &hMainProc, 0, FALSE,
-			DUPLICATE_SAME_ACCESS))
-    hMainProc = GetCurrentProcess ();
+  char padding[CYGTLS_PADSIZE];
+  _impure_ptr = &reent_data;
+  _impure_ptr->_stdin = &_impure_ptr->__sf[0];
+  _impure_ptr->_stdout = &_impure_ptr->__sf[1];
+  _impure_ptr->_stderr = &_impure_ptr->__sf[2];
+  _impure_ptr->_current_locale = "C";
 
-  DuplicateHandle (hMainProc, GetCurrentThread (), hMainProc,
-		   &hMainThread, 0, false, DUPLICATE_SAME_ACCESS);
-
-  GetStartupInfo (&si);
-  child_proc_info = (child_info *) si.lpReserved2;
-  if (si.cbReserved2 < EXEC_MAGIC_SIZE || !child_proc_info
-      || memcmp (child_proc_info->zero, zeros, sizeof (zeros)) != 0)
-    child_proc_info = NULL;
+  if (child_proc_info && child_proc_info->type == _PROC_FORK)
+    user_data->forkee = child_proc_info->cygpid;
   else
-    {
-      if ((child_proc_info->intro & OPROC_MAGIC_MASK) == OPROC_MAGIC_GENERIC)
-	multiple_cygwin_problem ("proc", child_proc_info->intro, 0);
-      else if (child_proc_info->intro == PROC_MAGIC_GENERIC
-	       && child_proc_info->magic != CHILD_INFO_MAGIC)
-	multiple_cygwin_problem ("proc", child_proc_info->magic,
-				 CHILD_INFO_MAGIC);
-      else if (child_proc_info->cygheap != (void *) &_cygheap_start)
-	multiple_cygwin_problem ("cygheap", (DWORD) child_proc_info->cygheap,
-				 (DWORD) &_cygheap_start);
-      unsigned should_be_cb = 0;
-      switch (child_proc_info->type)
-	{
-	  case _PROC_FORK:
-	    user_data->forkee = child_proc_info->cygpid;
-	    should_be_cb = sizeof (child_info_fork);
-	    /* fall through */;
-	  case _PROC_SPAWN:
-	  case _PROC_EXEC:
-	    if (!should_be_cb)
-	      should_be_cb = sizeof (child_info);
-	    if (should_be_cb != child_proc_info->cb)
-	      multiple_cygwin_problem ("proc size", child_proc_info->cb, should_be_cb);
-	    else if (sizeof (fhandler_union) != child_proc_info->fhandler_union_cb)
-	      multiple_cygwin_problem ("fhandler size", child_proc_info->fhandler_union_cb, sizeof (fhandler_union));
-	    else
-	      {
-		cygwin_mount_h = child_proc_info->mount_h;
-		mypid = child_proc_info->cygpid;
-		break;
-	      }
-	  default:
-	    system_printf ("unknown exec type %d", child_proc_info->type);
-	    /* intentionally fall through */
-	  case _PROC_WHOOPS:
-	    child_proc_info = NULL;
-	    break;
-	}
-    }
-  dll_crt0_1 ();
+    __sinit (_impure_ptr);
+
+  initialize_main_tls (padding);
+  dll_crt0_1 (padding);
 }
 
 void
 dll_crt0 (per_process *uptr)
 {
-  DECLARE_TLS_STORAGE;
   /* Set the local copy of the pointer into the user space. */
   if (uptr && uptr != user_data)
     {
@@ -947,44 +973,42 @@ cygwin_dll_init ()
   user_data->envptr = &envp;
   user_data->fmode_ptr = &_fmode;
 
-  dll_crt0_1 ();
+  dll_crt0_1 (NULL);
 }
 
 extern "C" void
 __main (void)
 {
-  do_global_ctors (user_data->ctors, FALSE);
+  do_global_ctors (user_data->ctors, false);
+  atexit (do_global_dtors);
 }
 
-enum exit_states
-  {
-    ES_NOT_EXITING = 0,
-    ES_THREADTERM,
-    ES_SIGNAL,
-    ES_CLOSEALL,
-    ES_SIGPROCTERMINATE,
-    ES_TITLE,
-    ES_HUP_PGRP,
-    ES_HUP_SID,
-    ES_TTY_TERMINATE,
-    ES_EVENTS_TERMINATE
-  };
-
 exit_states NO_COPY exit_state;
 extern CRITICAL_SECTION exit_lock;
 
-extern "C" void __stdcall
+void __stdcall
 do_exit (int status)
 {
-  EnterCriticalSection (&exit_lock);
-  UINT n = (UINT) status;
-
-  syscall_printf ("do_exit (%d)", n);
+  syscall_printf ("do_exit (%d), exit_state %d", status, exit_state);
 
+#ifdef NEWVFORK
   vfork_save *vf = vfork_storage.val ();
   if (vf != NULL && vf->pid < 0)
-    vf->restore_exit (status);
+    {
+      exit_state = ES_NOT_EXITING;
+      vf->restore_exit (status);
+    }
+#endif
 
+  EnterCriticalSection (&exit_lock);
+  muto::set_exiting_thread ();
+  if (exit_state < ES_EVENTS_TERMINATE)
+    {
+      exit_state = ES_EVENTS_TERMINATE;
+      events_terminate ();
+    }
+
+  UINT n = (UINT) status;
   if (exit_state < ES_THREADTERM)
     {
       exit_state = ES_THREADTERM;
@@ -1030,9 +1054,13 @@ do_exit (int status)
       /* Kill orphaned children on group leader exit */
       if (myself->has_pgid_children && myself->pid == myself->pgid)
 	{
+	  siginfo_t si;
+	  si.si_signo = -SIGHUP;
+	  si.si_code = SI_KERNEL;
+	  si.si_pid = si.si_uid = si.si_errno = 0;
 	  sigproc_printf ("%d == pgrp %d, send SIG{HUP,CONT} to stopped children",
 			  myself->pid, myself->pgid);
-	  kill_pgrp (myself->pgid, -SIGHUP);
+	  kill_pgrp (myself->pgid, si);
 	}
     }
 
@@ -1059,12 +1087,6 @@ do_exit (int status)
       tty_terminate ();
     }
 
-  if (exit_state < ES_EVENTS_TERMINATE)
-    {
-      exit_state = ES_EVENTS_TERMINATE;
-      events_terminate ();
-    }
-
   minimal_printf ("winpid %d, exit %d", GetCurrentProcessId (), n);
   myself->exit (n);
 }
@@ -1104,7 +1126,8 @@ __api_fatal (const char *fmt, ...)
   va_list ap;
 
   va_start (ap, fmt);
-  __small_vsprintf (buf, fmt, ap);
+  int n = __small_sprintf (buf, "%P (%u): *** ", cygwin_pid (GetCurrentProcessId ()));
+  __small_vsprintf (buf + n, fmt, ap);
   va_end (ap);
   strcat (buf, "\n");
   int len = strlen (buf);
