@@ -1,6 +1,6 @@
 /* dtable.cc: file descriptor support.
 
-   Copyright 1996, 1997, 1998, 1999, 2000, 2001, 2002, 2003 Red Hat, Inc.
+   Copyright 1996, 1997, 1998, 1999, 2000, 2001, 2002, 2003, 2004 Red Hat, Inc.
 
 This file is part of Cygwin.
 
@@ -26,25 +26,28 @@ details. */
 #include "cygerrno.h"
 #include "perprocess.h"
 #include "security.h"
-#include "fhandler.h"
 #include "path.h"
+#include "fhandler.h"
 #include "dtable.h"
 #include "cygheap.h"
 #include "ntdll.h"
+#include "tty.h"
+
+static const char NO_COPY unknown_file[] = "some disk file";
 
 static const NO_COPY DWORD std_consts[] = {STD_INPUT_HANDLE, STD_OUTPUT_HANDLE,
 					   STD_ERROR_HANDLE};
 
 static const char *handle_to_fn (HANDLE, char *);
 
-static const char NO_COPY unknown_file[] = "some disk file";
-
 /* Set aside space for the table of fds */
 void
-dtable_init (void)
+dtable_init ()
 {
   if (!cygheap->fdtab.size)
     cygheap->fdtab.extend (NOFILE_INCR);
+  cygheap->fdtab.init_lock ();
+
 }
 
 void __stdcall
@@ -56,6 +59,12 @@ set_std_handle (int fd)
     SetStdHandle (std_consts[fd], cygheap->fdtab[fd]->get_output_handle ());
 }
 
+void
+dtable::init_lock ()
+{
+  new_muto (lock_cs);
+}
+
 int
 dtable::extend (int howmuch)
 {
@@ -105,12 +114,12 @@ dtable::get_debugger_info ()
       for (int i = 0; i < 3; i++)
 	if (std[i][0])
 	  {
-	    path_conv pc;
 	    HANDLE h = GetStdHandle (std_consts[i]);
-	    fhandler_base *fh = build_fhandler_from_name (i, std[i], NULL, pc);
+	    fhandler_base *fh = build_fh_name (std[i]);
 	    if (!fh)
 	      continue;
-	    if (!fh->open (&pc, (i ? O_WRONLY : O_RDONLY) | O_BINARY, 0777))
+	    fds[i] = fh;
+	    if (!fh->open ((i ? O_WRONLY : O_RDONLY) | O_BINARY, 0777))
 	      release (i);
 	    else
 	      CloseHandle (h);
@@ -163,10 +172,43 @@ dtable::stdio_init ()
     set_console_ctty ();
 }
 
+const int dtable::initial_archetype_size;
+
+fhandler_base *
+dtable::find_archetype (device& dev)
+{
+  for (unsigned i = 0; i < farchetype; i++)
+    if (archetypes[i]->get_device () == (unsigned) dev)
+      return archetypes[i];
+  return NULL;
+}
+
+fhandler_base **
+dtable::add_archetype ()
+{
+  if (farchetype++ >= narchetypes)
+    archetypes = (fhandler_base **) crealloc (archetypes, (narchetypes += initial_archetype_size) * sizeof archetypes[0]);
+  return archetypes + farchetype - 1;
+}
+
+void
+dtable::delete_archetype (fhandler_base *fh)
+{
+  for (unsigned i = 0; i < farchetype; i++)
+    if (fh == archetypes[i])
+      {
+	debug_printf ("deleting element %d for %s", i, fh->get_name ());
+	if (i < --farchetype)
+	  archetypes[i] = archetypes[farchetype];
+	break;
+      }
+
+  delete fh;
+}
+
 int
 dtable::find_unused_handle (int start)
 {
-  AssertResourceOwner (LOCK_FD_LIST, READ_LOCK);
   do
     {
       for (size_t i = start; i < size; i++)
@@ -183,13 +225,12 @@ dtable::release (int fd)
 {
   if (!not_open (fd))
     {
-      switch (fds[fd]->get_device ())
-	{
-	case FH_SOCKET:
-	  dec_need_fixup_before ();
-	  break;
-	}
+      if (fds[fd]->need_fixup_before ())
+	dec_need_fixup_before ();
+      fhandler_base *arch = fds[fd]->archetype;
       delete fds[fd];
+      if (arch && !arch->usecount)
+	cygheap->fdtab.delete_archetype (arch);
       fds[fd] = NULL;
     }
 }
@@ -200,23 +241,24 @@ cygwin_attach_handle_to_fd (char *name, int fd, HANDLE handle, mode_t bin,
 {
   if (fd == -1)
     fd = cygheap->fdtab.find_unused_handle ();
-  path_conv pc;
-  fhandler_base *res = cygheap->fdtab.build_fhandler_from_name (fd, name, handle,
-								pc);
-  res->init (handle, myaccess, bin ?: pc.binmode ());
+  fhandler_base *fh = build_fh_name (name);
+  cygheap->fdtab[fd] = fh;
+  fh->init (handle, myaccess, bin ?: fh->pc_binmode ());
   return fd;
 }
 
 void
 dtable::init_std_file_from_handle (int fd, HANDLE handle)
 {
-  const char *name;
+  const char *name = NULL;
   CONSOLE_SCREEN_BUFFER_INFO buf;
   struct sockaddr sa;
   int sal = sizeof (sa);
   DCB dcb;
   unsigned bin = O_BINARY;
+  device dev;
 
+  dev.devn = 0;		/* FIXME: device */
   first_fd_for_open = 0;
 
   if (!not_open (fd))
@@ -224,54 +266,64 @@ dtable::init_std_file_from_handle (int fd, HANDLE handle)
 
   SetLastError (0);
   DWORD ft = GetFileType (handle);
-  if (ft == FILE_TYPE_UNKNOWN && GetLastError () == ERROR_INVALID_HANDLE)
-    name = NULL;
-  else
+  if (ft != FILE_TYPE_UNKNOWN || GetLastError () != ERROR_INVALID_HANDLE)
     {
       /* See if we can consoleify it */
       if (GetConsoleScreenBufferInfo (handle, &buf))
 	{
 	  if (ISSTATE (myself, PID_USETTY))
-	    name = "/dev/tty";
+	    dev.parse ("/dev/tty");
 	  else
-	    name = "/dev/conout";
+	    dev = *console_dev;
 	}
       else if (GetNumberOfConsoleInputEvents (handle, (DWORD *) &buf))
 	{
 	  if (ISSTATE (myself, PID_USETTY))
-	    name = "/dev/tty";
+	    dev.parse ("/dev/tty");
 	  else
-	    name = "/dev/conin";
+	    dev = *console_dev;
 	}
       else if (ft == FILE_TYPE_PIPE)
 	{
 	  if (fd == 0)
-	    name = "/dev/piper";
+	    dev = *piper_dev;
 	  else
-	    name = "/dev/pipew";
+	    dev = *pipew_dev;
 	}
       else if (wsock_started && getpeername ((SOCKET) handle, &sa, &sal) == 0)
-	name = "/dev/socket";
+	dev = *tcp_dev;
       else if (GetCommState (handle, &dcb))
-	name = "/dev/ttyS0"; // FIXME - determine correct device
+	dev.parse ("/dev/ttyS0");
       else
 	{
-	  name = handle_to_fn (handle, (char *) alloca (MAX_PATH + 100));
+	  name = handle_to_fn (handle, (char *) alloca (CYG_MAX_PATH + 100));
 	  bin = 0;
 	}
     }
 
-  if (!name)
+  if (!name && !dev)
     fds[fd] = NULL;
   else
     {
-      path_conv pc;
-      fhandler_base *fh = build_fhandler_from_name (fd, name, handle, pc);
+      fhandler_base *fh;
+
+      if (dev)
+	fh = build_fh_dev (dev);
+      else
+	fh = build_fh_name (name);
+
+      if (fh)
+	cygheap->fdtab[fd] = fh;
+
       if (!bin)
 	{
 	  bin = fh->get_default_fmode (O_RDWR);
-	  if (!bin && name != unknown_file)
-	    bin = pc.binmode ();
+	  if (bin)
+	    /* nothing */;
+	  else if (dev)
+	    bin = O_BINARY;
+	  else if (name != unknown_file)
+	    bin = fh->pc_binmode ();
 	}
 
       fh->init (handle, GENERIC_READ | GENERIC_WRITE, bin);
@@ -280,148 +332,149 @@ dtable::init_std_file_from_handle (int fd, HANDLE handle)
     }
 }
 
-fhandler_base *
-dtable::build_fhandler_from_name (int fd, const char *name, HANDLE handle,
-				  path_conv& pc, unsigned opt, suffix_info *si)
-{
-  pc.check (name, opt | PC_NULLEMPTY | PC_FULL | PC_POSIX, si);
-  if (pc.error)
-    {
-      set_errno (pc.error);
-      return NULL;
-    }
-
-  if (!pc.exists () && handle)
-    pc.fillin (handle);
-
-  char *posix_path = pc.return_and_clear_normalized_path ();
-  fhandler_base *fh = build_fhandler (fd, pc.get_devn (), posix_path,
-				      pc, pc.get_unitn ());
-  if (pc.issocket ()) /* Only true for files pretending an AF_LOCAL socket. */
-    {
-      fhandler_socket * fhs = (fhandler_socket *) fh;
-      fhs->set_addr_family (AF_LOCAL);
-      fhs->set_sun_path (posix_path);
-    }
-  return fh;
-}
-
-fhandler_base *
-dtable::build_fhandler (int fd, DWORD dev, const char *unix_name,
-			const char *win32_name, int unit)
-{
-  return build_fhandler (fd, dev, cstrdup (unix_name), win32_name, unit);
-}
-
 #define cnew(name) new ((void *) ccalloc (HEAP_FHANDLER, 1, sizeof (name))) name
 fhandler_base *
-dtable::build_fhandler (int fd, DWORD dev, char *unix_name,
-			const char *win32_name, int unit)
+build_fh_name (const char *name, HANDLE h, unsigned opt, suffix_info *si)
 {
-  fhandler_base *fh;
-
-  dev &= FH_DEVMASK;
-  switch (dev)
+  path_conv pc (name, opt | PC_NULLEMPTY | PC_FULL | PC_POSIX, si);
+  if (pc.error)
     {
-      case FH_TTYM:
-	fh = cnew (fhandler_tty_master) (unit);
+      fhandler_base *fh = cnew (fhandler_nodevice) ();
+      fh->set_error (pc.error);
+      set_errno (pc.error);
+      return fh;
+    }
+
+  if (!pc.exists () && h)
+    pc.fillin (h);
+
+  return build_fh_pc (pc);
+}
+
+fhandler_base *
+build_fh_dev (const device& dev, const char *unix_name)
+{
+  path_conv pc (dev);
+  if (unix_name)
+    pc.set_normalized_path (unix_name);
+  else
+    pc.set_normalized_path (dev.name);
+  return build_fh_pc (pc);
+}
+
+fhandler_base *
+build_fh_pc (path_conv& pc)
+{
+  fhandler_base *fh = NULL;
+
+    switch (pc.dev.major)
+      {
+      case DEV_TTYS_MAJOR:
+	fh = cnew (fhandler_tty_slave) ();
 	break;
-      case FH_CONSOLE:
-      case FH_CONIN:
-      case FH_CONOUT:
-	fh = cnew (fhandler_console) ();
+      case DEV_TTYM_MAJOR:
+	fh = cnew (fhandler_tty_master) ();
 	break;
-      case FH_PTYM:
-	fh = cnew (fhandler_pty_master) ();
+      case DEV_CYGDRIVE_MAJOR:
+	fh = cnew (fhandler_cygdrive) ();
 	break;
-      case FH_TTYS:
-	if (unit < 0)
-	  fh = cnew (fhandler_tty_slave) ();
-	else
-	  fh = cnew (fhandler_tty_slave) (unit);
+      case DEV_FLOPPY_MAJOR:
+      case DEV_CDROM_MAJOR:
+      case DEV_SD_MAJOR:
+	fh = cnew (fhandler_dev_floppy) ();
 	break;
-      case FH_WINDOWS:
-	fh = cnew (fhandler_windows) ();
+      case DEV_TAPE_MAJOR:
+	fh = cnew (fhandler_dev_tape) ();
 	break;
-      case FH_SERIAL:
-	fh = cnew (fhandler_serial) (unit);
-	break;
-      case FH_PIPE:
-      case FH_PIPER:
-      case FH_PIPEW:
-	fh = cnew (fhandler_pipe) (dev);
-	break;
-      case FH_SOCKET:
-	if ((fh = cnew (fhandler_socket) (unit)))
-	  inc_need_fixup_before ();
-	break;
-      case FH_DISK:
-	fh = cnew (fhandler_disk_file) ();
-	break;
-      case FH_CYGDRIVE:
-	fh = cnew (fhandler_cygdrive) (unit);
-	break;
-      case FH_FLOPPY:
-	fh = cnew (fhandler_dev_floppy) (unit);
-	break;
-      case FH_TAPE:
-	fh = cnew (fhandler_dev_tape) (unit);
-	break;
-      case FH_NULL:
-	fh = cnew (fhandler_dev_null) ();
-	break;
-      case FH_ZERO:
-	fh = cnew (fhandler_dev_zero) ();
-	break;
-      case FH_RANDOM:
-	fh = cnew (fhandler_dev_random) (unit);
-	break;
-      case FH_MEM:
-	fh = cnew (fhandler_dev_mem) (unit);
-	break;
-      case FH_CLIPBOARD:
-	fh = cnew (fhandler_dev_clipboard) ();
-	break;
-      case FH_OSS_DSP:
-	fh = cnew (fhandler_dev_dsp) ();
-	break;
-      case FH_PROC:
-	fh = cnew (fhandler_proc) ();
-	break;
-      case FH_REGISTRY:
-	fh = cnew (fhandler_registry) ();
-	break;
-      case FH_PROCESS:
-	fh = cnew (fhandler_process) ();
+      case DEV_SERIAL_MAJOR:
+	fh = cnew (fhandler_serial) ();
 	break;
       default:
-	system_printf ("internal error -- unknown device - %p", dev);
-	fh = NULL;
-    }
-
-  if (unix_name)
-    {
-      char new_win32_name[strlen (unix_name) + 1];
-      if (!win32_name)
-	{
-	  char *p;
-	  /* FIXME: ? Should we call win32_device_name here?
-	     It seems like overkill, but... */
-	  win32_name = strcpy (new_win32_name, unix_name);
-	  for (p = (char *) win32_name; (p = strchr (p, '/')); p++)
-	    *p = '\\';
+	switch (pc.dev)
+	  {
+	  case FH_CONSOLE:
+	  case FH_CONIN:
+	  case FH_CONOUT:
+	    fh = cnew (fhandler_console) ();
+	    break;
+	  case FH_PTYM:
+	    fh = cnew (fhandler_pty_master) ();
+	    break;
+	  case FH_WINDOWS:
+	    fh = cnew (fhandler_windows) ();
+	    break;
+	  case FH_FIFO:
+	    fh = cnew (fhandler_fifo) ();
+	    break;
+	  case FH_PIPE:
+	  case FH_PIPER:
+	  case FH_PIPEW:
+	    fh = cnew (fhandler_pipe) ();
+	    break;
+	  case FH_TCP:
+	  case FH_UDP:
+	  case FH_ICMP:
+	  case FH_UNIX:
+	  case FH_STREAM:
+	  case FH_DGRAM:
+	    fh = cnew (fhandler_socket) ();
+	    break;
+	  case FH_FS:
+	    fh = cnew (fhandler_disk_file) ();
+	    break;
+	  case FH_NULL:
+	    fh = cnew (fhandler_dev_null) ();
+	    break;
+	  case FH_ZERO:
+	    fh = cnew (fhandler_dev_zero) ();
+	    break;
+	  case FH_RANDOM:
+	  case FH_URANDOM:
+	    fh = cnew (fhandler_dev_random) ();
+	    break;
+	  case FH_MEM:
+	  case FH_PORT:
+	    fh = cnew (fhandler_dev_mem) ();
+	    break;
+	  case FH_CLIPBOARD:
+	    fh = cnew (fhandler_dev_clipboard) ();
+	    break;
+	  case FH_OSS_DSP:
+	    fh = cnew (fhandler_dev_dsp) ();
+	    break;
+	  case FH_PROC:
+	    fh = cnew (fhandler_proc) ();
+	    break;
+	  case FH_REGISTRY:
+	    fh = cnew (fhandler_registry) ();
+	    break;
+	  case FH_PROCESS:
+	    fh = cnew (fhandler_process) ();
+	    break;
+	  case FH_TTY:
+	    {
+	      if (myself->ctty == TTY_CONSOLE)
+		fh = cnew (fhandler_console) ();
+	      else if (myself->ctty >= 0)
+		fh = cnew (fhandler_tty_slave) ();
+	      break;
+	    }
 	}
-      if (!fh->set_name (unix_name, win32_name, fh->get_unit ()))
-	return NULL;
-    }
-  debug_printf ("fd %d, fh %p", fd, fh);
-  return fd >= 0 ? (fds[fd] = fh) : fh;
+      }
+
+  if (!fh)
+    fh = cnew (fhandler_nodevice) ();
+
+  fh->set_name (pc);
+
+  debug_printf ("fh %p", fh);
+  return fh;
 }
 
 fhandler_base *
 dtable::dup_worker (fhandler_base *oldfh)
 {
-  fhandler_base *newfh = build_fhandler (-1, oldfh->get_device ());
+  fhandler_base *newfh = build_fh_pc (oldfh->pc);
   *newfh = *oldfh;
   newfh->set_io_handle (NULL);
   if (oldfh->dup (newfh))
@@ -445,7 +498,7 @@ dtable::dup2 (int oldfd, int newfd)
 
   MALLOC_CHECK;
   debug_printf ("dup2 (%d, %d)", oldfd, newfd);
-  SetResourceLock (LOCK_FD_LIST, WRITE_LOCK | READ_LOCK, "dup");
+  lock ();
 
   if (not_open (oldfd))
     {
@@ -494,12 +547,25 @@ dtable::dup2 (int oldfd, int newfd)
 
 done:
   MALLOC_CHECK;
-  ReleaseResourceLock (LOCK_FD_LIST, WRITE_LOCK | READ_LOCK, "dup");
+  unlock ();
   syscall_printf ("%d = dup2 (%d, %d)", res, oldfd, newfd);
 
   return res;
 }
 
+fhandler_fifo *
+dtable::find_fifo (ATOM hill)
+{
+  lock ();
+  for (unsigned i = 0; i < size; i++)
+    {
+      fhandler_base *fh = fds[i];
+      if (fh && fh->isfifo () && ((fhandler_fifo *) fh)->get_atom () == hill)
+	return (fhandler_fifo *) fh;
+    }
+  return NULL;
+}
+
 select_record *
 dtable::select_read (int fd, select_record *s)
 {
@@ -556,7 +622,7 @@ dtable::select_except (int fd, select_record *s)
 void
 dtable::fixup_before_fork (DWORD target_proc_id)
 {
-  SetResourceLock (LOCK_FD_LIST, WRITE_LOCK | READ_LOCK, "fixup_before_fork");
+  lock ();
   fhandler_base *fh;
   for (size_t i = 0; i < size; i++)
     if ((fh = fds[i]) != NULL)
@@ -564,13 +630,13 @@ dtable::fixup_before_fork (DWORD target_proc_id)
 	debug_printf ("fd %d (%s)", i, fh->get_name ());
 	fh->fixup_before_fork_exec (target_proc_id);
       }
-  ReleaseResourceLock (LOCK_FD_LIST, WRITE_LOCK | READ_LOCK, "fixup_before_fork");
+  unlock ();
 }
 
 void
 dtable::fixup_before_exec (DWORD target_proc_id)
 {
-  SetResourceLock (LOCK_FD_LIST, WRITE_LOCK | READ_LOCK, "fixup_before_exec");
+  lock ();
   fhandler_base *fh;
   for (size_t i = 0; i < size; i++)
     if ((fh = fds[i]) != NULL && !fh->get_close_on_exec ())
@@ -578,34 +644,39 @@ dtable::fixup_before_exec (DWORD target_proc_id)
 	debug_printf ("fd %d (%s)", i, fh->get_name ());
 	fh->fixup_before_fork_exec (target_proc_id);
       }
-  ReleaseResourceLock (LOCK_FD_LIST, WRITE_LOCK | READ_LOCK, "fixup_before_exec");
+  unlock ();
 }
 
 void
 dtable::set_file_pointers_for_exec ()
 {
-  SetResourceLock (LOCK_FD_LIST, WRITE_LOCK | READ_LOCK, "set_file_pointers_for_exec");
+  lock ();
   fhandler_base *fh;
   for (size_t i = 0; i < size; i++)
     if ((fh = fds[i]) != NULL && fh->get_flags () & O_APPEND)
       SetFilePointer (fh->get_handle (), 0, 0, FILE_END);
-  ReleaseResourceLock (LOCK_FD_LIST, WRITE_LOCK | READ_LOCK, "fixup_before_exec");
+  unlock ();
 }
 
 void
-dtable::fixup_after_exec (HANDLE parent)
+dtable::fixup_after_exec ()
 {
   first_fd_for_open = 0;
   fhandler_base *fh;
+  cygheap->fdtab.init_lock ();
   for (size_t i = 0; i < size; i++)
     if ((fh = fds[i]) != NULL)
       {
 	fh->clear_readahead ();
 	if (fh->get_close_on_exec ())
-	  release (i);
+	  {
+	    if (fh->archetype)
+	      fh->close ();
+	    release (i);
+	  }
 	else
 	  {
-	    fh->fixup_after_exec (parent);
+	    fh->fixup_after_exec ();
 	    if (i == 0)
 	      SetStdHandle (std_consts[i], fh->get_io_handle ());
 	    else if (i <= 2)
@@ -618,6 +689,7 @@ void
 dtable::fixup_after_fork (HANDLE parent)
 {
   fhandler_base *fh;
+  cygheap->fdtab.init_lock ();
   for (size_t i = 0; i < size; i++)
     if ((fh = fds[i]) != NULL)
       {
@@ -633,16 +705,23 @@ dtable::fixup_after_fork (HANDLE parent)
       }
 }
 
+#ifdef NEWVFORK
 int
 dtable::vfork_child_dup ()
 {
   fhandler_base **newtable;
-  SetResourceLock (LOCK_FD_LIST, WRITE_LOCK | READ_LOCK, "dup");
+  lock ();
   newtable = (fhandler_base **) ccalloc (HEAP_ARGV, size, sizeof (fds[0]));
   int res = 1;
 
   /* Remove impersonation */
   cygheap->user.deimpersonate ();
+  if (cygheap->ctty)
+    {
+      cygheap->ctty->usecount++;
+      cygheap->open_fhs++;
+      report_tty_counts (cygheap->ctty, "vfork dup", "incremented ", "");
+    }
 
   for (size_t i = 0; i < size; i++)
     if (not_open (i))
@@ -663,22 +742,30 @@ out:
   /* Restore impersonation */
   cygheap->user.reimpersonate ();
 
-  ReleaseResourceLock (LOCK_FD_LIST, WRITE_LOCK | READ_LOCK, "dup");
+  unlock ();
   return 1;
 }
 
 void
 dtable::vfork_parent_restore ()
 {
-  SetResourceLock (LOCK_FD_LIST, WRITE_LOCK | READ_LOCK, "restore");
+  lock ();
 
+  fhandler_tty_slave *ctty_on_hold = cygheap->ctty_on_hold;
   close_all_files ();
   fhandler_base **deleteme = fds;
   fds = fds_on_hold;
   fds_on_hold = NULL;
   cfree (deleteme);
+  unlock ();
+
+  if (cygheap->ctty != ctty_on_hold)
+    {
+      cygheap->ctty = ctty_on_hold;		// revert
+      cygheap->ctty->close ();			// Undo previous bump of this archetype
+    }
+  cygheap->ctty_on_hold = NULL;
 
-  ReleaseResourceLock (LOCK_FD_LIST, WRITE_LOCK | READ_LOCK, "restore");
   return;
 }
 
@@ -692,16 +779,16 @@ dtable::vfork_child_fixup ()
   fds = fds_on_hold;
 
   fhandler_base *fh;
-  for (int i = 0; i < (int) cygheap->fdtab.size; i++)
-    if ((fh = cygheap->fdtab[i]) != NULL)
+  for (int i = 0; i < (int) size; i++)
+    if ((fh = fds[i]) != NULL)
       {
 	fh->clear_readahead ();
-	if (fh->get_close_on_exec ())
+	if (!fh->archetype && fh->get_close_on_exec ())
 	  release (i);
 	else
 	  {
 	    fh->close ();
-	    cygheap->fdtab.release (i);
+	    release (i);
 	  }
       }
 
@@ -709,8 +796,15 @@ dtable::vfork_child_fixup ()
   cfree (fds_on_hold);
   fds_on_hold = NULL;
 
+  if (cygheap->ctty_on_hold)
+    {
+      cygheap->ctty_on_hold->close ();
+      cygheap->ctty_on_hold = NULL;
+    }
+
   return;
 }
+#endif /*NEWVFORK*/
 
 #define DEVICE_PREFIX "\\device\\"
 #define DEVICE_PREFIX_LEN sizeof (DEVICE_PREFIX) - 1
@@ -746,24 +840,22 @@ handle_to_fn (HANDLE h, char *posix_fn)
 
   ntfn->Name.Buffer[ntfn->Name.Length / sizeof (WCHAR)] = 0;
 
-  char win32_fn[MAX_PATH + 100];
+  char win32_fn[CYG_MAX_PATH + 100];
   sys_wcstombs (win32_fn, ntfn->Name.Buffer, ntfn->Name.Length);
   debug_printf ("nt name '%s'", win32_fn);
   if (!strncasematch (win32_fn, DEVICE_PREFIX, DEVICE_PREFIX_LEN)
       || !QueryDosDevice (NULL, fnbuf, sizeof (fnbuf)))
     return strcpy (posix_fn, win32_fn);
 
-  char *p = strchr (win32_fn + DEVICE_PREFIX_LEN, '\\');
-  if (!p)
-    p = strchr (win32_fn + DEVICE_PREFIX_LEN, '\0');
+  char *p = strechr (win32_fn + DEVICE_PREFIX_LEN, '\\');
 
   int n = p - win32_fn;
   int maxmatchlen = 0;
   char *maxmatchdos = NULL;
   for (char *s = fnbuf; *s; s = strchr (s, '\0') + 1)
     {
-      char device[MAX_PATH + 10];
-      device[MAX_PATH + 9] = '\0';
+      char device[CYG_MAX_PATH + 10];
+      device[CYG_MAX_PATH + 9] = '\0';
       if (strchr (s, ':') == NULL)
 	continue;
       if (!QueryDosDevice (s, device, sizeof (device) - 1))
