@@ -15,6 +15,7 @@ details. */
 #include <stdio.h>
 #include "cygerrno.h"
 #include "thread.h"
+#include "cygtls.h"
 #undef _sys_nerr
 #undef sys_nerr
 #undef _sys_errlist
@@ -101,8 +102,8 @@ static NO_COPY struct
   X (NOT_OWNER,			EPERM),
   X (END_OF_MEDIA,		ENOSPC),
   X (EOM_OVERFLOW,		ENOSPC),
-  X (BEGINNING_OF_MEDIA,	ESPIPE),
-  X (SETMARK_DETECTED,		ESPIPE),
+  X (BEGINNING_OF_MEDIA,	EIO),
+  X (SETMARK_DETECTED,		EIO),
   X (NO_DATA_DETECTED,		ENOSPC),
   X (POSSIBLE_DEADLOCK,		EDEADLOCK),
   X (CRC,			EIO),
@@ -113,6 +114,8 @@ static NO_COPY struct
   X (FILE_INVALID,		ENXIO),
   X (INVALID_ADDRESS,		EOVERFLOW),
   X (INVALID_BLOCK_LENGTH,	EIO),
+  X (BUS_RESET,			EIO),
+  X (FILEMARK_DETECTED,		EIO),
   { 0, NULL, 0}
 };
 
@@ -681,13 +684,8 @@ strerror (int errnum)
       error = "Value too large for defined data type";
       break;
     default:
-#ifdef _MT_SAFE
-      char *buf= _reent_winsup ()->_strerror_buf;
-#else
-      static NO_COPY char buf[20];
-#endif
-      __small_sprintf (buf, "error %d", errnum);
-      error = buf;
+      __small_sprintf (_my_tls.locals.strerror_buf, "error %d", errnum);
+      error = _my_tls.locals.strerror_buf;
       break;
     }
 
