@@ -1,6 +1,6 @@
 /* fhandler.cc.  See console.cc for fhandler_console functions.
 
-   Copyright 1996, 1997, 1998, 1999, 2000, 2001, 2002 Red Hat, Inc.
+   Copyright 1996, 1997, 1998, 1999, 2000, 2001, 2002, 2003, 2004 Red Hat, Inc.
 
 This file is part of Cygwin.
 
@@ -18,8 +18,8 @@ details. */
 #include "perprocess.h"
 #include "security.h"
 #include "cygwin/version.h"
-#include "fhandler.h"
 #include "path.h"
+#include "fhandler.h"
 #include "dtable.h"
 #include "cygheap.h"
 #include "shared_info.h"
@@ -35,11 +35,10 @@ struct __cygwin_perfile *perfile_table;
 DWORD binmode;
 
 inline fhandler_base&
-fhandler_base::operator =(fhandler_base &x)
+fhandler_base::operator =(fhandler_base& x)
 {
   memcpy (this, &x, sizeof *this);
-  unix_path_name = x.unix_path_name ? cstrdup (x.unix_path_name) : NULL;
-  win32_path_name = x.win32_path_name ? cstrdup (x.win32_path_name) : NULL;
+  pc.set_normalized_path (x.pc.normalized_path);
   rabuf = NULL;
   ralen = 0;
   raixget = 0;
@@ -144,63 +143,13 @@ fhandler_base::get_readahead_into_buffer (char *buf, size_t buflen)
   return copied_chars;
 }
 
-/* Record the file name.
-   Filenames are used mostly for debugging messages, and it's hoped that
-   in cases where the name is really required, the filename wouldn't ever
-   be too long (e.g. devices or some such).
-   The unix_path_name is also used by virtual fhandlers.  */
-bool
-fhandler_base::set_name (const char *unix_path, const char *win32_path, int unit)
+/* Record the file name. and name hash */
+void
+fhandler_base::set_name (path_conv &in_pc)
 {
-  if (unix_path == NULL || !*unix_path)
-    return false;
-
-  if (win32_path)
-    win32_path_name = cstrdup (win32_path);
-  else
-    {
-      const char *fmt = get_native_name ();
-      char *w =  (char *) cmalloc (HEAP_STR, strlen (fmt) + 16);
-      if (w)
-	__small_sprintf (w, fmt, unit);
-      win32_path_name = w;
-    }
-
-  if (win32_path_name == NULL)
-    {
-      system_printf ("fatal error. strdup failed");
-      set_errno (ENOMEM);
-      return false;
-    }
-
-  assert (unix_path_name == NULL);
-  /* FIXME: This isn't really right.  It ignores the first argument if we're
-     building names for a device and just converts the device name from the
-     win32 name since it has theoretically been previously detected by
-     path_conv. Ideally, we should pass in a format string and build the
-     unix_path, too. */
-  if (!is_device () || *win32_path_name != '\\')
-    unix_path_name = unix_path;
-  else
-    {
-      char *p = cstrdup (win32_path_name);
-      unix_path_name = p;
-      if (p)
-	while ((p = strchr (p, '\\')) != NULL)
-	  *p++ = '/';
-      if (unix_path)
-	cfree ((void *) unix_path);
-    }
-
-  if (unix_path_name == NULL)
-    {
-      system_printf ("fatal error. strdup failed");
-      free ((void *) win32_path_name);
-      set_errno (ENOMEM);
-      return false;
-    }
-  namehash = hash_path_name (0, win32_path_name);
-  return true;
+  memcpy (&pc, &in_pc, in_pc.size ());
+  pc.set_normalized_path (in_pc.normalized_path);
+  namehash = hash_path_name (0, get_win32_name ());
 }
 
 /* Detect if we are sitting at EOF for conditions where Windows
@@ -308,7 +257,7 @@ fhandler_base::raw_read (void *ptr, size_t& ulen)
 	      break;
 	    }
 	default:
-	  syscall_printf ("ReadFile %s failed, %E", unix_path_name);
+	  syscall_printf ("ReadFile %s failed, %E", get_name ());
 	  __seterrno_from_win_error (errcode);
 	  bytes_read = (size_t) -1;
 	  break;
@@ -324,7 +273,7 @@ fhandler_base::raw_write (const void *ptr, size_t len)
 {
   DWORD bytes_written;
 
-  if (!WriteFile (get_handle (), ptr, len, &bytes_written, 0))
+  if (!WriteFile (get_output_handle (), ptr, len, &bytes_written, 0))
     {
       if (GetLastError () == ERROR_DISK_FULL && bytes_written > 0)
 	return bytes_written;
@@ -367,9 +316,115 @@ fhandler_base::get_default_fmode (int flags)
   return fmode;
 }
 
+bool
+fhandler_base::device_access_denied (int flags)
+{
+  int mode = 0;
+
+  if (flags & O_RDWR)
+    mode |= R_OK | W_OK;
+  if (flags & (O_WRONLY | O_APPEND))
+    mode |= W_OK;
+  if (!mode)
+    mode |= R_OK;
+
+  return fhaccess (mode);
+}
+
+int
+fhandler_base::fhaccess (int flags)
+{
+  int res = -1;
+  if (error ())
+    {
+      set_errno (error ());
+      goto done;
+    }
+
+  if (!exists ())
+    {
+      set_errno (ENOENT);
+      goto done;
+    }
+
+  if (!(flags & (R_OK | W_OK | X_OK)))
+    return 0;
+
+  if (is_fs_special ())
+    /* short circuit */;
+  else if (has_attribute (FILE_ATTRIBUTE_READONLY) && (flags & W_OK))
+    goto eaccess_done;
+  else if (has_acls () && allow_ntsec)
+    {
+      res = check_file_access (get_win32_name (), flags);
+      goto done;
+    }
+
+  struct __stat64 st;
+  if (fstat (&st))
+    goto done;
+
+  if (flags & R_OK)
+    {
+      if (st.st_uid == myself->uid)
+	{
+	  if (!(st.st_mode & S_IRUSR))
+	    goto eaccess_done;
+	}
+      else if (st.st_gid == myself->gid)
+	{
+	  if (!(st.st_mode & S_IRGRP))
+	    goto eaccess_done;
+	}
+      else if (!(st.st_mode & S_IROTH))
+	goto eaccess_done;
+    }
+
+  if (flags & W_OK)
+    {
+      if (st.st_uid == myself->uid)
+	{
+	  if (!(st.st_mode & S_IWUSR))
+	    goto eaccess_done;
+	}
+      else if (st.st_gid == myself->gid)
+	{
+	  if (!(st.st_mode & S_IWGRP))
+	    goto eaccess_done;
+	}
+      else if (!(st.st_mode & S_IWOTH))
+	goto eaccess_done;
+    }
+
+  if (flags & X_OK)
+    {
+      if (st.st_uid == myself->uid)
+	{
+	  if (!(st.st_mode & S_IXUSR))
+	    goto eaccess_done;
+	}
+      else if (st.st_gid == myself->gid)
+	{
+	  if (!(st.st_mode & S_IXGRP))
+	    goto eaccess_done;
+	}
+      else if (!(st.st_mode & S_IXOTH))
+	goto eaccess_done;
+    }
+
+  res = 0;
+  goto done;
+
+eaccess_done:
+  set_errno (EACCES);
+done:
+  debug_printf ("returning %d", res);
+  return res;
+}
+
 /* Open system call handler function. */
 int
-fhandler_base::open (path_conv *pc, int flags, mode_t mode)
+fhandler_base::open (int flags, mode_t mode)
 {
   int res = 0;
   HANDLE x;
@@ -377,6 +432,7 @@ fhandler_base::open (path_conv *pc, int flags, mode_t mode)
   int shared;
   int creation_distribution;
   SECURITY_ATTRIBUTES sa = sec_none;
+  security_descriptor sd;
 
   syscall_printf ("(%s, %p) query_open %d", get_win32_name (), flags, get_query_open ());
 
@@ -388,7 +444,7 @@ fhandler_base::open (path_conv *pc, int flags, mode_t mode)
 
   if (get_query_open ())
     access = 0;
-  else if (get_device () == FH_TAPE)
+  else if (get_major () == DEV_TAPE_MAJOR)
     access = GENERIC_READ | GENERIC_WRITE;
   else if ((flags & (O_RDONLY | O_WRONLY | O_RDWR)) == O_RDONLY)
     access = GENERIC_READ;
@@ -398,7 +454,7 @@ fhandler_base::open (path_conv *pc, int flags, mode_t mode)
     access = GENERIC_READ | GENERIC_WRITE;
 
   /* Allow reliable lseek on disk devices. */
-  if (get_device () == FH_FLOPPY)
+  if (get_major () == DEV_FLOPPY_MAJOR)
     access |= GENERIC_READ;
 
   /* FIXME: O_EXCL handling?  */
@@ -427,11 +483,11 @@ fhandler_base::open (path_conv *pc, int flags, mode_t mode)
   file_attributes = FILE_ATTRIBUTE_NORMAL;
   if (flags & O_DIROPEN)
     file_attributes |= FILE_FLAG_BACKUP_SEMANTICS;
-  if (get_device () == FH_SERIAL)
+  if (get_major () == DEV_SERIAL_MAJOR)
     file_attributes |= FILE_FLAG_OVERLAPPED;
 
 #ifdef HIDDEN_DOT_FILES
-  if (flags & O_CREAT && get_device () == FH_DISK)
+  if (flags & O_CREAT && get_device () == FH_FS)
     {
       char *c = strrchr (get_win32_name (), '\\');
       if ((c && c[1] == '.') || *get_win32_name () == '.')
@@ -443,7 +499,7 @@ fhandler_base::open (path_conv *pc, int flags, mode_t mode)
      share returns some handle, even if file doesn't exist. This code
      works around this bug. */
   if (get_query_open () && isremote () &&
-      creation_distribution == OPEN_EXISTING && pc && !pc->exists ())
+      creation_distribution == OPEN_EXISTING && !pc.exists ())
     {
       set_errno (ENOENT);
       goto done;
@@ -455,23 +511,23 @@ fhandler_base::open (path_conv *pc, int flags, mode_t mode)
 
   /* If the file should actually be created and ntsec is on,
      set files attributes. */
-  if (flags & O_CREAT && get_device () == FH_DISK && allow_ntsec && has_acls ())
-    set_security_attribute (mode, &sa, alloca (4096), 4096);
+  if (flags & O_CREAT && get_device () == FH_FS && allow_ntsec && has_acls ())
+    set_security_attribute (mode, &sa, sd);
 
   x = CreateFile (get_win32_name (), access, shared, &sa, creation_distribution,
 		  file_attributes, 0);
 
   if (x == INVALID_HANDLE_VALUE)
     {
-      if (!wincap.can_open_directories () && pc && pc->isdir ())
-       {
-	 if (flags & (O_CREAT | O_EXCL) == (O_CREAT | O_EXCL))
-	   set_errno (EEXIST);
-	 else if (flags & (O_WRONLY | O_RDWR))
-	   set_errno (EISDIR);
-	 else
-	   set_nohandle (true);
-       }
+      if (!wincap.can_open_directories () && pc.isdir ())
+	{
+	  if (flags & (O_CREAT | O_EXCL) == (O_CREAT | O_EXCL))
+	    set_errno (EEXIST);
+	  else if (flags & (O_WRONLY | O_RDWR))
+	    set_errno (EISDIR);
+	  else
+	    set_nohandle (true);
+	}
       else if (GetLastError () == ERROR_INVALID_HANDLE)
        set_errno (ENOENT);
       else
@@ -485,7 +541,7 @@ fhandler_base::open (path_conv *pc, int flags, mode_t mode)
 		  creation_distribution, file_attributes);
 
   set_io_handle (x);
-  set_flags (flags, pc ? pc->binmode () : 0);
+  set_flags (flags, pc.binmode ());
 
   res = 1;
   set_open_status ();
@@ -615,7 +671,7 @@ fhandler_base::write (const void *ptr, size_t len)
   int res;
 
   if (get_append_p ())
-    SetFilePointer (get_handle (), 0, 0, FILE_END);
+    SetFilePointer (get_output_handle (), 0, 0, FILE_END);
   else if (get_did_lseek ())
     {
       _off64_t actual_length, current_position;
@@ -624,10 +680,10 @@ fhandler_base::write (const void *ptr, size_t len)
 
       set_did_lseek (0); /* don't do it again */
 
-      actual_length = GetFileSize (get_handle (), &size_high);
+      actual_length = GetFileSize (get_output_handle (), &size_high);
       actual_length += ((_off64_t) size_high) << 32;
 
-      current_position = SetFilePointer (get_handle (), 0, &pos_high,
+      current_position = SetFilePointer (get_output_handle (), 0, &pos_high,
 					 FILE_CURRENT);
       current_position += ((_off64_t) pos_high) << 32;
 
@@ -640,7 +696,7 @@ fhandler_base::write (const void *ptr, size_t len)
 	         is writing after a long seek beyond EOF, convert the file to
 		 a sparse file. */
 	      DWORD dw;
-	      HANDLE h = get_handle ();
+	      HANDLE h = get_output_handle ();
 	      BOOL r = DeviceIoControl (h, FSCTL_SET_SPARSE, NULL, 0, NULL,
 	      				0, &dw, NULL);
 	      syscall_printf ("%d = DeviceIoControl(%p, FSCTL_SET_SPARSE, "
@@ -657,20 +713,20 @@ fhandler_base::write (const void *ptr, size_t len)
 	      char zeros[512];
 	      int number_of_zeros_to_write = current_position - actual_length;
 	      memset (zeros, 0, 512);
-	      SetFilePointer (get_handle (), 0, NULL, FILE_END);
+	      SetFilePointer (get_output_handle (), 0, NULL, FILE_END);
 	      while (number_of_zeros_to_write > 0)
 		{
 		  DWORD zeros_this_time = (number_of_zeros_to_write > 512
 					 ? 512 : number_of_zeros_to_write);
 		  DWORD written;
-		  if (!WriteFile (get_handle (), zeros, zeros_this_time,
+		  if (!WriteFile (get_output_handle (), zeros, zeros_this_time,
 				  &written, NULL))
 		    {
 		      __seterrno ();
 		      if (get_errno () == EPIPE)
 			raise (SIGPIPE);
 		      /* This might fail, but it's the best we can hope for */
-		      SetFilePointer (get_handle (), current_position, NULL,
+		      SetFilePointer (get_output_handle (), current_position, NULL,
 				      FILE_BEGIN);
 		      return -1;
 
@@ -679,7 +735,7 @@ fhandler_base::write (const void *ptr, size_t len)
 		    {
 		      set_errno (ENOSPC);
 		      /* This might fail, but it's the best we can hope for */
-		      SetFilePointer (get_handle (), current_position, NULL,
+		      SetFilePointer (get_output_handle (), current_position, NULL,
 				      FILE_BEGIN);
 		      return -1;
 		    }
@@ -886,12 +942,12 @@ fhandler_base::lseek (_off64_t offset, int whence)
       set_readahead_valid (0);
     }
 
-  debug_printf ("lseek (%s, %D, %d)", unix_path_name, offset, whence);
+  debug_printf ("lseek (%s, %D, %d)", get_name (), offset, whence);
 
   DWORD win32_whence = whence == SEEK_SET ? FILE_BEGIN
 		       : (whence == SEEK_CUR ? FILE_CURRENT : FILE_END);
 
-  LONG off_low = ((__uint64_t) offset) & 0xffffffffLL;
+  LONG off_low = ((__uint64_t) offset) & UINT32_MAX;
   LONG *poff_high, off_high;
   if (!wincap.has_64bit_file_access ())
     poff_high = NULL;
@@ -909,6 +965,9 @@ fhandler_base::lseek (_off64_t offset, int whence)
     }
   else
     {
+      if (poff_high)
+        res += (_off64_t) *poff_high << 32;
+
       /* When next we write(), we will check to see if *this* seek went beyond
 	 the end of the file, and back-seek and fill with zeros if so - DJ */
       set_did_lseek ();
@@ -944,17 +1003,26 @@ fhandler_base::close ()
 int
 fhandler_base::ioctl (unsigned int cmd, void *buf)
 {
-  if (cmd == FIONBIO)
-    syscall_printf ("ioctl (FIONBIO, %p)", buf);
-  else
-    syscall_printf ("ioctl (%x, %p)", cmd, buf);
+  int res;
 
-  set_errno (EINVAL);
-  return -1;
+  switch (cmd)
+    {
+    case FIONBIO:
+      set_nonblocking (*(int *) buf);
+      res = 0;
+      break;
+    default:
+      set_errno (EINVAL);
+      res = -1;
+      break;
+    }
+
+  syscall_printf ("%d = ioctl (%x, %p)", res, cmd, buf);
+  return res;
 }
 
 int
-fhandler_base::lock (int, struct flock *)
+fhandler_base::lock (int, struct __flock64 *)
 {
   set_errno (EINVAL);
   return -1;
@@ -994,9 +1062,13 @@ rootdir (char *full_path)
 }
 
 int __stdcall
-fhandler_base::fstat (struct __stat64 *buf, path_conv *)
+fhandler_base::fstat (struct __stat64 *buf)
 {
   debug_printf ("here");
+
+  if (is_fs_special ())
+    return fstat_fs (buf);
+
   switch (get_device ())
     {
     case FH_PIPE:
@@ -1008,9 +1080,6 @@ fhandler_base::fstat (struct __stat64 *buf, path_conv *)
     case FH_PIPER:
       buf->st_mode = S_IFIFO | STD_RBITS;
       break;
-    case FH_FLOPPY:
-      buf->st_mode = S_IFBLK | STD_RBITS | STD_WBITS | S_IWGRP | S_IWOTH;
-      break;
     default:
       buf->st_mode = S_IFCHR | STD_RBITS | STD_WBITS | S_IWGRP | S_IWOTH;
       break;
@@ -1066,6 +1135,7 @@ fhandler_base::dup (fhandler_base *child)
 	  return -1;
 	}
 
+      VerifyHandle (nh);
       child->set_io_handle (nh);
     }
   return 0;
@@ -1102,7 +1172,7 @@ int fhandler_base::fcntl (int cmd, void *arg)
 	   Set only the flag that has been passed in.  If both are set, just
 	   record O_NONBLOCK.   */
 	if ((new_flags & OLD_O_NDELAY) && (new_flags & O_NONBLOCK))
-	  new_flags = O_NONBLOCK;
+	  new_flags &= ~OLD_O_NDELAY;
 	set_flags ((get_flags () & ~allowed_flags) | new_flags);
       }
       res = 0;
@@ -1110,7 +1180,7 @@ int fhandler_base::fcntl (int cmd, void *arg)
     case F_GETLK:
     case F_SETLK:
     case F_SETLKW:
-      res = lock (cmd, (struct flock *) arg);
+      res = lock (cmd, (struct __flock64 *) arg);
       break;
     default:
       set_errno (EINVAL);
@@ -1186,8 +1256,8 @@ fhandler_base::operator delete (void *p)
 }
 
 /* Normal I/O constructor */
-fhandler_base::fhandler_base (DWORD devtype, int unit):
-  status (devtype),
+fhandler_base::fhandler_base ():
+  status (0),
   access (0),
   io_handle (NULL),
   namehash (0),
@@ -1197,31 +1267,26 @@ fhandler_base::fhandler_base (DWORD devtype, int unit):
   raixget (0),
   raixput (0),
   rabuflen (0),
-  unix_path_name (NULL),
-  win32_path_name (NULL),
   open_status (0),
   fs_flags (0),
-  read_state (NULL)
+  read_state (NULL),
+  archetype (NULL),
+  usecount (0)
 {
 }
 
 /* Normal I/O destructor */
 fhandler_base::~fhandler_base (void)
 {
-  if (unix_path_name != NULL)
-    cfree ((void *) unix_path_name);
-  if (win32_path_name != NULL)
-    cfree ((void *) win32_path_name);
   if (rabuf)
     free (rabuf);
-  unix_path_name = win32_path_name = NULL;
 }
 
 /**********************************************************************/
 /* /dev/null */
 
 fhandler_dev_null::fhandler_dev_null () :
-	fhandler_base (FH_NULL)
+	fhandler_base ()
 {
 }
 
@@ -1232,19 +1297,19 @@ fhandler_dev_null::dump (void)
 }
 
 void
-fhandler_base::set_inheritance (HANDLE &h, int not_inheriting)
+fhandler_base::set_no_inheritance (HANDLE &h, int not_inheriting)
 {
-#ifdef DEBUGGING_AND_FDS_PROTECTED
   HANDLE oh = h;
-#endif
   /* Note that we could use SetHandleInformation here but it is not available
      on all platforms.  Test cases seem to indicate that using DuplicateHandle
      in this fashion does not actually close the original handle, which is
      what we want.  If this changes in the future, we may be forced to use
      SetHandleInformation on newer OS's */
-  if (!DuplicateHandle (hMainProc, h, hMainProc, &h, 0, !not_inheriting,
+  if (!DuplicateHandle (hMainProc, oh, hMainProc, &h, 0, !not_inheriting,
 			     DUPLICATE_SAME_ACCESS | DUPLICATE_CLOSE_SOURCE))
     debug_printf ("DuplicateHandle failed, %E");
+  if (oh != h)
+    VerifyHandle (h);
 #ifdef DEBUGGING_AND_FDS_PROTECTED
   if (h)
     setclexec (oh, h, not_inheriting);
@@ -1254,24 +1319,21 @@ fhandler_base::set_inheritance (HANDLE &h, int not_inheriting)
 void
 fhandler_base::fork_fixup (HANDLE parent, HANDLE &h, const char *name)
 {
+  HANDLE oh = h;
   if (/* !is_socket () && */ !get_close_on_exec ())
     debug_printf ("handle %p already opened", h);
   else if (!DuplicateHandle (parent, h, hMainProc, &h, 0, !get_close_on_exec (),
 			     DUPLICATE_SAME_ACCESS))
     system_printf ("%s - %E, handle %s<%p>", get_name (), name, h);
-#ifdef DEBUGGING_AND_FDS_PROTECTED
-  else if (get_close_on_exec ())
-    ProtectHandle (h);	/* would have to be fancier than this */
-  else
-    /* ProtectHandleINH (h) */;	/* Should already be protected */
-#endif
+  else if (oh != h)
+    VerifyHandle (h);
 }
 
 void
 fhandler_base::set_close_on_exec (int val)
 {
   if (!get_nohandle ())
-    set_inheritance (io_handle, val);
+    set_no_inheritance (io_handle, val);
   set_close_on_exec_flag (val);
   debug_printf ("set close_on_exec for %s to %d", get_name (), val);
 }
@@ -1299,7 +1361,7 @@ fhandler_base::set_nonblocking (int yes)
 }
 
 DIR *
-fhandler_base::opendir (path_conv&)
+fhandler_base::opendir ()
 {
   set_errno (ENOTDIR);
   return NULL;
