@@ -1,6 +1,6 @@
 /* fhandler_console.cc
 
-   Copyright 1996, 1997, 1998, 1999, 2000, 2001, 2002, 2003 Red Hat, Inc.
+   Copyright 1996, 1997, 1998, 1999, 2000, 2001, 2002, 2003, 2004 Red Hat, Inc.
 
 This file is part of Cygwin.
 
@@ -21,8 +21,8 @@ details. */
 #include <sys/cygwin.h>
 #include "cygerrno.h"
 #include "security.h"
-#include "fhandler.h"
 #include "path.h"
+#include "fhandler.h"
 #include "dtable.h"
 #include "cygheap.h"
 #include "sigproc.h"
@@ -32,7 +32,7 @@ details. */
 
 #define CONVERT_LIMIT 16384
 
-static BOOL
+static bool
 cp_convert (UINT destcp, char *dest, UINT srccp, const char *src, DWORD size)
 {
   if (!size)
@@ -46,27 +46,12 @@ cp_convert (UINT destcp, char *dest, UINT srccp, const char *src, DWORD size)
     {
       WCHAR wbuffer[CONVERT_LIMIT]; /* same size as the maximum input, s.b. */
       if (!MultiByteToWideChar (srccp, 0, src, size, wbuffer, sizeof (wbuffer)))
-	return FALSE;
+	return false;
       if (!WideCharToMultiByte (destcp, 0, wbuffer, size, dest, size,
 				NULL, NULL))
-	return FALSE;
+	return false;
     }
-  return TRUE;
-}
-
-/* The results of GetConsoleCP() and GetConsoleOutputCP() cannot be
-   cached, because a program or the user can change these values at
-   any time. */
-inline BOOL
-con_to_str (char *d, const char *s, DWORD sz)
-{
-  return cp_convert (get_cp (), d, GetConsoleCP (), s, sz);
-}
-
-inline BOOL
-str_to_con (char *d, const char *s, DWORD sz)
-{
-  return cp_convert (GetConsoleOutputCP (), d, get_cp (), s, sz);
+  return true;
 }
 
 /*
@@ -88,8 +73,6 @@ static console_state NO_COPY *shared_console_info;
 
 dev_console NO_COPY *fhandler_console::dev_state;
 
-int NO_COPY fhandler_console::open_fhs;
-
 /* Allocate and initialize the shared record for the current console.
    Returns a pointer to shared_console_info. */
 tty_min *
@@ -109,7 +92,7 @@ fhandler_console::get_tty_stuff (int flags = 0)
     {
       shared_console_info->tty_min_state.setntty (TTY_CONSOLE);
       shared_console_info->tty_min_state.setsid (myself->sid);
-      shared_console_info->tty_min_state.set_ctty (TTY_CONSOLE, flags);
+      myself->set_ctty (&shared_console_info->tty_min_state, flags, NULL);
 
       dev_state->scroll_region.Bottom = -1;
       dev_state->dwLastCursorPosition.X = -1;
@@ -181,10 +164,31 @@ set_console_state_for_spawn ()
   return 1;
 }
 
-BOOL
-fhandler_console::set_raw_win32_keyboard_mode (BOOL new_mode)
+/* The results of GetConsoleCP() and GetConsoleOutputCP() cannot be
+   cached, because a program or the user can change these values at
+   any time. */
+inline bool
+dev_console::con_to_str (char *d, const char *s, DWORD sz)
 {
-  BOOL old_mode = dev_state->raw_win32_keyboard_mode;
+  return cp_convert (get_cp (), d, GetConsoleCP (), s, sz);
+}
+
+inline bool
+dev_console::str_to_con (char *d, const char *s, DWORD sz)
+{
+  if (alternate_charset_active)
+    {
+      /* no translation when alternate charset is active */
+      memcpy(d, s, sz);
+      return true;
+    }
+  return cp_convert (GetConsoleOutputCP (), d, get_cp (), s, sz);
+}
+
+bool
+fhandler_console::set_raw_win32_keyboard_mode (bool new_mode)
+{
+  bool old_mode = dev_state->raw_win32_keyboard_mode;
   dev_state->raw_win32_keyboard_mode = new_mode;
   syscall_printf ("raw keyboard mode %sabled", dev_state->raw_win32_keyboard_mode ? "en" : "dis");
   return old_mode;
@@ -294,6 +298,8 @@ fhandler_console::read (void *pv, size_t& buflen)
 #define virtual_key_code (input_rec.Event.KeyEvent.wVirtualKeyCode)
 #define control_key_state (input_rec.Event.KeyEvent.dwControlKeyState)
 
+	  dev_state->nModifiers = 0;
+
 #ifdef DEBUGGING
 	  /* allow manual switching to/from raw mode via ctrl-alt-scrolllock */
 	  if (input_rec.Event.KeyEvent.bKeyDown &&
@@ -340,13 +346,25 @@ fhandler_console::read (void *pv, size_t& buflen)
 		&& input_rec.Event.KeyEvent.wVirtualScanCode == 0x38))
 	    continue;
 
+	  if (control_key_state & SHIFT_PRESSED)
+	    dev_state->nModifiers |= 1;
+	  if (control_key_state & RIGHT_ALT_PRESSED)
+	    dev_state->nModifiers |= 2;
+	  if (control_key_state & CTRL_PRESSED)
+	    dev_state->nModifiers |= 4;
+	  if (control_key_state & LEFT_ALT_PRESSED)
+	    dev_state->nModifiers |= 8;
+
 	  if (wch == 0 ||
 	      /* arrow/function keys */
 	      (input_rec.Event.KeyEvent.dwControlKeyState & ENHANCED_KEY))
 	    {
 	      toadd = get_nonascii_key (input_rec, tmp);
 	      if (!toadd)
-		continue;
+		{
+		  dev_state->nModifiers = 0;
+		  continue;
+		}
 	      nread = strlen (toadd);
 	    }
 	  else
@@ -355,7 +373,7 @@ fhandler_console::read (void *pv, size_t& buflen)
 	      /* Need this check since US code page seems to have a bug when
 		 converting a CTRL-U. */
 	      if ((unsigned char) ich > 0x7f)
-		con_to_str (tmp + 1, tmp + 1, 1);
+		dev_state->con_to_str (tmp + 1, tmp + 1, 1);
 	      /* Determine if the keystroke is modified by META.  The tricky
 		 part is to distinguish whether the right Alt key should be
 		 recognized as Alt, or as AltGr. */
@@ -379,6 +397,7 @@ fhandler_console::read (void *pv, size_t& buflen)
 		  tmp[1] = cyg_tolower (tmp[1]);
 		  toadd = tmp;
 		  nread++;
+		  dev_state->nModifiers &= ~4;
 		}
 	    }
 #undef ich
@@ -523,10 +542,10 @@ fhandler_console::set_input_state ()
     input_tcsetattr (0, &tc->ti);
 }
 
-BOOL
+bool
 fhandler_console::fillin_info (void)
 {
-  BOOL ret;
+  bool ret;
   CONSOLE_SCREEN_BUFFER_INFO linfo;
 
   if ((ret = GetConsoleScreenBufferInfo (get_output_handle (), &linfo)))
@@ -594,7 +613,7 @@ fhandler_console::scroll_screen (int x1, int y1, int x2, int y2, int xn, int yn)
 }
 
 int
-fhandler_console::open (path_conv *, int flags, mode_t)
+fhandler_console::open (int flags, mode_t)
 {
   HANDLE h;
 
@@ -643,10 +662,10 @@ fhandler_console::open (path_conv *, int flags, mode_t)
 
   TTYCLEARF (RSTCONS);
   set_open_status ();
-  open_fhs++;
-  debug_printf ("incremented open_fhs, now %d", open_fhs);
-  debug_printf ("opened conin$ %p, conout$ %p",
-		get_io_handle (), get_output_handle ());
+  cygheap->open_fhs++;
+  debug_printf ("incremented open_fhs, now %d", cygheap->open_fhs);
+  debug_printf ("opened conin$ %p, conout$ %p", get_io_handle (),
+		get_output_handle ());
 
   return 1;
 }
@@ -658,14 +677,12 @@ fhandler_console::close (void)
   CloseHandle (get_output_handle ());
   set_io_handle (NULL);
   set_output_handle (NULL);
-  if (!cygheap->fdtab.in_vfork_cleanup () && --open_fhs <= 0
-      && myself->ctty != TTY_CONSOLE)
+  if (--(cygheap->open_fhs) <= 0 && myself->ctty != TTY_CONSOLE)
     {
-      syscall_printf ("open_fhs %d, freeing console %p",
-		      fhandler_console::open_fhs, myself->ctty);
+      syscall_printf ("open_fhs %d", cygheap->open_fhs);
       FreeConsole ();
     }
-  debug_printf ("decremented open_fhs, now %d", open_fhs);
+  debug_printf ("decremented open_fhs, now %d", cygheap->open_fhs);
   return 0;
 }
 
@@ -679,7 +696,7 @@ fhandler_console::dup (fhandler_base *child)
 {
   fhandler_console *fhc = (fhandler_console *) child;
 
-  if (!fhc->open (NULL, get_flags () & ~O_NOCTTY, 0))
+  if (!fhc->open (get_flags () & ~O_NOCTTY, 0))
     system_printf ("error opening console, %E");
 
   return 0;
@@ -715,6 +732,17 @@ fhandler_console::ioctl (unsigned int cmd, void *buf)
       case TIOCSWINSZ:
 	(void) bg_check (SIGTTOU);
 	return 0;
+      case TIOCLINUX:
+	if (* (int *) buf == 6)
+	  {
+	    * (int *) buf = dev_state->nModifiers;
+	    return 0;
+	  }
+	else
+	  {
+	    set_errno (EINVAL);
+	    return -1;
+	  }
     }
 
   return fhandler_base::ioctl (cmd, buf);
@@ -867,7 +895,7 @@ fhandler_console::tcgetattr (struct termios *t)
 }
 
 fhandler_console::fhandler_console () :
-  fhandler_termios (FH_CONSOLE, -1)
+  fhandler_termios ()
 {
 }
 
@@ -878,7 +906,7 @@ fhandler_console::fhandler_console () :
 void
 fhandler_console::set_default_attr ()
 {
-  dev_state->blink = dev_state->underline = dev_state->reverse = FALSE;
+  dev_state->blink = dev_state->underline = dev_state->reverse = false;
   dev_state->intensity = INTENSITY_NORMAL;
   dev_state->fg = dev_state->default_color & FOREGROUND_ATTR_MASK;
   dev_state->bg = dev_state->default_color & BACKGROUND_ATTR_MASK;
@@ -961,7 +989,7 @@ fhandler_console::clear_screen (int x1, int y1, int x2, int y2)
 }
 
 void
-fhandler_console::cursor_set (BOOL rel_to_top, int x, int y)
+fhandler_console::cursor_set (bool rel_to_top, int x, int y)
 {
   COORD pos;
 
@@ -989,7 +1017,7 @@ fhandler_console::cursor_rel (int x, int y)
   fillin_info ();
   x += dev_state->info.dwCursorPosition.X;
   y += dev_state->info.dwCursorPosition.Y;
-  cursor_set (FALSE, x, y);
+  cursor_set (false, x, y);
 }
 
 void
@@ -1074,10 +1102,10 @@ fhandler_console::char_command (char c)
 	       dev_state->underline = 1;
 	       break;
 	     case 5:    /* blink mode */
-	       dev_state->blink = TRUE;
+	       dev_state->blink = true;
 	       break;
 	     case 7:    /* reverse */
-	       dev_state->reverse = TRUE;
+	       dev_state->reverse = true;
 	       break;
 	     case 8:    /* invisible */
 	       dev_state->intensity = INTENSITY_INVISIBLE;
@@ -1085,11 +1113,17 @@ fhandler_console::char_command (char c)
 	     case 9:    /* dim */
 	       dev_state->intensity = INTENSITY_DIM;
 	       break;
+             case 10:   /* end alternate charset */
+               dev_state->alternate_charset_active = false;
+	       break;
+             case 11:   /* start alternate charset */
+               dev_state->alternate_charset_active = true;
+	       break;
 	     case 24:
-	       dev_state->underline = FALSE;
+	       dev_state->underline = false;
 	       break;
 	     case 27:
-	       dev_state->reverse = FALSE;
+	       dev_state->reverse = false;
 	       break;
 	     case 30:		/* BLACK foreground */
 	       dev_state->fg = 0;
@@ -1156,7 +1190,7 @@ fhandler_console::char_command (char c)
 	  switch (dev_state->args_[0])
 	    {
 	    case 4:    /* Insert mode */
-	      dev_state->insert_mode = (c == 'h') ? TRUE : FALSE;
+	      dev_state->insert_mode = (c == 'h') ? true : false;
 	      syscall_printf ("insert mode %sabled", dev_state->insert_mode ? "en" : "dis");
 	      break;
 	    }
@@ -1205,12 +1239,12 @@ fhandler_console::char_command (char c)
 	  break;
 
 	case 1000: /* Mouse support */
-	  dev_state->use_mouse = (c == 'h') ? TRUE : FALSE;
+	  dev_state->use_mouse = (c == 'h') ? true : false;
 	  syscall_printf ("mouse support %sabled", dev_state->use_mouse ? "en" : "dis");
 	  break;
 
 	case 2000: /* Raw keyboard mode */
-	  set_raw_win32_keyboard_mode ((c == 'h') ? TRUE : FALSE);
+	  set_raw_win32_keyboard_mode ((c == 'h') ? true : false);
 	  break;
 
 	default: /* Ignore */
@@ -1231,7 +1265,7 @@ fhandler_console::char_command (char c)
 	  break;
 	case 2:			/* Clear screen */
 	  clear_screen (0, 0, -1, -1);
-	  cursor_set (TRUE, 0,0);
+	  cursor_set (true, 0,0);
 	  break;
 	default:
 	  goto bad_escape;
@@ -1271,27 +1305,27 @@ fhandler_console::char_command (char c)
       break;
     case 'H':
     case 'f':
-      cursor_set (TRUE, (dev_state->args_[1] ? dev_state->args_[1] : 1) - 1,
+      cursor_set (true, (dev_state->args_[1] ? dev_state->args_[1] : 1) - 1,
 			(dev_state->args_[0] ? dev_state->args_[0] : 1) - 1);
       break;
     case 'G':   /* hpa - position cursor at column n - 1 */
       cursor_get (&x, &y);
-      cursor_set (FALSE, (dev_state->args_[0] ? dev_state->args_[0] - 1 : 0), y);
+      cursor_set (false, (dev_state->args_[0] ? dev_state->args_[0] - 1 : 0), y);
       break;
     case 'd':   /* vpa - position cursor at line n */
       cursor_get (&x, &y);
-      cursor_set (TRUE, x, (dev_state->args_[0] ? dev_state->args_[0] - 1 : 0));
+      cursor_set (true, x, (dev_state->args_[0] ? dev_state->args_[0] - 1 : 0));
       break;
     case 's':   /* Save cursor position */
       cursor_get (&dev_state->savex, &dev_state->savey);
       dev_state->savey -= dev_state->info.winTop;
       break;
     case 'u':   /* Restore cursor position */
-      cursor_set (TRUE, dev_state->savex, dev_state->savey);
+      cursor_set (true, dev_state->savex, dev_state->savey);
       break;
     case 'I':	/* TAB */
       cursor_get (&x, &y);
-      cursor_set (FALSE, 8 * (x / 8 + 1), y);
+      cursor_set (false, 8 * (x / 8 + 1), y);
       break;
     case 'L':				/* AL - insert blank lines */
       dev_state->args_[0] = dev_state->args_[0] ? dev_state->args_[0] : 1;
@@ -1330,7 +1364,7 @@ fhandler_console::char_command (char c)
       break;
     case 'Z':				/* Back tab */
       cursor_get (&x, &y);
-      cursor_set (FALSE, ((8 * (x / 8 + 1)) - 8), y);
+      cursor_set (false, ((8 * (x / 8 + 1)) - 8), y);
       break;
     case 'b':				/* Repeat char #1 #2 times */
       if (dev_state->insert_mode)
@@ -1362,7 +1396,7 @@ fhandler_console::char_command (char c)
     case 'r':				/* Set Scroll region */
       dev_state->scroll_region.Top = dev_state->args_[0] ? dev_state->args_[0] - 1 : 0;
       dev_state->scroll_region.Bottom = dev_state->args_[1] ? dev_state->args_[1] - 1 : -1;
-      cursor_set (TRUE, 0, 0);
+      cursor_set (true, 0, 0);
       break;
     case 'g':				/* TAB set/clear */
       break;
@@ -1396,7 +1430,7 @@ fhandler_console::write_normal (const unsigned char *src,
 	  DWORD buf_len;
 	  char buf[CONVERT_LIMIT];
 	  done = buf_len = min (sizeof (buf), len);
-	  if (!str_to_con (buf, (const char *) src, buf_len))
+	  if (!dev_state->str_to_con (buf, (const char *) src, buf_len))
 	    {
 	      debug_printf ("conversion error, handle %p",
 			    get_output_handle ());
@@ -1446,7 +1480,7 @@ fhandler_console::write_normal (const unsigned char *src,
 		  y--;
 		}
 	    }
-	  cursor_set (FALSE, ((tc->ti.c_oflag & ONLCR) ? 0 : x), y + 1);
+	  cursor_set (false, ((tc->ti.c_oflag & ONLCR) ? 0 : x), y + 1);
 	  break;
 	case BAK:
 	  cursor_rel (-1, 0);
@@ -1456,14 +1490,14 @@ fhandler_console::write_normal (const unsigned char *src,
 	  break;
 	case CR:
 	  cursor_get (&x, &y);
-	  cursor_set (FALSE, 0, y);
+	  cursor_set (false, 0, y);
 	  break;
 	case ERR:
 	  WriteFile (get_output_handle (), src, 1, &done, 0);
 	  break;
 	case TAB:
 	  cursor_get (&x, &y);
-	  cursor_set (FALSE, 8 * (x / 8 + 1), y);
+	  cursor_set (false, 8 * (x / 8 + 1), y);
 	  break;
 	}
       src ++;
@@ -1495,7 +1529,7 @@ fhandler_console::write (const void *vsrc, size_t len)
 	  if (*src == '[')
 	    {
 	      dev_state->state_ = gotsquare;
-	      dev_state->saw_question_mark = FALSE;
+	      dev_state->saw_question_mark = false;
 	      for (dev_state->nargs_ = 0; dev_state->nargs_ < MAXARGS; dev_state->nargs_++)
 		dev_state->args_[dev_state->nargs_] = 0;
 	      dev_state->nargs_ = 0;
@@ -1516,12 +1550,12 @@ fhandler_console::write (const void *vsrc, size_t len)
 	    {
 	      set_default_attr ();
 	      clear_screen (0, 0, -1, -1);
-	      cursor_set (TRUE, 0, 0);
+	      cursor_set (true, 0, 0);
 	      dev_state->state_ = normal;
 	    }
 	  else if (*src == '8')		/* Restore cursor position */
 	    {
-	      cursor_set (TRUE, dev_state->savex, dev_state->savey);
+	      cursor_set (true, dev_state->savex, dev_state->savey);
 	      dev_state->state_ = normal;
 	    }
 	  else if (*src == '7')		/* Save cursor position */
@@ -1603,7 +1637,7 @@ fhandler_console::write (const void *vsrc, size_t len)
 	  else if (*src != '@' && !isalpha (*src) && !isdigit (*src))
 	    {
 	      if (*src == '?')
-		dev_state->saw_question_mark = TRUE;
+		dev_state->saw_question_mark = true;
 	      /* ignore any extra chars between [ and first arg or command */
 	      src++;
 	    }
@@ -1696,7 +1730,7 @@ fhandler_console::init (HANDLE f, DWORD a, mode_t bin)
     flags = O_WRONLY;
   if (a == (GENERIC_READ | GENERIC_WRITE))
     flags = O_RDWR;
-  open ((path_conv *) NULL, flags | O_BINARY);
+  open (flags | O_BINARY);
   if (f != INVALID_HANDLE_VALUE)
     CloseHandle (f);	/* Reopened by open */
 
@@ -1713,7 +1747,7 @@ void
 fhandler_console::set_close_on_exec (int val)
 {
   fhandler_base::set_close_on_exec (val);
-  set_inheritance (output_handle, val);
+  set_no_inheritance (output_handle, val);
 }
 
 void
@@ -1725,7 +1759,8 @@ fhandler_console::fixup_after_fork (HANDLE)
   /* Windows does not allow duplication of console handles between processes
      so open the console explicitly. */
 
-  if (!open (NULL, O_NOCTTY | get_flags (), 0))
+  cygheap->open_fhs--;		/* The downside of storing this in cygheap. */
+  if (!open (O_NOCTTY | get_flags (), 0))
     system_printf ("error opening console after fork, %E");
 
   if (!get_close_on_exec ())
@@ -1750,12 +1785,13 @@ set_console_title (char *title)
 }
 
 void
-fhandler_console::fixup_after_exec (HANDLE)
+fhandler_console::fixup_after_exec ()
 {
   HANDLE h = get_handle ();
   HANDLE oh = get_output_handle ();
 
-  if (!open (NULL, O_NOCTTY | get_flags (), 0))
+  cygheap->open_fhs--;		/* The downside of storing this in cygheap. */
+  if (!open (O_NOCTTY | get_flags (), 0))
     {
       int sawerr = 0;
       if (!get_io_handle ())
