@@ -1,7 +1,7 @@
 /* fhandler_floppy.cc.  See fhandler.h for a description of the
    fhandler classes.
 
-   Copyright 1999, 2000, 2001, 2002 Red Hat, Inc.
+   Copyright 1999, 2000, 2001, 2002, 2003, 2004 Red Hat, Inc.
 
 This file is part of Cygwin.
 
@@ -17,6 +17,7 @@ details. */
 #include <cygwin/hdreg.h>
 #include <cygwin/fs.h>
 #include "security.h"
+#include "path.h"
 #include "fhandler.h"
 #include "cygerrno.h"
 
@@ -41,12 +42,13 @@ fhandler_dev_floppy::is_eof (int)
   return ret;
 }
 
-fhandler_dev_floppy::fhandler_dev_floppy (int unit) : fhandler_dev_raw (FH_FLOPPY, unit)
+fhandler_dev_floppy::fhandler_dev_floppy ()
+  : fhandler_dev_raw ()
 {
 }
 
 int
-fhandler_dev_floppy::open (path_conv *real_path, int flags, mode_t)
+fhandler_dev_floppy::open (int flags, mode_t)
 {
   /* The correct size of the buffer would be 512 bytes,
    * which is the atomic size, supported by WinNT.
@@ -61,7 +63,7 @@ fhandler_dev_floppy::open (path_conv *real_path, int flags, mode_t)
    * and cpio buffer sizes by default!
   */
   devbufsiz = 61440L; /* 512L; */
-  return fhandler_dev_raw::open (real_path, flags);
+  return fhandler_dev_raw::open (flags);
 }
 
 int
@@ -170,7 +172,7 @@ fhandler_dev_floppy::lseek (_off64_t offset, int whence)
 	return ret;
       devbufstart = devbufend = 0;
 
-      low = sector_aligned_offset & 0xffffffff;
+      low = sector_aligned_offset & UINT32_MAX;
       high = sector_aligned_offset >> 32;
       if (SetFilePointer (get_handle (), low, &high, FILE_BEGIN)
 	  == INVALID_SET_FILE_POINTER && GetLastError ())
