@@ -1,6 +1,6 @@
 /* fhandler_tty.cc
 
-   Copyright 1997, 1998, 2000, 2001, 2002, 2003 Red Hat, Inc.
+   Copyright 1997, 1998, 2000, 2001, 2002, 2003, 2004 Red Hat, Inc.
 
 This file is part of Cygwin.
 
@@ -18,14 +18,14 @@ details. */
 #include <limits.h>
 #include "cygerrno.h"
 #include "security.h"
-#include "fhandler.h"
 #include "path.h"
+#include "fhandler.h"
 #include "dtable.h"
 #include "sigproc.h"
 #include "pinfo.h"
 #include "cygheap.h"
 #include "shared_info.h"
-#include "cygwin/cygserver.h"
+#include "cygserver.h"
 #include "cygthread.h"
 
 /* Tty master stuff */
@@ -36,11 +36,17 @@ static DWORD WINAPI process_input (void *);		// Input queue thread
 static DWORD WINAPI process_output (void *);		// Output queue thread
 static DWORD WINAPI process_ioctl (void *);		// Ioctl requests thread
 
-fhandler_tty_master::fhandler_tty_master (int unit)
-  : fhandler_pty_master (FH_TTYM, unit), console (NULL)
+fhandler_tty_master::fhandler_tty_master ()
+  : fhandler_pty_master (), console (NULL)
 {
 }
 
+int
+fhandler_tty_slave::get_unit ()
+{
+  return dev () == FH_TTY ? myself->ctty : dev ().minor;
+}
+
 void
 fhandler_tty_master::set_winsize (bool sendSIGWINCH)
 {
@@ -52,9 +58,10 @@ fhandler_tty_master::set_winsize (bool sendSIGWINCH)
 }
 
 int
-fhandler_tty_master::init (int ntty)
+fhandler_tty_master::init ()
 {
-  termios_printf ("Creating master for tty%d", ntty);
+  slave = dev ();
+  termios_printf ("Creating master for tty%d", get_unit ());
 
   if (init_console ())
     {
@@ -66,13 +73,12 @@ fhandler_tty_master::init (int ntty)
   memset (&ti, 0, sizeof (ti));
   console->tcsetattr (0, &ti);
 
-  ttynum = ntty;
-
-  cygwin_shared->tty[ttynum]->common_init (this);
+  cygwin_shared->tty[get_unit ()]->common_init (this);
 
   set_winsize (false);
 
   inuse = get_ttyp ()->create_inuse (TTY_MASTER_ALIVE);
+  set_close_on_exec (true);
 
   cygthread *h;
   h = new cygthread (process_input, cygself, "ttyin");
@@ -431,14 +437,8 @@ process_ioctl (void *)
 /**********************************************************************/
 /* Tty slave stuff */
 
-fhandler_tty_slave::fhandler_tty_slave (int num)
-  : fhandler_tty_common (FH_TTYS, num)
-{
-  set_r_no_interrupt (1);
-}
-
 fhandler_tty_slave::fhandler_tty_slave ()
-  : fhandler_tty_common (FH_TTYS, 0)
+  : fhandler_tty_common ()
 {
   set_r_no_interrupt (1);
 }
@@ -446,12 +446,23 @@ fhandler_tty_slave::fhandler_tty_slave ()
 /* FIXME: This function needs to close handles when it has
    a failing condition. */
 int
-fhandler_tty_slave::open (path_conv *, int flags, mode_t)
+fhandler_tty_slave::open (int flags, mode_t)
 {
-  tcinit (cygwin_shared->tty[ttynum]);
+  if (get_device () == FH_TTY)
+    pc.dev.tty_to_real_device ();
+  fhandler_tty_slave *arch = (fhandler_tty_slave *)
+    cygheap->fdtab.find_archetype (pc.dev);
+  if (arch)
+    {
+      *this = *(fhandler_tty_slave *) arch;
+      termios_printf ("copied tty fhandler archetype");
+      cygheap->open_fhs++;
+      goto out;
+    }
 
-  attach_tty (ttynum);
-  tc->set_ctty (ttynum, flags);
+  tcinit (cygwin_shared->tty[get_unit ()]);
+
+  attach_tty (get_unit ());
 
   set_flags ((flags & ~O_TEXT) | O_BINARY);
   /* Create synchronisation events */
@@ -462,7 +473,7 @@ fhandler_tty_slave::open (path_conv *, int flags, mode_t)
      startup if use_tty is non-zero.  It will not exist if this is a
      pty opened by fhandler_pty_master::open.  In the former case, tty
      output is handled by a separate thread which controls output.  */
-  __small_sprintf (buf, OUTPUT_DONE_EVENT, ttynum);
+  __small_sprintf (buf, OUTPUT_DONE_EVENT, get_unit ());
   output_done_event = OpenEvent (EVENT_ALL_ACCESS, TRUE, buf);
 
   if (!(output_mutex = get_ttyp ()->open_output_mutex ()))
@@ -477,7 +488,7 @@ fhandler_tty_slave::open (path_conv *, int flags, mode_t)
       __seterrno ();
       return 0;
     }
-  __small_sprintf (buf, INPUT_AVAILABLE_EVENT, ttynum);
+  __small_sprintf (buf, INPUT_AVAILABLE_EVENT, get_unit ());
   if (!(input_available_event = OpenEvent (EVENT_ALL_ACCESS, TRUE, buf)))
     {
       termios_printf ("open input event failed, %E");
@@ -487,16 +498,16 @@ fhandler_tty_slave::open (path_conv *, int flags, mode_t)
 
   /* The ioctl events may or may not exist.  See output_done_event,
      above.  */
-  __small_sprintf (buf, IOCTL_REQUEST_EVENT, ttynum);
+  __small_sprintf (buf, IOCTL_REQUEST_EVENT, get_unit ());
   ioctl_request_event = OpenEvent (EVENT_ALL_ACCESS, TRUE, buf);
-  __small_sprintf (buf, IOCTL_DONE_EVENT, ttynum);
+  __small_sprintf (buf, IOCTL_DONE_EVENT, get_unit ());
   ioctl_done_event = OpenEvent (EVENT_ALL_ACCESS, TRUE, buf);
 
   /* FIXME: Needs a method to eliminate tty races */
   {
     acquire_output_mutex (500);
     inuse = get_ttyp ()->create_inuse (TTY_SLAVE_ALIVE);
-    get_ttyp ()->was_opened = TRUE;
+    get_ttyp ()->was_opened = true;
     release_output_mutex ();
   }
 
@@ -509,7 +520,9 @@ fhandler_tty_slave::open (path_conv *, int flags, mode_t)
       return 0;
     }
 
-  HANDLE from_master_local, to_master_local;
+  HANDLE from_master_local;
+  HANDLE to_master_local;
+  from_master_local = to_master_local = NULL;
 
 #ifdef USE_SERVER
   if (!wincap.has_security ()
@@ -517,15 +530,16 @@ fhandler_tty_slave::open (path_conv *, int flags, mode_t)
       || !cygserver_attach_tty (&from_master_local, &to_master_local))
 #endif
     {
+#ifdef USE_SERVER
       termios_printf ("cannot dup handles via server. using old method.");
-
+#endif
       HANDLE tty_owner = OpenProcess (PROCESS_DUP_HANDLE, FALSE,
 				      get_ttyp ()->master_pid);
       termios_printf ("tty own handle %p",tty_owner);
       if (tty_owner == NULL)
 	{
 	  termios_printf ("can't open tty (%d) handle process %d",
-			  ttynum, get_ttyp ()->master_pid);
+			  get_unit (), get_ttyp ()->master_pid);
 	  __seterrno ();
 	  return 0;
 	}
@@ -539,6 +553,7 @@ fhandler_tty_slave::open (path_conv *, int flags, mode_t)
 	  return 0;
 	}
 
+      VerifyHandle (from_master_local);
       if (!DuplicateHandle (tty_owner, get_ttyp ()->to_master,
 			  hMainProc, &to_master_local, 0, TRUE,
 			  DUPLICATE_SAME_ACCESS))
@@ -547,6 +562,7 @@ fhandler_tty_slave::open (path_conv *, int flags, mode_t)
 	  __seterrno ();
 	  return 0;
 	}
+      VerifyHandle (to_master_local);
       CloseHandle (tty_owner);
     }
 
@@ -559,8 +575,8 @@ fhandler_tty_slave::open (path_conv *, int flags, mode_t)
   set_output_handle (to_master_local);
 
   set_open_status ();
-  if (fhandler_console::open_fhs++ == 0 && !GetConsoleCP ()
-      && !output_done_event && wincap.pty_needs_alloc_console ())
+  if (cygheap->open_fhs++ == 0 && !GetConsoleCP () && !output_done_event
+      && wincap.pty_needs_alloc_console () && !GetProcessWindowStation ())
     {
       BOOL b;
       HWINSTA h = CreateWindowStation (NULL, 0, GENERIC_READ | GENERIC_WRITE, &sec_none_nih);
@@ -576,8 +592,18 @@ fhandler_tty_slave::open (path_conv *, int flags, mode_t)
       if (b)
 	init_console_handler ();
     }
-  termios_printf ("incremented open_fhs %d", fhandler_console::open_fhs);
-  termios_printf ("tty%d opened", ttynum);
+
+  // FIXME: Do this better someday
+  arch = (fhandler_tty_slave *) cmalloc (HEAP_ARCHETYPES, sizeof (*this));
+  *((fhandler_tty_slave **) cygheap->fdtab.add_archetype ()) = arch;
+  archetype = arch;
+  *arch = *this;
+
+out:
+  usecount = 0;
+  archetype->usecount++;
+  report_tty_counts (this, "opened", "incremented ", "");
+  myself->set_ctty (get_ttyp (), flags, arch);
 
   return 1;
 }
@@ -585,9 +611,23 @@ fhandler_tty_slave::open (path_conv *, int flags, mode_t)
 int
 fhandler_tty_slave::close ()
 {
-  if (!--fhandler_console::open_fhs && myself->ctty == -1)
+  if (!--cygheap->open_fhs && myself->ctty == -1)
     FreeConsole ();
-  termios_printf ("decremented open_fhs %d", fhandler_console::open_fhs);
+
+  archetype->usecount--;
+  report_tty_counts (this, "closed", "decremented ", "");
+
+  if (archetype->usecount)
+    {
+#ifdef DEBUGGING
+      if (archetype->usecount < 0)
+	system_printf ("error: usecount %d", archetype->usecount);
+#endif
+      termios_printf ("just returning because archetype usecount is != 0");
+      return 0;
+    }
+
+  termios_printf ("closing last open %s handle", ttyname ());
   return fhandler_tty_common::close ();
 }
 
@@ -618,17 +658,17 @@ fhandler_tty_slave::cygserver_attach_tty (LPHANDLE from_master_ptr,
 void
 fhandler_tty_slave::init (HANDLE, DWORD a, mode_t)
 {
-  int mode = 0;
+  int flags = 0;
 
   a &= GENERIC_READ | GENERIC_WRITE;
   if (a == GENERIC_READ)
-    mode = O_RDONLY;
+    flags = O_RDONLY;
   if (a == GENERIC_WRITE)
-    mode = O_WRONLY;
+    flags = O_WRONLY;
   if (a == (GENERIC_READ | GENERIC_WRITE))
-    mode = O_RDWR;
+    flags = O_RDWR;
 
-  open (0, mode);
+  open (flags);
 }
 
 int
@@ -636,7 +676,7 @@ fhandler_tty_slave::write (const void *ptr, size_t len)
 {
   DWORD n, towrite = len;
 
-  termios_printf ("tty%d, write(%x, %d)", ttynum, ptr, len);
+  termios_printf ("tty%d, write(%x, %d)", get_unit (), ptr, len);
 
   acquire_output_mutex (INFINITE);
 
@@ -871,9 +911,14 @@ fhandler_tty_slave::read (void *ptr, size_t& len)
 int
 fhandler_tty_slave::dup (fhandler_base *child)
 {
-  fhandler_console::open_fhs++;
-  termios_printf ("incremented open_fhs %d", fhandler_console::open_fhs);
-  return fhandler_tty_common::dup (child);
+  fhandler_tty_slave *arch = (fhandler_tty_slave *) archetype;
+  *(fhandler_tty_slave *) child = *arch;
+  child->usecount = 0;
+  arch->usecount++;
+  cygheap->open_fhs++;
+  report_tty_counts (child, "duped", "incremented ", "");
+  myself->set_ctty (get_ttyp (), openflags, arch);
+  return 0;
 }
 
 int
@@ -882,11 +927,9 @@ fhandler_tty_common::dup (fhandler_base *child)
   fhandler_tty_slave *fts = (fhandler_tty_slave *) child;
   int errind;
 
-  fts->ttynum = ttynum;
   fts->tcinit (get_ttyp ());
 
-  attach_tty (ttynum);
-  tc->set_ctty (ttynum, openflags);
+  attach_tty (get_unit ());
 
   HANDLE nh;
 
@@ -965,6 +1008,7 @@ fhandler_tty_common::dup (fhandler_base *child)
       errind = 9;
       goto err;
     }
+
   return 0;
 
 err:
@@ -1017,7 +1061,7 @@ fhandler_tty_slave::ioctl (unsigned int cmd, void *arg)
   termios_printf ("ioctl (%x)", cmd);
 
   if (myself->pgid && get_ttyp ()->getpgid () != myself->pgid
-      && myself->ctty == ttynum && (get_ttyp ()->ti.c_lflag & TOSTOP))
+      && myself->ctty == get_unit () && (get_ttyp ()->ti.c_lflag & TOSTOP))
     {
       /* background process */
       termios_printf ("bg ioctl pgid %d, tpgid %d, ctty %d",
@@ -1025,13 +1069,16 @@ fhandler_tty_slave::ioctl (unsigned int cmd, void *arg)
       raise (SIGTTOU);
     }
 
+  int retval;
   switch (cmd)
     {
     case TIOCGWINSZ:
     case TIOCSWINSZ:
+    case TIOCLINUX:
       break;
     case FIONBIO:
       set_nonblocking (*(int *) arg);
+      retval = 0;
       goto out;
     default:
       set_errno (EINVAL);
@@ -1060,54 +1107,75 @@ fhandler_tty_slave::ioctl (unsigned int cmd, void *arg)
 	  get_ttyp ()->arg.winsize = *(struct winsize *) arg;
 	  if (ioctl_request_event)
 	    {
-	      get_ttyp ()->ioctl_retval = -1;
+	      get_ttyp ()->ioctl_retval = -EINVAL;
 	      SetEvent (ioctl_request_event);
 	    }
 	  else
 	    {
 	      get_ttyp ()->winsize = *(struct winsize *) arg;
-	      kill (-get_ttyp ()->getpgid (), SIGWINCH);
+	      killsys (-get_ttyp ()->getpgid (), SIGWINCH);
 	    }
 	  if (ioctl_done_event)
 	    WaitForSingleObject (ioctl_done_event, INFINITE);
 	}
       break;
+    case TIOCLINUX:
+      int val = *(unsigned char *) arg;
+      if (val != 6 || !ioctl_request_event || !ioctl_done_event)
+	  get_ttyp ()->ioctl_retval = -EINVAL;
+      else
+	{
+	  get_ttyp ()->arg.value = val;
+	  SetEvent (ioctl_request_event);
+	  WaitForSingleObject (ioctl_done_event, INFINITE);
+	  *(unsigned char *) arg = get_ttyp ()->arg.value & 0xFF;
+	}
+      break;
     }
 
   release_output_mutex ();
+  retval = get_ttyp ()->ioctl_retval;
+  if (retval < 0)
+    {
+      set_errno (-retval);
+      retval = -1;
+    }
 
 out:
-  termios_printf ("%d = ioctl (%x)", get_ttyp ()->ioctl_retval, cmd);
-  return get_ttyp ()->ioctl_retval;
+  termios_printf ("%d = ioctl (%x)", retval, cmd);
+  return retval;
 }
 
 /*******************************************************
  fhandler_pty_master
 */
-fhandler_pty_master::fhandler_pty_master (DWORD devtype, int unit)
-  : fhandler_tty_common (devtype, unit)
+fhandler_pty_master::fhandler_pty_master ()
+  : fhandler_tty_common ()
 {
 }
 
 int
-fhandler_pty_master::open (path_conv *, int flags, mode_t)
+fhandler_pty_master::open (int flags, mode_t)
 {
-  ttynum = cygwin_shared->tty.allocate_tty (0);
-  if (ttynum < 0)
+  int ntty = cygwin_shared->tty.allocate_tty (0);
+  if (ntty < 0)
     return 0;
 
-  cygwin_shared->tty[ttynum]->common_init (this);
+  slave = *ttys_dev;
+  slave.setunit (ntty);
+  cygwin_shared->tty[ntty]->common_init (this);
   inuse = get_ttyp ()->create_inuse (TTY_MASTER_ALIVE);
   set_flags ((flags & ~O_TEXT) | O_BINARY);
   set_open_status ();
 
-  termios_printf ("opened pty master tty%d<%p>", ttynum, this);
+  termios_printf ("opened pty master tty%d", get_unit ());
   return 1;
 }
 
 int
 fhandler_tty_common::close ()
 {
+  termios_printf ("tty%d <%p,%p> closing", get_unit (), get_handle (), get_output_handle ());
   if (output_done_event && !CloseHandle (output_done_event))
     termios_printf ("CloseHandle (output_done_event), %E");
   if (ioctl_done_event && !CloseHandle (ioctl_done_event))
@@ -1124,7 +1192,7 @@ fhandler_tty_common::close ()
   /* Send EOF to slaves if master side is closed */
   if (!get_ttyp ()->master_alive ())
     {
-      termios_printf ("no more masters left. sending EOF" );
+      termios_printf ("no more masters left. sending EOF");
       SetEvent (input_available_event);
     }
 
@@ -1136,7 +1204,7 @@ fhandler_tty_common::close ()
     termios_printf ("CloseHandle (get_output_handle ()<%p>), %E", get_output_handle ());
 
   inuse = NULL;
-  termios_printf ("tty%d <%p,%p> closed", ttynum, get_handle (), get_output_handle ());
+  set_io_handle (NULL);
   return 0;
 }
 
@@ -1151,7 +1219,7 @@ fhandler_pty_master::close ()
 
   if (!get_ttyp ()->master_alive ())
     {
-      termios_printf ("freeing tty%d (%d)", ttynum, get_ttyp ()->ntty);
+      termios_printf ("freeing tty%d (%d)", get_unit (), get_ttyp ()->ntty);
 #if 0
       if (get_ttyp ()->to_slave)
 	ForceCloseHandle1 (get_ttyp ()->to_slave, to_slave);
@@ -1198,14 +1266,14 @@ fhandler_pty_master::read (void *ptr, size_t& len)
 int
 fhandler_pty_master::tcgetattr (struct termios *t)
 {
-  *t = cygwin_shared->tty[ttynum]->ti;
+  *t = cygwin_shared->tty[get_unit ()]->ti;
   return 0;
 }
 
 int
 fhandler_pty_master::tcsetattr (int, const struct termios *t)
 {
-  cygwin_shared->tty[ttynum]->ti = *t;
+  cygwin_shared->tty[get_unit ()]->ti = *t;
   return 0;
 }
 
@@ -1233,7 +1301,7 @@ fhandler_pty_master::ioctl (unsigned int cmd, void *arg)
   switch (cmd)
     {
       case TIOCPKT:
-	pktmode = * (int *) arg;
+	pktmode = *(int *) arg;
 	break;
       case TIOCGWINSZ:
 	*(struct winsize *) arg = get_ttyp ()->winsize;
@@ -1242,8 +1310,8 @@ fhandler_pty_master::ioctl (unsigned int cmd, void *arg)
 	if (get_ttyp ()->winsize.ws_row != ((struct winsize *) arg)->ws_row
 	    || get_ttyp ()->winsize.ws_col != ((struct winsize *) arg)->ws_col)
 	  {
-	    get_ttyp ()->winsize = * (struct winsize *) arg;
-	    kill (-get_ttyp ()->getpgid (), SIGWINCH);
+	    get_ttyp ()->winsize = *(struct winsize *) arg;
+	    killsys (-get_ttyp ()->getpgid (), SIGWINCH);
 	  }
 	break;
       case FIONBIO:
@@ -1257,46 +1325,50 @@ fhandler_pty_master::ioctl (unsigned int cmd, void *arg)
 }
 
 char *
-fhandler_pty_master::ptsname (void)
+fhandler_pty_master::ptsname ()
 {
   static char buf[32];
 
-  __small_sprintf (buf, "/dev/tty%d", ttynum);
+  __small_sprintf (buf, "/dev/tty%d", get_unit ());
   return buf;
 }
 
 void
 fhandler_tty_common::set_close_on_exec (int val)
 {
+  if (archetype)
+    set_close_on_exec_flag (val);
+  else
+    {
+      if (output_done_event)
+	set_no_inheritance (output_done_event, val);
+      if (ioctl_request_event)
+	set_no_inheritance (ioctl_request_event, val);
+      if (ioctl_done_event)
+	set_no_inheritance (ioctl_done_event, val);
+      if (inuse)
+	set_no_inheritance (inuse, val);
+      set_no_inheritance (output_mutex, val);
+      set_no_inheritance (input_mutex, val);
+      set_no_inheritance (input_available_event, val);
+      set_no_inheritance (output_handle, val);
 #ifndef DEBUGGING
-  fhandler_base::set_close_on_exec (val);
+      fhandler_base::set_close_on_exec (val);
 #else
-  /* FIXME: This is a duplication from fhandler_base::set_close_on_exec.
-     It is here because we need to specify the "from_pty" stuff here or
-     we'll get warnings from ForceCloseHandle when debugging. */
-  set_inheritance (get_io_handle (), val);
-  set_close_on_exec_flag (val);
+      /* FIXME: This is a duplication from fhandler_base::set_close_on_exec.
+	 It is here because we need to specify the "from_pty" stuff here or
+	 we'll get warnings from ForceCloseHandle when debugging. */
+      set_no_inheritance (get_io_handle (), val);
+      set_close_on_exec_flag (val);
 #endif
-  if (output_done_event)
-    set_inheritance (output_done_event, val);
-  if (ioctl_request_event)
-    set_inheritance (ioctl_request_event, val);
-  if (ioctl_done_event)
-    set_inheritance (ioctl_done_event, val);
-  if (inuse)
-    set_inheritance (inuse, val);
-  set_inheritance (output_mutex, val);
-  set_inheritance (input_mutex, val);
-  set_inheritance (input_available_event, val);
-  set_inheritance (output_handle, val);
+    }
 }
 
 void
 fhandler_tty_slave::fixup_after_fork (HANDLE parent)
 {
-  fhandler_console::open_fhs++;
-  termios_printf ("incremented open_fhs %d", fhandler_console::open_fhs);
-  fhandler_tty_common::fixup_after_fork (parent);
+  // fhandler_tty_common::fixup_after_fork (parent);
+  report_tty_counts (this, "inherited", "", "");
 }
 
 void
@@ -1333,29 +1405,15 @@ fhandler_pty_master::set_close_on_exec (int val)
     }
 }
 
-void
-fhandler_tty_master::fixup_after_fork (HANDLE child)
-{
-  fhandler_pty_master::fixup_after_fork (child);
-  console->fixup_after_fork (child);
-}
-
-void
-fhandler_tty_master::fixup_after_exec (HANDLE)
-{
-  console->close ();
-  init_console ();
-}
-
 int
 fhandler_tty_master::init_console ()
 {
-  console = (fhandler_console *) cygheap->fdtab.build_fhandler (-1, FH_CONSOLE, "/dev/ttym");
+  console = (fhandler_console *) build_fh_dev (*console_dev, "/dev/ttym");
   if (console == NULL)
     return -1;
 
   console->init (INVALID_HANDLE_VALUE, GENERIC_READ | GENERIC_WRITE, O_BINARY);
-  fhandler_console::open_fhs--;  /* handled when individual fds are opened */
+  cygheap->open_fhs--;  	/* handled when individual fds are opened */
   console->set_r_no_interrupt (1);
   return 0;
 }
