@@ -1,6 +1,6 @@
 /* fhandler_virtual.cc: base fhandler class for virtual filesystems
 
-   Copyright 2002 Red Hat, Inc.
+   Copyright 2002, 2003, 2004 Red Hat, Inc.
 
 This file is part of Cygwin.
 
@@ -14,8 +14,8 @@ details. */
 #include <sys/cygwin.h>
 #include "cygerrno.h"
 #include "security.h"
-#include "fhandler.h"
 #include "path.h"
+#include "fhandler.h"
 #include "dtable.h"
 #include "shared_info.h"
 #include "cygheap.h"
@@ -23,8 +23,8 @@ details. */
 
 #include <dirent.h>
 
-fhandler_virtual::fhandler_virtual (DWORD devtype):
-  fhandler_base (devtype), filebuf (NULL), bufalloc ((size_t) -1),
+fhandler_virtual::fhandler_virtual ():
+  fhandler_base (), filebuf (NULL), bufalloc ((size_t) -1),
   fileid (-1)
 {
 }
@@ -37,14 +37,14 @@ fhandler_virtual::~fhandler_virtual ()
 }
 
 void
-fhandler_virtual::fixup_after_exec (HANDLE)
+fhandler_virtual::fixup_after_exec ()
 {
   if (filebuf)
     filebuf = NULL;
 }
 
 DIR *
-fhandler_virtual::opendir (path_conv& pc)
+fhandler_virtual::opendir ()
 {
   DIR *dir;
   DIR *res = NULL;
@@ -52,7 +52,7 @@ fhandler_virtual::opendir (path_conv& pc)
 
   if (exists () <= 0)
     set_errno (ENOTDIR);
-  else if ((len = strlen (get_name ())) > MAX_PATH - 3)
+  else if ((len = strlen (get_name ())) > CYG_MAX_PATH - 3)
     set_errno (ENAMETOOLONG);
   else if ((dir = (DIR *) malloc (sizeof (DIR))) == NULL)
     set_errno (ENOMEM);
@@ -167,7 +167,7 @@ fhandler_virtual::close ()
     free (filebuf);
   filebuf = NULL;
   bufalloc = (size_t) -1;
-  cygwin_shared->delqueue.process_queue ();
+  user_shared->delqueue.process_queue ();
   return 0;
 }
 
@@ -205,14 +205,11 @@ fhandler_virtual::write (const void *ptr, size_t len)
 
 /* low-level open for all proc files */
 int
-fhandler_virtual::open (path_conv *, int flags, mode_t mode)
+fhandler_virtual::open (int flags, mode_t mode)
 {
   set_r_binary (1);
   set_w_binary (1);
 
-  set_has_acls (false);
-  set_isremote (false);
-
   /* what to do about symlinks? */
   set_symlink_p (false);
   set_execable_p (not_executable);
