@@ -1,6 +1,6 @@
 /* fhandler_windows.cc: code to access windows message queues.
 
-   Copyright 1998, 1999, 2000, 2001, 2002 Red Hat, Inc.
+   Copyright 1998, 1999, 2000, 2001, 2002, 2003, 2004 Red Hat, Inc.
 
    Written by Sergey S. Okhapkin (sos@prospect.com.ru).
    Feedback and testing by Andy Piper (andyp@parallax.co.uk).
@@ -16,6 +16,7 @@ details. */
 #include <winuser.h>
 #include "cygerrno.h"
 #include "security.h"
+#include "path.h"
 #include "fhandler.h"
 
 /*
@@ -46,12 +47,12 @@ The following unix-style calls are supported:
 */
 
 fhandler_windows::fhandler_windows ()
-  : fhandler_base (FH_WINDOWS), hWnd_ (NULL), method_ (WINDOWS_POST)
+  : fhandler_base (), hWnd_ (NULL), method_ (WINDOWS_POST)
 {
 }
 
 int
-fhandler_windows::open (path_conv *, int flags, mode_t)
+fhandler_windows::open (int flags, mode_t)
 {
   set_flags ((flags & ~O_TEXT) | O_BINARY);
   set_close_on_exec_flag (1);
@@ -132,7 +133,7 @@ fhandler_windows::set_close_on_exec (int val)
     fhandler_base::set_close_on_exec_flag (val);
   void *h = hWnd_;
   if (h)
-    set_inheritance (h, val);
+    set_no_inheritance (h, val);
 }
 
 void
