@@ -1,6 +1,6 @@
 /* fhandler_dev_zero.cc: code to access /dev/zero
 
-   Copyright 2000, 2001, 2002 Red Hat, Inc.
+   Copyright 2000, 2001, 2002, 2003, 2004 Red Hat, Inc.
 
    Written by DJ Delorie (dj@cygnus.com)
 
@@ -13,15 +13,16 @@ details. */
 #include "winsup.h"
 #include <errno.h>
 #include "security.h"
+#include "path.h"
 #include "fhandler.h"
 
 fhandler_dev_zero::fhandler_dev_zero ()
-  : fhandler_base (FH_ZERO)
+  : fhandler_base ()
 {
 }
 
 int
-fhandler_dev_zero::open (path_conv *, int flags, mode_t)
+fhandler_dev_zero::open (int flags, mode_t)
 {
   set_flags ((flags & ~O_TEXT) | O_BINARY);
   set_nohandle (true);
