@@ -1,6 +1,6 @@
 /* net.cc: network-related routines.
 
-   Copyright 1996, 1997, 1998, 1999, 2000, 2001, 2002, 2003 Red Hat, Inc.
+   Copyright 1996, 1997, 1998, 1999, 2000, 2001, 2002, 2003, 2004 Red Hat, Inc.
 
 This file is part of Cygwin.
 
@@ -28,14 +28,15 @@ details. */
 #include <assert.h>
 #include "cygerrno.h"
 #include "security.h"
-#include "fhandler.h"
 #include "path.h"
+#include "fhandler.h"
 #include "dtable.h"
 #include "cygheap.h"
 #include "sigproc.h"
 #include "pinfo.h"
 #include "registry.h"
 #include "wsock_event.h"
+#include "cygtls.h"
 
 extern "C"
 {
@@ -129,6 +130,7 @@ set_socket_inheritance (SOCKET sock)
     system_printf ("DuplicateHandle failed %E");
   else
     debug_printf ("DuplicateHandle succeeded osock %p, sock %p", osock, sock);
+  VerifyHandle ((HANDLE) sock);
   return sock;
 }
 
@@ -175,22 +177,16 @@ dump_protoent (struct protoent *p)
 extern "C" char *
 cygwin_inet_ntoa (struct in_addr in)
 {
-#ifdef _MT_SAFE
-#define ntoa_buf  _reent_winsup ()->_ntoa_buf
-#else
-  static char *ntoa_buf = NULL;
-#endif
-
   char *res = inet_ntoa (in);
 
-  if (ntoa_buf)
+  if (_my_tls.locals.ntoa_buf)
     {
-      free (ntoa_buf);
-      ntoa_buf = NULL;
+      free (_my_tls.locals.ntoa_buf);
+      _my_tls.locals.ntoa_buf = NULL;
     }
   if (res)
-    ntoa_buf = strdup (res);
-  return ntoa_buf;
+    _my_tls.locals.ntoa_buf = strdup (res);
+  return _my_tls.locals.ntoa_buf;
 }
 
 /* exported as inet_addr: BSD 4.3 */
@@ -499,8 +495,20 @@ dup_ent (void *old, void *src0, struct_type type)
   /* Do servent/hostent specific processing */
   int protolen = 0;
   int addr_list_len = 0;
+  char *s_proto = NULL;
   if (type == is_servent)
-    sz += (protolen = strlen_round (src->s_proto));
+    {
+      if (src->s_proto)
+	{
+	  /* Windows 95 idiocy.  Structure is misaligned on Windows 95.
+	     Kludge around this by trying a different pointer alignment.  */
+	  if (!IsBadStringPtr (src->s_proto, INT32_MAX))
+	    s_proto = src->s_proto;
+	  else if (!IsBadStringPtr (((pservent *) src)->s_proto, INT32_MAX))
+	    s_proto = ((pservent *) src)->s_proto;
+	  sz += (protolen = strlen_round (s_proto));
+	}
+    }
   else if (type == is_hostent)
     {
       /* Calculate the length and storage used for h_addr_list */
@@ -549,16 +557,8 @@ dup_ent (void *old, void *src0, struct_type type)
       /* Do servent/hostent specific processing. */
       if (type == is_servent)
 	{
-	  if (src->s_proto)
+	  if (s_proto)
 	    {
-	      char *s_proto;
-	      /* Windows 95 idiocy.  Structure is misaligned on Windows 95.
-		 Kludge around this by trying a different pointer alignment.  */
-	      if (IsBadReadPtr (src->s_proto, sizeof (src->s_proto))
-		  && !IsBadReadPtr (((pservent *) src)->s_proto, sizeof (src->s_proto)))
-		s_proto = ((pservent *) src)->s_proto;
-	      else
-		s_proto = src->s_proto;
 	      strcpy (dst->s_proto = dp, s_proto);
 	      dp += protolen;
 	    }
@@ -583,42 +583,38 @@ dup_ent (void *old, void *src0, struct_type type)
   return dst;
 }
 
-#ifdef _MT_SAFE
-#define protoent_buf  _reent_winsup ()->_protoent_buf
-#else
-static struct protoent *protoent_buf = NULL;
-#endif
-
 /* exported as getprotobyname: standards? */
 extern "C" struct protoent *
 cygwin_getprotobyname (const char *p)
 {
   if (check_null_str_errno (p))
     return NULL;
-  protoent_buf = (protoent *) dup_ent (protoent_buf, getprotobyname (p),
-				       is_protoent);
-  if (!protoent_buf)
+  _my_tls.locals.protoent_buf =
+    (protoent *) dup_ent (_my_tls.locals.protoent_buf, getprotobyname (p),
+			  is_protoent);
+  if (!_my_tls.locals.protoent_buf)
     set_winsock_errno ();
 
-  dump_protoent (protoent_buf);
-  return protoent_buf;
+  dump_protoent (_my_tls.locals.protoent_buf);
+  return _my_tls.locals.protoent_buf;
 }
 
 /* exported as getprotobynumber: standards? */
 extern "C" struct protoent *
 cygwin_getprotobynumber (int number)
 {
-  protoent_buf = (protoent *) dup_ent (protoent_buf, getprotobynumber (number),
-				       is_protoent);
-  if (!protoent_buf)
+  _my_tls.locals.protoent_buf =
+    (protoent *) dup_ent (_my_tls.locals.protoent_buf,
+			  getprotobynumber (number), is_protoent);
+  if (!_my_tls.locals.protoent_buf)
     set_winsock_errno ();
 
-  dump_protoent (protoent_buf);
-  return protoent_buf;
+  dump_protoent (_my_tls.locals.protoent_buf);
+  return _my_tls.locals.protoent_buf;
 }
 
-fhandler_socket *
-fdsock (int &fd, const char *name, SOCKET soc)
+bool
+fdsock (cygheap_fdmanip& fd, const device *dev, SOCKET soc)
 {
   if (!winsock2_active)
     soc = set_socket_inheritance (soc);
@@ -632,16 +628,15 @@ fdsock (int &fd, const char *name, SOCKET soc)
   else
     debug_printf ("not setting socket inheritance since winsock2_active %d",
 		  winsock2_active);
-  fhandler_socket *fh = (fhandler_socket *)
-	cygheap->fdtab.build_fhandler (fd, FH_SOCKET, name, NULL,
-				       tolower (name[5]) - 'a');
-  if (!fh)
-    return NULL;
-  fh->set_io_handle ((HANDLE) soc);
-  fh->set_flags (O_RDWR | O_BINARY);
-  fh->set_r_no_interrupt (winsock2_active);
-  debug_printf ("fd %d, name '%s', soc %p", fd, name, soc);
-  return fh;
+  fd = build_fh_dev (*dev);
+  if (!fd.isopen ())
+    return false;
+  fd->set_io_handle ((HANDLE) soc);
+  fd->set_flags (O_RDWR | O_BINARY);
+  fd->set_r_no_interrupt (winsock2_active);
+  cygheap->fdtab.inc_need_fixup_before ();
+  debug_printf ("fd %d, name '%s', soc %p", (int) fd, dev->name, soc);
+  return true;
 }
 
 /* exported as socket: standards? */
@@ -650,7 +645,6 @@ cygwin_socket (int af, int type, int protocol)
 {
   int res = -1;
   SOCKET soc = 0;
-  fhandler_socket *fh = NULL;
 
   debug_printf ("socket (%d, %d, %d)", af, type, protocol);
 
@@ -662,25 +656,23 @@ cygwin_socket (int af, int type, int protocol)
       goto done;
     }
 
-  const char *name;
+  const device *dev;
 
   if (af == AF_INET)
-    name = (type == SOCK_STREAM ? "/dev/tcp" : "/dev/udp");
+    dev = type == SOCK_STREAM ? tcp_dev : udp_dev;
   else
-    name = (type == SOCK_STREAM ? "/dev/streamsocket" : "/dev/dgsocket");
+    dev = type == SOCK_STREAM ? stream_dev : dgram_dev;
 
   {
     cygheap_fdnew fd;
-    if (fd >= 0)
-      fh = fdsock (fd, name, soc);
-    if (fh)
+    if (fd < 0 || !fdsock (fd, dev, soc))
+      closesocket (soc);
+    else
       {
-	fh->set_addr_family (af);
-	fh->set_socket_type (type);
+	((fhandler_socket *) fd)->set_addr_family (af);
+	((fhandler_socket *) fd)->set_socket_type (type);
 	res = fd;
       }
-    else
-	closesocket (soc);
   }
 
 done:
@@ -695,7 +687,6 @@ cygwin_sendto (int fd, const void *buf, int len, int flags,
 {
   int res;
   sig_dispatch_pending ();
-  sigframe thisframe (mainthread);
 
   fhandler_socket *fh = get (fd);
 
@@ -719,7 +710,6 @@ cygwin_recvfrom (int fd, void *buf, int len, int flags,
 {
   int res;
   sig_dispatch_pending ();
-  sigframe thisframe (mainthread);
 
   fhandler_socket *fh = get (fd);
 
@@ -876,7 +866,6 @@ cygwin_connect (int fd, const struct sockaddr *name, int namelen)
 {
   int res;
   sig_dispatch_pending ();
-  sigframe thisframe (mainthread);
 
   fhandler_socket *fh = get (fd);
 
@@ -939,29 +928,22 @@ cygwin_connect (int fd, const struct sockaddr *name, int namelen)
   return res;
 }
 
-#ifdef _MT_SAFE
-#define servent_buf  _reent_winsup ()->_servent_buf
-#else
-static struct servent *servent_buf = NULL;
-#endif
-
 /* exported as getservbyname: standards? */
 extern "C" struct servent *
 cygwin_getservbyname (const char *name, const char *proto)
 {
   sig_dispatch_pending ();
-  sigframe thisframe (mainthread);
   if (check_null_str_errno (name)
       || (proto != NULL && check_null_str_errno (proto)))
     return NULL;
 
-  servent_buf = (servent *) dup_ent (servent_buf, getservbyname (name, proto),
+  _my_tls.locals.servent_buf = (servent *) dup_ent (_my_tls.locals.servent_buf, getservbyname (name, proto),
 				     is_servent);
-  if (!servent_buf)
+  if (!_my_tls.locals.servent_buf)
     set_winsock_errno ();
 
-  syscall_printf ("%x = getservbyname (%s, %s)", servent_buf, name, proto);
-  return servent_buf;
+  syscall_printf ("%x = getservbyname (%s, %s)", _my_tls.locals.servent_buf, name, proto);
+  return _my_tls.locals.servent_buf;
 }
 
 /* exported as getservbyport: standards? */
@@ -969,24 +951,22 @@ extern "C" struct servent *
 cygwin_getservbyport (int port, const char *proto)
 {
   sig_dispatch_pending ();
-  sigframe thisframe (mainthread);
   if (proto != NULL && check_null_str_errno (proto))
     return NULL;
 
-  servent_buf = (servent *) dup_ent (servent_buf, getservbyport (port, proto),
+  _my_tls.locals.servent_buf = (servent *) dup_ent (_my_tls.locals.servent_buf, getservbyport (port, proto),
 				     is_servent);
-  if (!servent_buf)
+  if (!_my_tls.locals.servent_buf)
     set_winsock_errno ();
 
-  syscall_printf ("%x = getservbyport (%d, %s)", servent_buf, port, proto);
-  return servent_buf;
+  syscall_printf ("%x = getservbyport (%d, %s)", _my_tls.locals.servent_buf, port, proto);
+  return _my_tls.locals.servent_buf;
 }
 
 extern "C" int
 cygwin_gethostname (char *name, size_t len)
 {
   sig_dispatch_pending ();
-  sigframe thisframe (mainthread);
   if (__check_null_invalid_struct_errno (name, len))
     return -1;
 
@@ -1005,12 +985,6 @@ cygwin_gethostname (char *name, size_t len)
   return 0;
 }
 
-#ifdef _MT_SAFE
-#define hostent_buf  _reent_winsup ()->_hostent_buf
-#else
-static struct hostent *hostent_buf = NULL;
-#endif
-
 /* exported as gethostbyname: standards? */
 extern "C" struct hostent *
 cygwin_gethostbyname (const char *name)
@@ -1022,7 +996,6 @@ cygwin_gethostbyname (const char *name)
   static int a, b, c, d;
 
   sig_dispatch_pending ();
-  sigframe thisframe (mainthread);
   if (check_null_str_errno (name))
     return NULL;
 
@@ -1043,19 +1016,19 @@ cygwin_gethostbyname (const char *name)
       return &tmp;
     }
 
-  hostent_buf = (hostent *) dup_ent (hostent_buf, gethostbyname (name),
+  _my_tls.locals.hostent_buf = (hostent *) dup_ent (_my_tls.locals.hostent_buf, gethostbyname (name),
 				     is_hostent);
-  if (!hostent_buf)
+  if (!_my_tls.locals.hostent_buf)
     {
       set_winsock_errno ();
       set_host_errno ();
     }
   else
     {
-      debug_printf ("h_name %s", hostent_buf->h_name);
+      debug_printf ("h_name %s", _my_tls.locals.hostent_buf->h_name);
       h_errno = 0;
     }
-  return hostent_buf;
+  return _my_tls.locals.hostent_buf;
 }
 
 /* exported as gethostbyaddr: standards? */
@@ -1063,24 +1036,23 @@ extern "C" struct hostent *
 cygwin_gethostbyaddr (const char *addr, int len, int type)
 {
   sig_dispatch_pending ();
-  sigframe thisframe (mainthread);
   if (__check_invalid_read_ptr_errno (addr, len))
     return NULL;
 
-  hostent_buf = (hostent *) dup_ent (hostent_buf,
+  _my_tls.locals.hostent_buf = (hostent *) dup_ent (_my_tls.locals.hostent_buf,
 				     gethostbyaddr (addr, len, type),
 				     is_hostent);
-  if (!hostent_buf)
+  if (!_my_tls.locals.hostent_buf)
     {
       set_winsock_errno ();
       set_host_errno ();
     }
   else
     {
-      debug_printf ("h_name %s", hostent_buf->h_name);
+      debug_printf ("h_name %s", _my_tls.locals.hostent_buf->h_name);
       h_errno = 0;
     }
-  return hostent_buf;
+  return _my_tls.locals.hostent_buf;
 }
 
 /* exported as accept: standards? */
@@ -1089,7 +1061,6 @@ cygwin_accept (int fd, struct sockaddr *peer, int *len)
 {
   int res;
   sig_dispatch_pending ();
-  sigframe thisframe (mainthread);
 
   fhandler_socket *fh = get (fd);
 
@@ -1122,7 +1093,6 @@ cygwin_bind (int fd, const struct sockaddr *my_addr, int addrlen)
 {
   int res;
   sig_dispatch_pending ();
-  sigframe thisframe (mainthread);
   fhandler_socket *fh = get (fd);
 
   if (__check_invalid_read_ptr_errno (my_addr, addrlen) || !fh)
@@ -1140,7 +1110,6 @@ cygwin_getsockname (int fd, struct sockaddr *addr, int *namelen)
 {
   int res;
   sig_dispatch_pending ();
-  sigframe thisframe (mainthread);
 
   fhandler_socket *fh = get (fd);
 
@@ -1161,7 +1130,6 @@ cygwin_listen (int fd, int backlog)
 {
   int res;
   sig_dispatch_pending ();
-  sigframe thisframe (mainthread);
   fhandler_socket *fh = get (fd);
 
   if (!fh)
@@ -1179,7 +1147,6 @@ cygwin_shutdown (int fd, int how)
 {
   int res;
   sig_dispatch_pending ();
-  sigframe thisframe (mainthread);
 
   fhandler_socket *fh = get (fd);
 
@@ -1245,7 +1212,6 @@ cygwin_getpeername (int fd, struct sockaddr *name, int *len)
 {
   int res;
   sig_dispatch_pending ();
-  sigframe thisframe (mainthread);
 
   fhandler_socket *fh = get (fd);
 
@@ -1286,7 +1252,6 @@ getdomainname (char *domain, size_t len)
    * Punt for now and assume MS-TCP on Win95.
    */
   sig_dispatch_pending ();
-  sigframe thisframe (mainthread);
   if (__check_null_invalid_struct_errno (domain, len))
     return -1;
 
@@ -1900,7 +1865,6 @@ get_ifconf (struct ifconf *ifc, int what)
   struct sockaddr_in *sa;
 
   sig_dispatch_pending ();
-  sigframe thisframe (mainthread);
   if (check_null_invalid_struct_errno (ifc))
     return -1;
 
@@ -1979,7 +1943,6 @@ cygwin_rcmd (char **ahost, unsigned short inport, char *locuser,
   SOCKET fd2s;
 
   sig_dispatch_pending ();
-  sigframe thisframe (mainthread);
 
   if (check_null_invalid_struct_errno (ahost) ||
       check_null_empty_str_errno (*ahost) ||
@@ -1990,14 +1953,11 @@ cygwin_rcmd (char **ahost, unsigned short inport, char *locuser,
   res = rcmd (ahost, inport, locuser, remuser, cmd, fd2p ? &fd2s : NULL);
   if (res != (int) INVALID_SOCKET)
     {
-      fhandler_socket *fh = NULL;
       cygheap_fdnew res_fd;
 
-      if (res_fd >= 0)
-	fh = fdsock (res_fd, "/dev/tcp", res);
-      if (fh)
+      if (res_fd >= 0 && fdsock (res_fd, tcp_dev, res))
 	{
-	  fh->set_connect_state (CONNECTED);
+	  ((fhandler_socket *) res_fd)->set_connect_state (CONNECTED);
 	  res = res_fd;
 	}
       else
@@ -2009,14 +1969,12 @@ cygwin_rcmd (char **ahost, unsigned short inport, char *locuser,
       if (res >= 0 && fd2p)
 	{
 	  cygheap_fdnew newfd (res_fd, false);
+	  cygheap_fdget fd (*fd2p);
 
-	  fh = NULL;
-	  if (newfd >= 0)
-	    fh = fdsock (newfd, "/dev/tcp", fd2s);
-	  if (fh)
+	  if (newfd >= 0 && fdsock (newfd, tcp_dev, fd2s))
 	    {
 	      *fd2p = newfd;
-	      fh->set_connect_state (CONNECTED);
+	      ((fhandler_socket *) fd2p)->set_connect_state (CONNECTED);
 	    }
 	  else
 	    {
@@ -2037,7 +1995,6 @@ cygwin_rresvport (int *port)
 {
   int res;
   sig_dispatch_pending ();
-  sigframe thisframe (mainthread);
 
   if (check_null_invalid_struct_errno (port))
     return -1;
@@ -2046,13 +2003,10 @@ cygwin_rresvport (int *port)
 
   if (res != (int) INVALID_SOCKET)
     {
-      fhandler_socket *fh = NULL;
       cygheap_fdnew res_fd;
 
-      if (res_fd >= 0)
-	fh = fdsock (res_fd, "/dev/tcp", res);
-      if (fh)
-	res = res_fd;
+      if (res_fd >= 0 && fdsock (res_fd, tcp_dev, res))
+        res = res_fd;
       else
 	res = -1;
     }
@@ -2069,7 +2023,6 @@ cygwin_rexec (char **ahost, unsigned short inport, char *locuser,
   int res = -1;
   SOCKET fd2s;
   sig_dispatch_pending ();
-  sigframe thisframe (mainthread);
 
   if (check_null_invalid_struct_errno (ahost) ||
       check_null_empty_str_errno (*ahost) ||
@@ -2080,14 +2033,11 @@ cygwin_rexec (char **ahost, unsigned short inport, char *locuser,
   res = rexec (ahost, inport, locuser, password, cmd, fd2p ? &fd2s : NULL);
   if (res != (int) INVALID_SOCKET)
     {
-      fhandler_socket *fh = NULL;
       cygheap_fdnew res_fd;
 
-      if (res_fd >= 0)
-	fh = fdsock (res_fd, "/dev/tcp", res);
-      if (fh)
+      if (res_fd >= 0 && fdsock (res_fd, tcp_dev, res))
 	{
-	  fh->set_connect_state (CONNECTED);
+	  ((fhandler_socket *) res_fd)->set_connect_state (CONNECTED);
 	  res = res_fd;
 	}
       else
@@ -2099,13 +2049,11 @@ cygwin_rexec (char **ahost, unsigned short inport, char *locuser,
       if (res >= 0 && fd2p)
 	{
 	  cygheap_fdnew newfd (res_fd, false);
+	  cygheap_fdget fd (*fd2p);
 
-	  fh = NULL;
-	  if (newfd >= 0)
-	    fh = fdsock (newfd, "/dev/tcp", fd2s);
-	  if (fh)
+	  if (newfd >= 0 && fdsock (newfd, tcp_dev, fd2s))
 	    {
-	      fh->set_connect_state (CONNECTED);
+	      ((fhandler_socket *) fd2p)->set_connect_state (CONNECTED);
 	      *fd2p = newfd;
 	    }
 	  else
@@ -2132,7 +2080,6 @@ socketpair (int family, int type, int protocol, int *sb)
   int len;
 
   sig_dispatch_pending ();
-  sigframe thisframe (mainthread);
   if (__check_null_invalid_struct_errno (sb, 2 * sizeof (int)))
     return -1;
 
@@ -2267,35 +2214,29 @@ socketpair (int family, int type, int protocol, int *sb)
     }
 
   {
-    fhandler_socket *fh = NULL;
     cygheap_fdnew sb0;
-    const char *name;
+    const device *dev;
 
     if (family == AF_INET)
-      name = (type == SOCK_STREAM ? "/dev/tcp" : "/dev/udp");
+      dev = (type == SOCK_STREAM ? tcp_dev : udp_dev);
     else
-      name = (type == SOCK_STREAM ? "/dev/streamsocket" : "/dev/dgsocket");
+      dev = (type == SOCK_STREAM ? stream_dev : dgram_dev);
 
-    if (sb0 >= 0)
-      fh = fdsock (sb0, name, insock);
-    if (fh)
+    if (sb0 >= 0 && fdsock (sb0, dev, insock))
       {
-	fh->set_sun_path ("");
-	fh->set_addr_family (family);
-	fh->set_socket_type (type);
-	fh->set_connect_state (CONNECTED);
+	((fhandler_socket *) sb0)->set_sun_path ("");
+	((fhandler_socket *) sb0)->set_addr_family (family);
+	((fhandler_socket *) sb0)->set_socket_type (type);
+	((fhandler_socket *) sb0)->set_connect_state (CONNECTED);
 
 	cygheap_fdnew sb1 (sb0, false);
 
-	fh = NULL;
-	if (sb1 >= 0)
-	  fh = fdsock (sb1, name, outsock);
-	if (fh)
+	if (sb1 >= 0 && fdsock (sb1, dev, outsock))
 	  {
-	    fh->set_sun_path ("");
-	    fh->set_addr_family (family);
-	    fh->set_socket_type (type);
-	    fh->set_connect_state (CONNECTED);
+	    ((fhandler_socket *) sb1)->set_sun_path ("");
+	    ((fhandler_socket *) sb1)->set_addr_family (family);
+	    ((fhandler_socket *) sb1)->set_socket_type (type);
+	    ((fhandler_socket *) sb1)->set_connect_state (CONNECTED);
 
 	    sb[0] = sb0;
 	    sb[1] = sb1;
@@ -2333,7 +2274,6 @@ cygwin_recvmsg (int fd, struct msghdr *msg, int flags)
 {
   int res;
   sig_dispatch_pending ();
-  sigframe thisframe (mainthread);
 
   fhandler_socket *fh = get (fd);
 
@@ -2360,7 +2300,6 @@ cygwin_sendmsg (int fd, const struct msghdr *msg, int flags)
 {
   int res;
   sig_dispatch_pending ();
-  sigframe thisframe (mainthread);
 
   fhandler_socket *fh = get (fd);
 
