@@ -1,6 +1,6 @@
 /* pinfo.cc: process table support
 
-   Copyright 1996, 1997, 1998, 2000, 2001, 2002, 2003 Red Hat, Inc.
+   Copyright 1996, 1997, 1998, 2000, 2001, 2002, 2003, 2004 Red Hat, Inc.
 
 This file is part of Cygwin.
 
@@ -12,9 +12,10 @@ details. */
 #include <stdlib.h>
 #include <time.h>
 #include <limits.h>
+#include <stdarg.h>
 #include "security.h"
-#include "fhandler.h"
 #include "path.h"
+#include "fhandler.h"
 #include "dtable.h"
 #include "cygerrno.h"
 #include "sigproc.h"
@@ -27,6 +28,10 @@ details. */
 #include "ntdll.h"
 #include "cygthread.h"
 #include "shared_info.h"
+#include "cygheap.h"
+#include "fhandler.h"
+#include "cygmalloc.h"
+#include "cygtls.h"
 
 static char NO_COPY pinfo_dummy[sizeof (_pinfo)] = {0};
 
@@ -39,13 +44,14 @@ pinfo_fixup_after_fork ()
 {
   if (hexec_proc)
     CloseHandle (hexec_proc);
-
+  /* Keeps the cygpid from being reused. No rights required */
   if (!DuplicateHandle (hMainProc, hMainProc, hMainProc, &hexec_proc, 0,
-			TRUE, DUPLICATE_SAME_ACCESS))
+			TRUE, 0))
     {
       system_printf ("couldn't save current process handle %p, %E", hMainProc);
       hexec_proc = NULL;
     }
+  VerifyHandle (hexec_proc);
 }
 
 /* Initialize the process table.
@@ -89,7 +95,7 @@ pinfo_init (char **envp, int envc)
       myself->pgid = myself->sid = myself->pid;
       myself->ctty = -1;
       myself->uid = ILLEGAL_UID;
-
+      myself->gid = UNKNOWN_GID;
       environ_init (NULL, 0);	/* call after myself has been set up */
     }
 
@@ -99,6 +105,10 @@ pinfo_init (char **envp, int envc)
 void
 _pinfo::exit (UINT n, bool norecord)
 {
+  exit_state = ES_FINAL;
+  cygthread::terminate ();
+  if (norecord)
+    sigproc_terminate ();
   if (this)
     {
       if (!norecord)
@@ -111,8 +121,9 @@ _pinfo::exit (UINT n, bool norecord)
       add_rusage (&rusage_self, &r);
     }
 
-  cygthread::terminate ();
   sigproc_printf ("Calling ExitProcess %d", n);
+  _my_tls.stacklock = 0;
+  _my_tls.stackptr = _my_tls.stack;
   ExitProcess (n);
 }
 
@@ -138,11 +149,13 @@ pinfo::init (pid_t n, DWORD flag, HANDLE in_h)
     }
 
   int createit = flag & (PID_IN_USE | PID_EXECED);
+  DWORD access = FILE_MAP_READ
+		 | (flag & (PID_IN_USE | PID_EXECED | PID_MAP_RW) ? FILE_MAP_WRITE : 0);
   for (int i = 0; i < 10; i++)
     {
       int created;
-      char mapname[MAX_PATH];
-      __small_sprintf (mapname, "cygpid.%x", n);
+      char mapname[CYG_MAX_PATH]; /* XXX Not a path */
+      shared_name (mapname, "cygpid", n);
 
       int mapsize;
       if (flag & PID_EXECED)
@@ -157,12 +170,16 @@ pinfo::init (pid_t n, DWORD flag, HANDLE in_h)
 	}
       else if (!createit)
 	{
-	  h = OpenFileMappingA (FILE_MAP_READ | FILE_MAP_WRITE, FALSE, mapname);
+	  h = OpenFileMappingA (access, FALSE, mapname);
 	  created = 0;
 	}
       else
 	{
-	  h = CreateFileMapping (INVALID_HANDLE_VALUE, &sec_all_nih,
+	  char sa_buf[1024];
+	  PSECURITY_ATTRIBUTES sec_attribs =
+	    sec_user_nih (sa_buf, cygheap->user.sid(), well_known_world_sid,
+			  FILE_MAP_READ);
+	  h = CreateFileMapping (INVALID_HANDLE_VALUE, sec_attribs,
 				 PAGE_READWRITE, 0, mapsize, mapname);
 	  created = h && GetLastError () != ERROR_ALREADY_EXISTS;
 	}
@@ -175,8 +192,7 @@ pinfo::init (pid_t n, DWORD flag, HANDLE in_h)
 	  return;
 	}
 
-      procinfo = (_pinfo *) MapViewOfFileEx (h, FILE_MAP_READ | FILE_MAP_WRITE,
-					     0, 0, 0, mapaddr);
+      procinfo = (_pinfo *) MapViewOfFileEx (h, access, 0, 0, 0, mapaddr);
       ProtectHandle1 (h, pinfo_shared_handle);
 
       if ((procinfo->process_state & PID_INITIALIZING) && (flag & PID_NOREDIR)
@@ -230,6 +246,67 @@ pinfo::init (pid_t n, DWORD flag, HANDLE in_h)
   destroy = 1;
 }
 
+void
+pinfo::set_acl()
+{
+  char sa_buf[1024];
+  SECURITY_DESCRIPTOR sd;
+
+  sec_acl ((PACL) sa_buf, true, true, cygheap->user.sid (),
+	   well_known_world_sid, FILE_MAP_READ);
+  if (!InitializeSecurityDescriptor (&sd, SECURITY_DESCRIPTOR_REVISION))
+    debug_printf ("InitializeSecurityDescriptor %E");
+  else if (!SetSecurityDescriptorDacl (&sd, TRUE, (PACL) sa_buf, FALSE))
+    debug_printf ("SetSecurityDescriptorDacl %E");
+  else if (!SetKernelObjectSecurity (h, DACL_SECURITY_INFORMATION, &sd))
+    debug_printf ("SetKernelObjectSecurity %E");
+}
+
+void
+_pinfo::set_ctty (tty_min *tc, int flags, fhandler_tty_slave *arch)
+{
+  debug_printf ("checking if /dev/tty%d changed", ctty);
+  if ((ctty < 0 || ctty == tc->ntty) && !(flags & O_NOCTTY))
+    {
+      ctty = tc->ntty;
+      syscall_printf ("attached tty%d sid %d, pid %d, tty->pgid %d, tty->sid %d",
+		      tc->ntty, sid, pid, pgid, tc->getsid ());
+
+      pinfo p (tc->getsid ());
+      if (sid == pid && (!p || p->pid == pid || !proc_exists (p)))
+	{
+	  paranoid_printf ("resetting tty%d sid.  Was %d, now %d.  pgid was %d, now %d.",
+			   tc->ntty, tc->getsid (), sid, tc->getpgid (), pgid);
+	  /* We are the session leader */
+	  tc->setsid (sid);
+	  tc->setpgid (pgid);
+	}
+      else
+	sid = tc->getsid ();
+      if (tc->getpgid () == 0)
+	tc->setpgid (pgid);
+      if (cygheap->ctty != arch)
+	{
+	  debug_printf ("cygheap->ctty %p, arch %p", cygheap->ctty, arch);
+	  if (!cygheap->ctty)
+	    syscall_printf ("ctty NULL");
+	  else
+	    {
+	      syscall_printf ("ctty %p, usecount %d", cygheap->ctty,
+			      cygheap->ctty->usecount);
+	      cygheap->ctty->close ();
+	    }
+	  cygheap->ctty = arch;
+	  if (arch)
+	    {
+	      arch->usecount++;
+	      cygheap->open_fhs++;
+	      report_tty_counts (cygheap->ctty, "ctty", "incremented ", "");
+	    }
+	}
+    }
+}
+
 bool
 _pinfo::alive ()
 {
@@ -293,7 +370,7 @@ _pinfo::commune_recv ()
 	const char *argv[__argc_safe + 1];
 	for (int i = 0; i < __argc_safe; i++)
 	  {
-	    if (IsBadStringPtr (__argv[i], 0x7fffffff))
+	    if (IsBadStringPtr (__argv[i], INT32_MAX))
 	      argv[i] = "";
 	    else
 	      argv[i] = __argv[i];
@@ -312,11 +389,34 @@ _pinfo::commune_recv ()
 		sigproc_printf ("WriteFile arg %d failed, %E", a - argv);
 		break;
 	      }
-	  if (!WriteFile (__tothem, "", 1, &nr, NULL))
-	    {
-	      sigproc_printf ("WriteFile null failed, %E");
-	      break;
-	    }
+	if (!WriteFile (__tothem, "", 1, &nr, NULL))
+	  {
+	    sigproc_printf ("WriteFile null failed, %E");
+	    break;
+	  }
+	break;
+      }
+    case PICOM_FIFO:
+      {
+	int formic;
+	if (!ReadFile (__fromthem, &formic, sizeof formic, &nr, NULL)
+	    || nr != sizeof formic)
+	  {
+	    /* __seterrno ();*/	// this is run from the signal thread, so don't set errno
+	    goto out;
+	  }
+
+	fhandler_fifo *fh = cygheap->fdtab.find_fifo ((ATOM) formic);
+	HANDLE it[] = {(fh->get_handle ()), (fh->get_output_handle ())};
+
+	if (!WriteFile (__tothem, it, sizeof (it), &nr, NULL))
+	  {
+	    /*__seterrno ();*/	// this is run from the signal thread, so don't set errno
+	    sigproc_printf ("WriteFile read handle failed, %E");
+	  }
+
+	(void) ReadFile (__fromthem, &nr, sizeof (nr), &nr, NULL);
+	break;
       }
     }
 
@@ -330,17 +430,20 @@ out:
 #define PIPEBUFSIZE (16 * sizeof (DWORD))
 
 commune_result
-_pinfo::commune_send (DWORD code)
+_pinfo::commune_send (DWORD code, ...)
 {
   HANDLE fromthem = NULL, tome = NULL;
   HANDLE fromme = NULL, tothem = NULL;
   DWORD nr;
   commune_result res;
+  va_list args;
+
+  va_start (args, code);
 
   res.s = NULL;
   res.n = 0;
 
-  if (!pid || !this || (dwProcessId != (DWORD) pid && !pinfo (myself->dwProcessId)))
+  if (!this || !pid)
     {
       set_errno (ESRCH);
       goto err;
@@ -367,6 +470,20 @@ _pinfo::commune_send (DWORD code)
       goto err;
     }
 
+  switch (code)
+    {
+    case PICOM_FIFO:
+      {
+	int formic = va_arg (args, int);
+	if (!WriteFile (tothem, &formic, sizeof formic, &nr, NULL) || nr != sizeof formic)
+	  {
+	    __seterrno ();
+	    goto err;
+	  }
+	break;
+      }
+    }
+
   if (sig_send (this, __SIGCOMMUNE))
     goto err;
 
@@ -396,14 +513,14 @@ _pinfo::commune_send (DWORD code)
     }
 
   size_t n;
-  if (!ReadFile (fromthem, &n, sizeof n, &nr, NULL) || nr != sizeof n)
-    {
-      __seterrno ();
-      goto err;
-    }
   switch (code)
     {
     case PICOM_CMDLINE:
+      if (!ReadFile (fromthem, &n, sizeof n, &nr, NULL) || nr != sizeof n)
+	{
+	  __seterrno ();
+	  goto err;
+	}
       res.s = (char *) malloc (n);
       char *p;
       for (p = res.s; ReadFile (fromthem, p, n, &nr, NULL); p += nr)
@@ -415,6 +532,19 @@ _pinfo::commune_send (DWORD code)
 	}
       res.n = n;
       break;
+    case PICOM_FIFO:
+      {
+	DWORD x = ReadFile (fromthem, res.handles, sizeof (res.handles), &nr, NULL);
+	WriteFile (tothem, &x, sizeof (x), &x, NULL);
+	if (!x)
+	  goto err;
+	if (nr != sizeof (res.handles))
+	  {
+	    set_errno (EPIPE);
+	    goto err;
+	  }
+	break;
+      }
     }
   CloseHandle (tothem);
   CloseHandle (fromthem);
@@ -429,7 +559,8 @@ err:
     CloseHandle (tothem);
   if (fromme)
     CloseHandle (fromme);
-  res.n = 0;
+  memset (&res, 0, sizeof (res));
+
 out:
   myself->hello_pid = 0;
   LeaveCriticalSection (&myself->lock);
@@ -471,7 +602,8 @@ pinfo::release ()
   if (h)
     {
 #ifdef DEBUGGING
-      if (((DWORD) procinfo & 0x77000000) == 0x61000000) try_to_debug ();
+      if (((DWORD) procinfo & 0x77000000) == 0x61000000)
+	try_to_debug ();
 #endif
       UnmapViewOfFile (procinfo);
       procinfo = NULL;
@@ -485,12 +617,12 @@ pinfo::release ()
 <sect1 id="func-cygwin-winpid-to-pid">
   <title>cygwin_winpid_to_pid</title>
 
-  <funcsynopsis>
+  <funcsynopsis><funcprototype>
     <funcdef>extern "C" pid_t
       <function>cygwin_winpid_to_pid</function>
       </funcdef>
       <paramdef>int <parameter>winpid</parameter></paramdef>
-  </funcsynopsis>
+  </funcprototype></funcsynopsis>
 
   <para>Given a windows pid, converts to the corresponding Cygwin
 pid, if any.  Returns -1 if windows pid does not correspond to
@@ -535,22 +667,31 @@ winpids::add (DWORD& nelem, bool winpid, DWORD pid)
       pinfolist = (pinfo *) realloc (pinfolist, size_pinfolist (npidlist + 1));
     }
 
-  pinfolist[nelem].init (cygpid, PID_NOREDIR | (winpid ? PID_ALLPIDS : 0));
+  pinfolist[nelem].init (cygpid, PID_NOREDIR | (winpid ? PID_ALLPIDS : 0)
+			 | pinfo_access);
   if (winpid)
-    /* nothing to do */;
-  else if (!pinfolist[nelem])
-    return;
-  else
-    /* Scan list of previously recorded pids to make sure that this pid hasn't
-       shown up before.  This can happen when a process execs. */
-    for (unsigned i = 0; i < nelem; i++)
-      if (pinfolist[i]->pid == pinfolist[nelem]->pid)
-	{
-	  if ((_pinfo *) pinfolist[nelem] != (_pinfo *) myself)
-	    pinfolist[nelem].release ();
-	  return;
-	}
+    goto out;
 
+  if (!pinfolist[nelem])
+    {
+      if (!pinfo_access)
+	return;
+      pinfolist[nelem].init (cygpid, PID_NOREDIR | (winpid ? PID_ALLPIDS : 0));
+      if (!pinfolist[nelem])
+	return;
+      }
+
+  /* Scan list of previously recorded pids to make sure that this pid hasn't
+     shown up before.  This can happen when a process execs. */
+  for (unsigned i = 0; i < nelem; i++)
+    if (pinfolist[i]->pid == pinfolist[nelem]->pid)
+      {
+	if ((_pinfo *) pinfolist[nelem] != (_pinfo *) myself)
+	  pinfolist[nelem].release ();
+	return;
+      }
+
+out:
   pidlist[nelem++] = pid;
 }
 
@@ -621,22 +762,14 @@ winpids::enum9x (bool winpid)
   return nelem;
 }
 
-NO_COPY CRITICAL_SECTION winpids::cs;
-
 void
 winpids::set (bool winpid)
 {
-  EnterCriticalSection (&cs);
+  __malloc_lock ();
   npids = (this->*enum_processes) (winpid);
   if (pidlist)
     pidlist[npids] = 0;
-  LeaveCriticalSection (&cs);
-}
-
-void
-winpids::init ()
-{
-  InitializeCriticalSection (&cs);
+  __malloc_unlock ();
 }
 
 DWORD
