@@ -1,6 +1,6 @@
 /* sec_acl.cc: Sun compatible ACL functions.
 
-   Copyright 2000, 2001, 2002, 2003 Red Hat, Inc.
+   Copyright 2000, 2001, 2002, 2003, 2004 Red Hat, Inc.
 
    Written by Corinna Vinschen <corinna@vinschen.de>
 
@@ -24,8 +24,8 @@ details. */
 #include <winuser.h>
 #include "cygerrno.h"
 #include "security.h"
-#include "fhandler.h"
 #include "path.h"
+#include "fhandler.h"
 #include "dtable.h"
 #include "pinfo.h"
 #include "cygheap.h"
@@ -49,11 +49,9 @@ searchace (__aclent32_t *aclp, int nentries, int type, __uid32_t id = ILLEGAL_UI
 static int
 setacl (const char *file, int nentries, __aclent32_t *aclbufp)
 {
-  DWORD sd_size = 4096;
-  char sd_buf[4096];
-  PSECURITY_DESCRIPTOR psd = (PSECURITY_DESCRIPTOR) sd_buf;
+  security_descriptor sd_ret;
 
-  if (read_sd (file, psd, &sd_size) <= 0)
+  if (read_sd (file, sd_ret) <= 0)
     {
       debug_printf ("read_sd %E");
       return -1;
@@ -63,7 +61,7 @@ setacl (const char *file, int nentries, __aclent32_t *aclbufp)
 
   /* Get owner SID. */
   PSID owner_sid;
-  if (!GetSecurityDescriptorOwner (psd, &owner_sid, &dummy))
+  if (!GetSecurityDescriptorOwner (sd_ret, &owner_sid, &dummy))
     {
       __seterrno ();
       return -1;
@@ -72,7 +70,7 @@ setacl (const char *file, int nentries, __aclent32_t *aclbufp)
 
   /* Get group SID. */
   PSID group_sid;
-  if (!GetSecurityDescriptorGroup (psd, &group_sid, &dummy))
+  if (!GetSecurityDescriptorGroup (sd_ret, &group_sid, &dummy))
     {
       __seterrno ();
       return -1;
@@ -206,21 +204,26 @@ setacl (const char *file, int nentries, __aclent32_t *aclbufp)
       __seterrno ();
       return -1;
     }
-  /* Make self relative security descriptor in psd. */
-  sd_size = 0;
-  MakeSelfRelativeSD (&sd, psd, &sd_size);
+  /* Make self relative security descriptor in sd_ret. */
+  DWORD sd_size = 0;
+  MakeSelfRelativeSD (&sd, sd_ret, &sd_size);
   if (sd_size <= 0)
     {
       __seterrno ();
       return -1;
     }
-  if (!MakeSelfRelativeSD (&sd, psd, &sd_size))
+  if (!sd_ret.realloc (sd_size))
+    {
+      set_errno (ENOMEM);
+      return -1;
+    }
+  if (!MakeSelfRelativeSD (&sd, sd_ret, &sd_size))
     {
       __seterrno ();
       return -1;
     }
-  debug_printf ("Created SD-Size: %d", sd_size);
-  return write_sd (file, psd, sd_size);
+  debug_printf ("Created SD-Size: %d", sd_ret.size ());
+  return write_sd (file, sd_ret);
 }
 
 /* Temporary access denied bits */
@@ -235,19 +238,19 @@ getace (__aclent32_t &acl, int type, int id, DWORD win_ace_mask,
   acl.a_type = type;
   acl.a_id = id;
 
-  if ((win_ace_mask & FILE_READ_DATA) && !(acl.a_perm & (S_IROTH | DENY_R)))
+  if ((win_ace_mask & FILE_READ_BITS) && !(acl.a_perm & (S_IROTH | DENY_R)))
     if (win_ace_type == ACCESS_ALLOWED_ACE_TYPE)
       acl.a_perm |= S_IROTH;
     else if (win_ace_type == ACCESS_DENIED_ACE_TYPE)
       acl.a_perm |= DENY_R;
 
-  if ((win_ace_mask & FILE_WRITE_DATA) && !(acl.a_perm & (S_IWOTH | DENY_W)))
+  if ((win_ace_mask & FILE_WRITE_BITS) && !(acl.a_perm & (S_IWOTH | DENY_W)))
     if (win_ace_type == ACCESS_ALLOWED_ACE_TYPE)
       acl.a_perm |= S_IWOTH;
     else if (win_ace_type == ACCESS_DENIED_ACE_TYPE)
       acl.a_perm |= DENY_W;
 
-  if ((win_ace_mask & FILE_EXECUTE) && !(acl.a_perm & (S_IXOTH | DENY_X)))
+  if ((win_ace_mask & FILE_EXEC_BITS) && !(acl.a_perm & (S_IXOTH | DENY_X)))
     if (win_ace_type == ACCESS_ALLOWED_ACE_TYPE)
       acl.a_perm |= S_IXOTH;
     else if (win_ace_type == ACCESS_DENIED_ACE_TYPE)
@@ -257,12 +260,10 @@ getace (__aclent32_t &acl, int type, int id, DWORD win_ace_mask,
 static int
 getacl (const char *file, DWORD attr, int nentries, __aclent32_t *aclbufp)
 {
-  DWORD sd_size = 4096;
-  char sd_buf[4096];
-  PSECURITY_DESCRIPTOR psd = (PSECURITY_DESCRIPTOR) sd_buf;
+  security_descriptor sd;
 
   int ret;
-  if ((ret = read_sd (file, psd, &sd_size)) <= 0)
+  if ((ret = read_sd (file, sd)) <= 0)
     {
       debug_printf ("read_sd %E");
       return ret;
@@ -274,7 +275,7 @@ getacl (const char *file, DWORD attr, int nentries, __aclent32_t *aclbufp)
   __uid32_t uid;
   __gid32_t gid;
 
-  if (!GetSecurityDescriptorOwner (psd, (PSID *) &owner_sid, &dummy))
+  if (!GetSecurityDescriptorOwner (sd, (PSID *) &owner_sid, &dummy))
     {
       debug_printf ("GetSecurityDescriptorOwner %E");
       __seterrno ();
@@ -282,7 +283,7 @@ getacl (const char *file, DWORD attr, int nentries, __aclent32_t *aclbufp)
     }
   uid = owner_sid.get_uid ();
 
-  if (!GetSecurityDescriptorGroup (psd, (PSID *) &group_sid, &dummy))
+  if (!GetSecurityDescriptorGroup (sd, (PSID *) &group_sid, &dummy))
     {
       debug_printf ("GetSecurityDescriptorGroup %E");
       __seterrno ();
@@ -305,7 +306,7 @@ getacl (const char *file, DWORD attr, int nentries, __aclent32_t *aclbufp)
   PACL acl;
   BOOL acl_exists;
 
-  if (!GetSecurityDescriptorDacl (psd, &acl_exists, &acl, &dummy))
+  if (!GetSecurityDescriptorDacl (sd, &acl_exists, &acl, &dummy))
     {
       __seterrno ();
       debug_printf ("GetSecurityDescriptorDacl %E");
@@ -356,7 +357,7 @@ getacl (const char *file, DWORD attr, int nentries, __aclent32_t *aclbufp)
 	      id = ILLEGAL_GID;
 	    }
 	  else
-	    id = ace_sid.get_id (TRUE, &type);
+	    id = ace_sid.get_id (true, &type);
 
 	  if (!type)
 	    continue;
@@ -520,16 +521,16 @@ facl32 (int fd, int cmd, int nentries, __aclent32_t *aclbufp)
 extern "C" int
 aclcheck32 (__aclent32_t *aclbufp, int nentries, int *which)
 {
-  BOOL has_user_obj = FALSE;
-  BOOL has_group_obj = FALSE;
-  BOOL has_other_obj = FALSE;
-  BOOL has_class_obj = FALSE;
-  BOOL has_ug_objs = FALSE;
-  BOOL has_def_user_obj = FALSE;
-  BOOL has_def_group_obj = FALSE;
-  BOOL has_def_other_obj = FALSE;
-  BOOL has_def_class_obj = FALSE;
-  BOOL has_def_ug_objs = FALSE;
+  bool has_user_obj = false;
+  bool has_group_obj = false;
+  bool has_other_obj = false;
+  bool has_class_obj = false;
+  bool has_ug_objs = false;
+  bool has_def_user_obj = false;
+  bool has_def_group_obj = false;
+  bool has_def_other_obj = false;
+  bool has_def_class_obj = false;
+  bool has_def_ug_objs = false;
   int pos2;
 
   for (int pos = 0; pos < nentries; ++pos)
@@ -542,7 +543,7 @@ aclcheck32 (__aclent32_t *aclbufp, int nentries, int *which)
 	      *which = pos;
 	    return USER_ERROR;
 	  }
-	has_user_obj = TRUE;
+	has_user_obj = true;
 	break;
       case GROUP_OBJ:
 	if (has_group_obj)
@@ -551,7 +552,7 @@ aclcheck32 (__aclent32_t *aclbufp, int nentries, int *which)
 	      *which = pos;
 	    return GRP_ERROR;
 	  }
-	has_group_obj = TRUE;
+	has_group_obj = true;
 	break;
       case OTHER_OBJ:
 	if (has_other_obj)
@@ -560,7 +561,7 @@ aclcheck32 (__aclent32_t *aclbufp, int nentries, int *which)
 	      *which = pos;
 	    return OTHER_ERROR;
 	  }
-	has_other_obj = TRUE;
+	has_other_obj = true;
 	break;
       case CLASS_OBJ:
 	if (has_class_obj)
@@ -569,7 +570,7 @@ aclcheck32 (__aclent32_t *aclbufp, int nentries, int *which)
 	      *which = pos;
 	    return CLASS_ERROR;
 	  }
-	has_class_obj = TRUE;
+	has_class_obj = true;
 	break;
       case USER:
       case GROUP:
@@ -580,7 +581,7 @@ aclcheck32 (__aclent32_t *aclbufp, int nentries, int *which)
 	      *which = pos2;
 	    return DUPLICATE_ERROR;
 	  }
-	has_ug_objs = TRUE;
+	has_ug_objs = true;
 	break;
       case DEF_USER_OBJ:
 	if (has_def_user_obj)
@@ -589,7 +590,7 @@ aclcheck32 (__aclent32_t *aclbufp, int nentries, int *which)
 	      *which = pos;
 	    return USER_ERROR;
 	  }
-	has_def_user_obj = TRUE;
+	has_def_user_obj = true;
 	break;
       case DEF_GROUP_OBJ:
 	if (has_def_group_obj)
@@ -598,7 +599,7 @@ aclcheck32 (__aclent32_t *aclbufp, int nentries, int *which)
 	      *which = pos;
 	    return GRP_ERROR;
 	  }
-	has_def_group_obj = TRUE;
+	has_def_group_obj = true;
 	break;
       case DEF_OTHER_OBJ:
 	if (has_def_other_obj)
@@ -607,7 +608,7 @@ aclcheck32 (__aclent32_t *aclbufp, int nentries, int *which)
 	      *which = pos;
 	    return OTHER_ERROR;
 	  }
-	has_def_other_obj = TRUE;
+	has_def_other_obj = true;
 	break;
       case DEF_CLASS_OBJ:
 	if (has_def_class_obj)
@@ -616,7 +617,7 @@ aclcheck32 (__aclent32_t *aclbufp, int nentries, int *which)
 	      *which = pos;
 	    return CLASS_ERROR;
 	  }
-	has_def_class_obj = TRUE;
+	has_def_class_obj = true;
 	break;
       case DEF_USER:
       case DEF_GROUP:
@@ -627,7 +628,7 @@ aclcheck32 (__aclent32_t *aclbufp, int nentries, int *which)
 	      *which = pos2;
 	    return DUPLICATE_ERROR;
 	  }
-	has_def_ug_objs = TRUE;
+	has_def_ug_objs = true;
 	break;
       default:
 	return ENTRY_ERROR;
@@ -785,13 +786,13 @@ acltotext32 (__aclent32_t *aclbufp, int aclcnt)
     }
   char buf[32000];
   buf[0] = '\0';
-  BOOL first = TRUE;
+  bool first = true;
 
   for (int pos = 0; pos < aclcnt; ++pos)
     {
       if (!first)
 	strcat (buf, ",");
-      first = FALSE;
+      first = false;
       if (aclbufp[pos].a_type & ACL_DEFAULT)
 	strcat (buf, "default");
       switch (aclbufp[pos].a_type)
