@@ -1,6 +1,6 @@
 /* security.cc: NT security functions
 
-   Copyright 1997, 1998, 1999, 2000, 2001, 2002, 2003 Red Hat, Inc.
+   Copyright 1997, 1998, 1999, 2000, 2001, 2002, 2003, 2004 Red Hat, Inc.
 
    Originaly written by Gunther Ebert, gunther.ebert@ixos-leipzig.de
    Completely rewritten by Corinna Vinschen <corinna@vinschen.de>
@@ -30,8 +30,8 @@ details. */
 #include <aclapi.h>
 #include "cygerrno.h"
 #include "security.h"
-#include "fhandler.h"
 #include "path.h"
+#include "fhandler.h"
 #include "dtable.h"
 #include "pinfo.h"
 #include "cygheap.h"
@@ -261,7 +261,7 @@ get_lsa_srv_inf (LSA_HANDLE lsa, char *logonserver, char *domain)
 }
 #endif
 
-BOOL
+bool
 get_logon_server (const char *domain, char *server, WCHAR *wserver)
 {
   WCHAR wdomain[INTERNET_MAX_HOST_NAME_LENGTH + 1];
@@ -276,7 +276,7 @@ get_logon_server (const char *domain, char *server, WCHAR *wserver)
       server[0] = server[1] = '\\';
       if (wserver)
 	sys_mbstowcs (wserver, server, INTERNET_MAX_HOST_NAME_LENGTH + 1);
-      return TRUE;
+      return true;
     }
 
   /* Try to get the primary domain controller for the domain */
@@ -288,13 +288,13 @@ get_logon_server (const char *domain, char *server, WCHAR *wserver)
 	for (WCHAR *ptr1 = buf; (*wserver++ = *ptr1++);)
 	  ;
       NetApiBufferFree (buf);
-      return TRUE;
+      return true;
     }
   __seterrno_from_win_error (ret);
-  return FALSE;
+  return false;
 }
 
-static BOOL
+static bool
 get_user_groups (WCHAR *wlogonserver, cygsidlist &grp_list, char *user,
 		 char *domain)
 {
@@ -338,36 +338,36 @@ get_user_groups (WCHAR *wlogonserver, cygsidlist &grp_list, char *user,
     }
 
   NetApiBufferFree (buf);
-  return TRUE;
+  return true;
 }
 
-static BOOL
+static bool
 is_group_member (WCHAR *wgroup, PSID pusersid, cygsidlist &grp_list)
 {
   LPLOCALGROUP_MEMBERS_INFO_0 buf;
   DWORD cnt, tot;
   NET_API_STATUS ret;
-  BOOL retval = FALSE;
+  bool retval = false;
 
   /* Members can be users or global groups */
   ret = NetLocalGroupGetMembers (NULL, wgroup, 0, (LPBYTE *) &buf,
 				 MAX_PREFERRED_LENGTH, &cnt, &tot, NULL);
   if (ret)
-    return FALSE;
+    return false;
 
   for (DWORD bidx = 0; !retval && bidx < cnt; ++bidx)
     if (EqualSid (pusersid, buf[bidx].lgrmi0_sid))
-      retval = TRUE;
+      retval = true;
     else
       for (int glidx = 0; !retval && glidx < grp_list.count; ++glidx)
 	if (EqualSid (grp_list.sids[glidx], buf[bidx].lgrmi0_sid))
-	  retval = TRUE;
+	  retval = true;
 
   NetApiBufferFree (buf);
   return retval;
 }
 
-static BOOL
+static bool
 get_user_local_groups (cygsidlist &grp_list, PSID pusersid)
 {
   LPLOCALGROUP_INFO_0 buf;
@@ -379,7 +379,7 @@ get_user_local_groups (cygsidlist &grp_list, PSID pusersid)
   if (ret)
     {
       __seterrno_from_win_error (ret);
-      return FALSE;
+      return false;
     }
 
   char bgroup[INTERNET_MAX_HOST_NAME_LENGTH + GNLEN + 2];
@@ -392,7 +392,7 @@ get_user_local_groups (cygsidlist &grp_list, PSID pusersid)
       || !GetComputerNameA (lgroup, &(llen = INTERNET_MAX_HOST_NAME_LENGTH + 1)))
     {
       __seterrno ();
-      return FALSE;
+      return false;
     }
   bgroup[blen++] = lgroup[llen++] = '\\';
 
@@ -421,18 +421,18 @@ get_user_local_groups (cygsidlist &grp_list, PSID pusersid)
 	  grp_list += gsid;
       }
   NetApiBufferFree (buf);
-  return TRUE;
+  return true;
 }
 
-static BOOL
-sid_in_token_groups (PTOKEN_GROUPS grps, cygsid &sid)
+static bool
+sid_in_token_groups (PTOKEN_GROUPS grps, cygpsid sid)
 {
   if (!grps)
-    return FALSE;
+    return false;
   for (DWORD i = 0; i < grps->GroupCount; ++i)
     if (sid == grps->Groups[i].Sid)
-      return TRUE;
-  return FALSE;
+      return true;
+  return false;
 }
 
 #if 0				/* Unused */
@@ -529,11 +529,11 @@ get_token_group_sidlist (cygsidlist &grp_list, PTOKEN_GROUPS my_grps,
     }
 }
 
-static BOOL
+static bool
 get_initgroups_sidlist (cygsidlist &grp_list,
 			PSID usersid, PSID pgrpsid, struct passwd *pw,
 			PTOKEN_GROUPS my_grps, LUID auth_luid, int &auth_pos,
-			BOOL &special_pgrp)
+			bool &special_pgrp)
 {
   grp_list += well_known_world_sid;
   grp_list += well_known_authenticated_users_sid;
@@ -556,12 +556,12 @@ get_initgroups_sidlist (cygsidlist &grp_list,
 	get_user_groups (wserver, grp_list, user, domain);
       get_unix_group_sidlist (pw, grp_list);
       if (!get_user_local_groups (grp_list, usersid))
-	return FALSE;
+	return false;
     }
   /* special_pgrp true if pgrpsid is not in normal groups */
   if ((special_pgrp = !grp_list.contains (pgrpsid)))
     grp_list += pgrpsid;
-  return TRUE;
+  return true;
 }
 
 static void
@@ -704,11 +704,11 @@ get_priv_list (LSA_HANDLE lsa, cygsid &usersid, cygsidlist &grp_list)
 	except if the token is internal and the group is in the token SD
 	(see create_token). In that latter case that group must match the
 	requested primary group.  */
-BOOL
-verify_token (HANDLE token, cygsid &usersid, user_groups &groups, BOOL *pintern)
+bool
+verify_token (HANDLE token, cygsid &usersid, user_groups &groups, bool *pintern)
 {
   DWORD size;
-  BOOL intern = FALSE;
+  bool intern = false;
 
   if (pintern)
     {
@@ -725,7 +725,7 @@ verify_token (HANDLE token, cygsid &usersid, user_groups &groups, BOOL *pintern)
 			    &tok_usersid, sizeof tok_usersid, &size))
     debug_printf ("GetTokenInformation(): %E");
   if (usersid != tok_usersid)
-    return FALSE;
+    return false;
 
   /* For an internal token, if setgroups was not called and if the sd group
      is not well_known_null_sid, it must match pgrpsid */
@@ -805,18 +805,17 @@ create_token (cygsid &usersid, user_groups &new_groups, struct passwd *pw)
     { sizeof sqos, SecurityImpersonation, SECURITY_STATIC_TRACKING, FALSE };
   OBJECT_ATTRIBUTES oa = { sizeof oa, 0, 0, 0, 0, &sqos };
   PSECURITY_ATTRIBUTES psa;
-  BOOL special_pgrp = FALSE;
+  bool special_pgrp = false;
   char sa_buf[1024];
   LUID auth_luid = SYSTEM_LUID;
-  LARGE_INTEGER exp = { QuadPart:0x7fffffffffffffffLL };
+  LARGE_INTEGER exp = { QuadPart:INT64_MAX };
 
   TOKEN_USER user;
   PTOKEN_GROUPS new_tok_gsids = NULL;
   PTOKEN_PRIVILEGES privs = NULL;
   TOKEN_OWNER owner;
   TOKEN_PRIMARY_GROUP pgrp;
-  char acl_buf[MAX_DACL_LEN (5)];
-  TOKEN_DEFAULT_DACL dacl;
+  TOKEN_DEFAULT_DACL dacl = {};
   TOKEN_SOURCE source;
   TOKEN_STATISTICS stats;
   memcpy (source.SourceName, "Cygwin.1", 8);
@@ -905,13 +904,6 @@ create_token (cygsid &usersid, user_groups &new_groups, struct passwd *pw)
   if (!(privs = get_priv_list (lsa, usersid, tmp_gsids)))
     goto out;
 
-  /* Create default dacl. */
-  if (!sec_acl ((PACL) acl_buf, FALSE,
-		tmp_gsids.contains (well_known_admins_sid) ?
-		well_known_admins_sid : usersid))
-    goto out;
-  dacl.DefaultDacl = (PACL) acl_buf;
-
   /* Let's be heroic... */
   ret = NtCreateToken (&token, TOKEN_ALL_ACCESS, &oa, TokenImpersonation,
 		       &auth_luid, &exp, &user, new_tok_gsids, privs, &owner,
@@ -926,7 +918,7 @@ create_token (cygsid &usersid, user_groups &new_groups, struct passwd *pw)
   else
     {
       /* Set security descriptor and primary group */
-      psa = __sec_user (sa_buf, usersid, TRUE);
+      psa = sec_user (sa_buf, usersid);
       if (psa->lpSecurityDescriptor &&
 	  !SetSecurityDescriptorGroup ((PSECURITY_DESCRIPTOR)
 				       psa->lpSecurityDescriptor,
@@ -1075,23 +1067,18 @@ out:
    of the SD on success. Unfortunately NT returns
    0 in `len' on success, while W2K returns the
    correct size!
+   
+   2003-11-26: Now the function allocates the space needed by itself so
+   it knows the real size and returns it in the security_descriptor object.
 */
 
 LONG
-read_sd (const char *file, PSECURITY_DESCRIPTOR sd_buf, LPDWORD sd_size)
+read_sd (const char *file, security_descriptor &sd)
 {
-  /* Check parameters */
-  if (!sd_size)
-    {
-      set_errno (EINVAL);
-      return -1;
-    }
-
-  debug_printf ("file = %s", file);
 
   DWORD len = 0;
   const char *pfile = file;
-  char fbuf[PATH_MAX];
+  char fbuf[CYG_MAX_PATH];
   if (current_codepage == oem_cp)
     {
       DWORD fname_len = min (sizeof (fbuf) - 1, strlen (file));
@@ -1104,34 +1091,38 @@ read_sd (const char *file, PSECURITY_DESCRIPTOR sd_buf, LPDWORD sd_size)
 			OWNER_SECURITY_INFORMATION
 			| GROUP_SECURITY_INFORMATION
 			| DACL_SECURITY_INFORMATION,
-			sd_buf, *sd_size, &len))
+			NULL, 0, &len)
+      && GetLastError () != ERROR_INSUFFICIENT_BUFFER)
     {
+      debug_printf ("file = %s", file);
       __seterrno ();
       return -1;
     }
   debug_printf ("file = %s: len=%d", file, len);
-  if (len > *sd_size)
+  if (!sd.malloc (len))
     {
-      *sd_size = len;
-      return 0;
+      set_errno (ENOMEM);
+      return -1;
     }
-  return 1;
+  if (!GetFileSecurity (pfile,
+			OWNER_SECURITY_INFORMATION
+			| GROUP_SECURITY_INFORMATION
+			| DACL_SECURITY_INFORMATION,
+			sd, len, &len))
+    {
+      __seterrno ();
+      return -1;
+    }
+  return sd.size ();
 }
 
 LONG
-write_sd (const char *file, PSECURITY_DESCRIPTOR sd_buf, DWORD sd_size)
+write_sd (const char *file, security_descriptor &sd)
 {
-  /* Check parameters */
-  if (!sd_buf || !sd_size)
-    {
-      set_errno (EINVAL);
-      return -1;
-    }
-
   BOOL dummy;
   cygpsid owner;
 
-  if (!GetSecurityDescriptorOwner (sd_buf, (PSID *) &owner, &dummy))
+  if (!GetSecurityDescriptorOwner (sd, (PSID *) &owner, &dummy))
     {
       __seterrno ();
       return -1;
@@ -1176,7 +1167,7 @@ write_sd (const char *file, PSECURITY_DESCRIPTOR sd_buf, DWORD sd_size)
   header.dwStreamId = BACKUP_SECURITY_DATA;
   header.dwStreamAttributes = STREAM_CONTAINS_SECURITY;
   header.Size.HighPart = 0;
-  header.Size.LowPart = sd_size;
+  header.Size.LowPart = sd.size ();
   header.dwStreamNameSize = 0;
   if (!BackupWrite (fh, (LPBYTE) &header,
 		    3 * sizeof (DWORD) + sizeof (LARGE_INTEGER),
@@ -1188,7 +1179,7 @@ write_sd (const char *file, PSECURITY_DESCRIPTOR sd_buf, DWORD sd_size)
     }
 
   /* write new security descriptor */
-  if (!BackupWrite (fh, (LPBYTE) sd_buf,
+  if (!BackupWrite (fh, (LPBYTE) (PSECURITY_DESCRIPTOR) sd,
 		    header.Size.LowPart + header.dwStreamNameSize,
 		    &bytes_written, FALSE, TRUE, &context))
     {
@@ -1213,7 +1204,7 @@ write_sd (const char *file, PSECURITY_DESCRIPTOR sd_buf, DWORD sd_size)
 
 static void
 get_attribute_from_acl (mode_t *attribute, PACL acl, PSID owner_sid,
-			PSID group_sid, BOOL grp_member)
+			PSID group_sid, bool grp_member)
 {
   ACCESS_ALLOWED_ACE *ace;
   int allow = 0;
@@ -1243,15 +1234,15 @@ get_attribute_from_acl (mode_t *attribute, PACL acl, PSID owner_sid,
       cygpsid ace_sid ((PSID) &ace->SidStart);
       if (ace_sid == well_known_world_sid)
 	{
-	  if (ace->Mask & FILE_READ_DATA)
+	  if (ace->Mask & FILE_READ_BITS)
 	    *flags |= ((!(*anti & S_IROTH)) ? S_IROTH : 0)
 		      | ((!(*anti & S_IRGRP)) ? S_IRGRP : 0)
 		      | ((!(*anti & S_IRUSR)) ? S_IRUSR : 0);
-	  if (ace->Mask & FILE_WRITE_DATA)
+	  if (ace->Mask & FILE_WRITE_BITS)
 	    *flags |= ((!(*anti & S_IWOTH)) ? S_IWOTH : 0)
 		      | ((!(*anti & S_IWGRP)) ? S_IWGRP : 0)
 		      | ((!(*anti & S_IWUSR)) ? S_IWUSR : 0);
-	  if (ace->Mask & FILE_EXECUTE)
+	  if (ace->Mask & FILE_EXEC_BITS)
 	    *flags |= ((!(*anti & S_IXOTH)) ? S_IXOTH : 0)
 		      | ((!(*anti & S_IXGRP)) ? S_IXGRP : 0)
 		      | ((!(*anti & S_IXUSR)) ? S_IXUSR : 0);
@@ -1272,22 +1263,22 @@ get_attribute_from_acl (mode_t *attribute, PACL acl, PSID owner_sid,
 	}
       else if (ace_sid == owner_sid)
 	{
-	  if (ace->Mask & FILE_READ_DATA)
+	  if (ace->Mask & FILE_READ_BITS)
 	    *flags |= ((!(*anti & S_IRUSR)) ? S_IRUSR : 0);
-	  if (ace->Mask & FILE_WRITE_DATA)
+	  if (ace->Mask & FILE_WRITE_BITS)
 	    *flags |= ((!(*anti & S_IWUSR)) ? S_IWUSR : 0);
-	  if (ace->Mask & FILE_EXECUTE)
+	  if (ace->Mask & FILE_EXEC_BITS)
 	    *flags |= ((!(*anti & S_IXUSR)) ? S_IXUSR : 0);
 	}
       else if (ace_sid == group_sid)
 	{
-	  if (ace->Mask & FILE_READ_DATA)
+	  if (ace->Mask & FILE_READ_BITS)
 	    *flags |= ((!(*anti & S_IRGRP)) ? S_IRGRP : 0)
 		      | ((grp_member && !(*anti & S_IRUSR)) ? S_IRUSR : 0);
-	  if (ace->Mask & FILE_WRITE_DATA)
+	  if (ace->Mask & FILE_WRITE_BITS)
 	    *flags |= ((!(*anti & S_IWGRP)) ? S_IWGRP : 0)
 		      | ((grp_member && !(*anti & S_IWUSR)) ? S_IWUSR : 0);
-	  if (ace->Mask & FILE_EXECUTE)
+	  if (ace->Mask & FILE_EXEC_BITS)
 	    *flags |= ((!(*anti & S_IXGRP)) ? S_IXGRP : 0)
 		      | ((grp_member && !(*anti & S_IXUSR)) ? S_IXUSR : 0);
 	}
@@ -1334,7 +1325,7 @@ get_info_from_sd (PSECURITY_DESCRIPTOR psd, mode_t *attribute,
 
   __uid32_t uid;
   __gid32_t gid;
-  BOOL grp_member = get_sids_info (owner_sid, group_sid, &uid, &gid);
+  bool grp_member = get_sids_info (owner_sid, group_sid, &uid, &gid);
   if (uidret)
     *uidret = uid;
   if (gidret)
@@ -1369,19 +1360,11 @@ static void
 get_nt_attribute (const char *file, mode_t *attribute,
 		  __uid32_t *uidret, __gid32_t *gidret)
 {
-  /* Yeah, sounds too much, but I've seen SDs of 2100 bytes! */
-  DWORD sd_size = 4096;
-  char sd_buf[4096];
-  PSECURITY_DESCRIPTOR psd = (PSECURITY_DESCRIPTOR) sd_buf;
+  security_descriptor sd;
 
-  if (read_sd (file, psd, &sd_size) <= 0)
-    {
-      debug_printf ("read_sd %E");
-      psd = NULL;
-    }
-
-  get_info_from_sd (psd, attribute, uidret, gidret);
-  return;
+  if (read_sd (file, sd) <= 0)
+    debug_printf ("read_sd %E");
+  get_info_from_sd (sd, attribute, uidret, gidret);
 }
 
 int
@@ -1419,46 +1402,46 @@ get_file_attribute (int use_ntsec, const char *file,
 
 static void
 get_nt_object_attribute (HANDLE handle, SE_OBJECT_TYPE object_type,
-			 mode_t *attribute, __uid32_t *uidret, __gid32_t *gidret)
+			 mode_t *attribute, __uid32_t *uidret,
+			 __gid32_t *gidret)
 {
-  PSECURITY_DESCRIPTOR psd;
-  char sd_buf[4096];
+  security_descriptor sd;
+  PSECURITY_DESCRIPTOR psd = NULL;
+  LONG ret;
 
   if (object_type == SE_REGISTRY_KEY)
     {
       /* use different code for registry handles, for performance reasons */
-      psd = (PSECURITY_DESCRIPTOR) & sd_buf[0];
-      DWORD len = sizeof (sd_buf);
-      if (ERROR_SUCCESS != RegGetKeySecurity ((HKEY) handle,
-					      DACL_SECURITY_INFORMATION |
-					      GROUP_SECURITY_INFORMATION |
-					      OWNER_SECURITY_INFORMATION,
-					      psd, &len))
-	{
-	  __seterrno ();
-	  debug_printf ("RegGetKeySecurity %E");
-	  psd = NULL;
-	}
-      }
+      DWORD len = 0;
+      if ((ret = RegGetKeySecurity ((HKEY) handle,
+				    DACL_SECURITY_INFORMATION
+				    | GROUP_SECURITY_INFORMATION
+				    | OWNER_SECURITY_INFORMATION,
+				    sd, &len)) != ERROR_INSUFFICIENT_BUFFER)
+	__seterrno_from_win_error (ret);
+      else if (!sd.malloc (len))
+	set_errno (ENOMEM);
+      else if ((ret = RegGetKeySecurity ((HKEY) handle,
+					 DACL_SECURITY_INFORMATION
+					 | GROUP_SECURITY_INFORMATION
+					 | OWNER_SECURITY_INFORMATION,
+					 sd, &len)) != ERROR_SUCCESS)
+	__seterrno_from_win_error (ret);
+      else
+        psd = sd;
+      get_info_from_sd (psd, attribute, uidret, gidret);
+    }
+  else if ((ret = GetSecurityInfo (handle, object_type,
+				   DACL_SECURITY_INFORMATION
+				   | GROUP_SECURITY_INFORMATION
+				   | OWNER_SECURITY_INFORMATION,
+				   NULL, NULL, NULL, NULL, &psd)))
+    __seterrno_from_win_error (ret);
   else
     {
-      if (ERROR_SUCCESS != GetSecurityInfo (handle, object_type,
-					    DACL_SECURITY_INFORMATION |
-					    GROUP_SECURITY_INFORMATION |
-					    OWNER_SECURITY_INFORMATION,
-					    NULL, NULL, NULL, NULL, &psd))
-	{
-	  __seterrno ();
-	  debug_printf ("GetSecurityInfo %E");
-	  psd = NULL;
-	}
+      get_info_from_sd (psd, attribute, uidret, gidret);
+      LocalFree (psd);
     }
-
-  get_info_from_sd (psd, attribute, uidret, gidret);
-  if (psd != (PSECURITY_DESCRIPTOR) & sd_buf[0])
-    LocalFree (psd);
-
-  return;
 }
 
 int
@@ -1474,50 +1457,45 @@ get_object_attribute (HANDLE handle, SE_OBJECT_TYPE object_type,
   return -1;
 }
 
-BOOL
+bool
 add_access_allowed_ace (PACL acl, int offset, DWORD attributes,
 			PSID sid, size_t &len_add, DWORD inherit)
 {
   if (!AddAccessAllowedAce (acl, ACL_REVISION, attributes, sid))
     {
       __seterrno ();
-      return FALSE;
+      return false;
     }
   ACCESS_ALLOWED_ACE *ace;
   if (inherit && GetAce (acl, offset, (PVOID *) &ace))
     ace->Header.AceFlags |= inherit;
   len_add += sizeof (ACCESS_ALLOWED_ACE) - sizeof (DWORD) + GetLengthSid (sid);
-  return TRUE;
+  return true;
 }
 
-BOOL
+bool
 add_access_denied_ace (PACL acl, int offset, DWORD attributes,
 		       PSID sid, size_t &len_add, DWORD inherit)
 {
   if (!AddAccessDeniedAce (acl, ACL_REVISION, attributes, sid))
     {
       __seterrno ();
-      return FALSE;
+      return false;
     }
   ACCESS_DENIED_ACE *ace;
   if (inherit && GetAce (acl, offset, (PVOID *) &ace))
     ace->Header.AceFlags |= inherit;
   len_add += sizeof (ACCESS_DENIED_ACE) - sizeof (DWORD) + GetLengthSid (sid);
-  return TRUE;
+  return true;
 }
 
-PSECURITY_DESCRIPTOR
+static PSECURITY_DESCRIPTOR
 alloc_sd (__uid32_t uid, __gid32_t gid, int attribute,
-	  PSECURITY_DESCRIPTOR sd_ret, DWORD *sd_size_ret)
+	  security_descriptor &sd_ret)
 {
   BOOL dummy;
 
   debug_printf("uid %d, gid %d, attribute %x", uid, gid, attribute);
-  if (!sd_ret || !sd_size_ret)
-    {
-      set_errno (EINVAL);
-      return NULL;
-    }
 
   /* Get owner and group from current security descriptor. */
   PSID cur_owner_sid = NULL;
@@ -1654,7 +1632,7 @@ alloc_sd (__uid32_t uid, __gid32_t gid, int attribute,
 
   /* Add owner and group permissions if SIDs are equal
      and construct deny attributes for group and owner. */
-  BOOL isownergroup;
+  bool isownergroup;
   if ((isownergroup = (owner_sid == group_sid)))
     owner_allow |= group_allow;
 
@@ -1745,7 +1723,7 @@ alloc_sd (__uid32_t uid, __gid32_t gid, int attribute,
 	}
 
   /* Construct appropriate inherit attribute for new directories */
-  if (S_ISDIR (attribute) && !acl_exists )
+  if (S_ISDIR (attribute) && !acl_exists)
     {
       const DWORD inherit = SUB_CONTAINERS_AND_OBJECTS_INHERIT | INHERIT_ONLY;
 
@@ -1796,33 +1774,37 @@ alloc_sd (__uid32_t uid, __gid32_t gid, int attribute,
     }
 
   /* Make self relative security descriptor. */
-  *sd_size_ret = 0;
-  MakeSelfRelativeSD (&sd, sd_ret, sd_size_ret);
-  if (*sd_size_ret <= 0)
+  DWORD sd_size = 0;
+  MakeSelfRelativeSD (&sd, sd_ret, &sd_size);
+  if (sd_size <= 0)
     {
       __seterrno ();
       return NULL;
     }
-  if (!MakeSelfRelativeSD (&sd, sd_ret, sd_size_ret))
+  if (!sd_ret.malloc (sd_size))
+    {
+      set_errno (ENOMEM);
+      return NULL;
+    }
+  if (!MakeSelfRelativeSD (&sd, sd_ret, &sd_size))
     {
       __seterrno ();
       return NULL;
     }
-  debug_printf ("Created SD-Size: %d", *sd_size_ret);
+  debug_printf ("Created SD-Size: %u", sd_ret.size ());
 
   return sd_ret;
 }
 
 void
 set_security_attribute (int attribute, PSECURITY_ATTRIBUTES psa,
-			void *sd_buf, DWORD sd_buf_size)
+			security_descriptor &sd)
 {
-  psa->lpSecurityDescriptor = sd_buf;
-  InitializeSecurityDescriptor ((PSECURITY_DESCRIPTOR) sd_buf,
+  psa->lpSecurityDescriptor = sd.malloc (SECURITY_DESCRIPTOR_MIN_LENGTH);
+  InitializeSecurityDescriptor ((PSECURITY_DESCRIPTOR)psa->lpSecurityDescriptor,
 				SECURITY_DESCRIPTOR_REVISION);
-  psa->lpSecurityDescriptor = alloc_sd (geteuid32 (), getegid32 (), attribute,
-					(PSECURITY_DESCRIPTOR) sd_buf,
-					&sd_buf_size);
+  psa->lpSecurityDescriptor = alloc_sd (geteuid32 (), getegid32 (),
+  					attribute, sd);
 }
 
 static int
@@ -1832,22 +1814,19 @@ set_nt_attribute (const char *file, __uid32_t uid, __gid32_t gid,
   if (!wincap.has_security ())
     return 0;
 
-  DWORD sd_size = 4096;
-  char sd_buf[4096];
-  PSECURITY_DESCRIPTOR psd = (PSECURITY_DESCRIPTOR) sd_buf;
+  security_descriptor sd;
 
   int ret;
-  if ((ret = read_sd (file, psd, &sd_size)) <= 0)
+  if ((ret = read_sd (file, sd)) <= 0)
     {
       debug_printf ("read_sd %E");
       return -1;
     }
 
-  sd_size = 4096;
-  if (!(psd = alloc_sd (uid, gid, attribute, psd, &sd_size)))
+  if (!alloc_sd (uid, gid, attribute, sd))
     return -1;
 
-  return write_sd (file, psd, sd_size);
+  return write_sd (file, sd);
 }
 
 int
@@ -1880,9 +1859,9 @@ int
 check_file_access (const char *fn, int flags)
 {
   int ret = -1;
-  char sd_buf[4096];
-  DWORD sd_size = sizeof sd_buf;
-  PSECURITY_DESCRIPTOR psd = (PSECURITY_DESCRIPTOR) sd_buf;
+
+  security_descriptor sd;
+
   HANDLE hToken, hIToken;
   BOOL status;
   char pbuf[sizeof (PRIVILEGE_SET) + 3 * sizeof (LUID_AND_ATTRIBUTES)];
@@ -1891,7 +1870,7 @@ check_file_access (const char *fn, int flags)
 					     FILE_GENERIC_WRITE,
 					     FILE_GENERIC_EXECUTE,
 					     FILE_ALL_ACCESS };
-  if (read_sd (fn, psd, &sd_size) <= 0)
+  if (read_sd (fn, sd) <= 0)
     goto done;
 
   if (cygheap->user.issetuid ())
@@ -1914,7 +1893,7 @@ check_file_access (const char *fn, int flags)
     desired |= FILE_WRITE_DATA;
   if (flags & X_OK)
     desired |= FILE_EXECUTE;
-  if (!AccessCheck (psd, hIToken, desired, &mapping,
+  if (!AccessCheck (sd, hIToken, desired, &mapping,
 		    (PPRIVILEGE_SET) pbuf, &plength, &granted, &status))
     __seterrno ();
   else if (!status)
