@@ -1,6 +1,6 @@
 /* select.cc
 
-   Copyright 1996, 1997, 1998, 1999, 2000, 2001, 2002 Red Hat, Inc.
+   Copyright 1996, 1997, 1998, 1999, 2000, 2001, 2002, 2003, 2004 Red Hat, Inc.
 
    Written by Christopher Faylor of Cygnus Solutions
    cgf@cygnus.com
@@ -34,8 +34,8 @@ details. */
 #include "select.h"
 #include "cygerrno.h"
 #include "security.h"
-#include "fhandler.h"
 #include "path.h"
+#include "fhandler.h"
 #include "dtable.h"
 #include "cygheap.h"
 #include "sigproc.h"
@@ -102,7 +102,6 @@ cygwin_select (int maxfds, fd_set *readfds, fd_set *writefds, fd_set *exceptfds,
   fd_set *dummy_readfds = allocfd_set (maxfds);
   fd_set *dummy_writefds = allocfd_set (maxfds);
   fd_set *dummy_exceptfds = allocfd_set (maxfds);
-  sigframe thisframe (mainthread);
 
   select_printf ("%d, %p, %p, %p, %p", maxfds, readfds, writefds, exceptfds, to);
 
@@ -282,7 +281,7 @@ select_stuff::wait (fd_set *readfds, fd_set *writefds, fd_set *exceptfds,
 
       select_printf ("woke up.  wait_ret %d.  verifying", wait_ret);
       s = &start;
-      int gotone = FALSE;
+      bool gotone = false;
       /* Some types of object (e.g., consoles) wake up on "inappropriate" events
 	 like mouse movements.  The verify function will detect these situations.
 	 If it returns false, then this wakeup was a false alarm and we should go
@@ -335,7 +334,7 @@ set_bits (select_record *me, fd_set *readfds, fd_set *writefds,
   if (me->write_selected && me->write_ready)
     {
       UNIX_FD_SET (me->fd, writefds);
-      if (me->except_on_write && me->fh->get_device () == FH_SOCKET)
+      if (me->except_on_write && me->fh->is_socket ())
 	((fhandler_socket *) me->fh)->set_connect_state (CONNECTED);
       ready++;
     }
@@ -344,7 +343,7 @@ set_bits (select_record *me, fd_set *readfds, fd_set *writefds,
       if (me->except_on_write) /* Only on sockets */
 	{
 	  UNIX_FD_SET (me->fd, writefds);
-	  if (me->fh->get_device () == FH_SOCKET)
+	  if (me->fh->is_socket ())
 	    ((fhandler_socket *) me->fh)->set_connect_state (CONNECTED);
 	}
       if (me->except_selected)
@@ -512,7 +511,7 @@ static int start_thread_pipe (select_record *me, select_stuff *stuff);
 struct pipeinf
   {
     cygthread *thread;
-    BOOL stop_thread_pipe;
+    bool stop_thread_pipe;
     select_record *start;
   };
 
@@ -520,7 +519,7 @@ static DWORD WINAPI
 thread_pipe (void *arg)
 {
   pipeinf *pi = (pipeinf *) arg;
-  BOOL gotone = FALSE;
+  bool gotone = false;
 
   for (;;)
     {
@@ -553,32 +552,32 @@ out:
 static int
 start_thread_pipe (select_record *me, select_stuff *stuff)
 {
-  if (stuff->device_specific[FHDEVN (FH_PIPE)])
+  if (stuff->device_specific_pipe)
     {
-      me->h = *((pipeinf *) stuff->device_specific[FHDEVN (FH_PIPE)])->thread;
+      me->h = *((pipeinf *) stuff->device_specific_pipe)->thread;
       return 1;
     }
   pipeinf *pi = new pipeinf;
   pi->start = &stuff->start;
-  pi->stop_thread_pipe = FALSE;
+  pi->stop_thread_pipe = false;
   pi->thread = new cygthread (thread_pipe, (LPVOID) pi, "select_pipe");
   me->h = *pi->thread;
   if (!me->h)
     return 0;
-  stuff->device_specific[FHDEVN (FH_PIPE)] = (void *) pi;
+  stuff->device_specific_pipe = (void *) pi;
   return 1;
 }
 
 static void
 pipe_cleanup (select_record *, select_stuff *stuff)
 {
-  pipeinf *pi = (pipeinf *) stuff->device_specific[FHDEVN (FH_PIPE)];
+  pipeinf *pi = (pipeinf *) stuff->device_specific_pipe;
   if (pi && pi->thread)
     {
       pi->stop_thread_pipe = true;
       pi->thread->detach ();
       delete pi;
-      stuff->device_specific[FHDEVN (FH_PIPE)] = NULL;
+      stuff->device_specific_pipe = NULL;
     }
 }
 
@@ -849,7 +848,7 @@ static int start_thread_serial (select_record *me, select_stuff *stuff);
 struct serialinf
   {
     cygthread *thread;
-    BOOL stop_thread_serial;
+    bool stop_thread_serial;
     select_record *start;
   };
 
@@ -957,7 +956,7 @@ static DWORD WINAPI
 thread_serial (void *arg)
 {
   serialinf *si = (serialinf *) arg;
-  BOOL gotone= FALSE;
+  bool gotone = false;
 
   for (;;)
     {
@@ -984,30 +983,30 @@ thread_serial (void *arg)
 static int
 start_thread_serial (select_record *me, select_stuff *stuff)
 {
-  if (stuff->device_specific[FHDEVN (FH_SERIAL)])
+  if (stuff->device_specific_serial)
     {
-      me->h = *((serialinf *) stuff->device_specific[FHDEVN (FH_SERIAL)])->thread;
+      me->h = *((serialinf *) stuff->device_specific_serial)->thread;
       return 1;
     }
   serialinf *si = new serialinf;
   si->start = &stuff->start;
-  si->stop_thread_serial = FALSE;
+  si->stop_thread_serial = false;
   si->thread = new cygthread (thread_serial, (LPVOID) si, "select_serial");
   me->h = *si->thread;
-  stuff->device_specific[FHDEVN (FH_SERIAL)] = (void *) si;
+  stuff->device_specific_serial = (void *) si;
   return 1;
 }
 
 static void
 serial_cleanup (select_record *, select_stuff *stuff)
 {
-  serialinf *si = (serialinf *) stuff->device_specific[FHDEVN (FH_SERIAL)];
+  serialinf *si = (serialinf *) stuff->device_specific_serial;
   if (si && si->thread)
     {
       si->stop_thread_serial = true;
       si->thread->detach ();
       delete si;
-      stuff->device_specific[FHDEVN (FH_SERIAL)] = NULL;
+      stuff->device_specific_serial = NULL;
     }
 }
 
@@ -1086,6 +1085,7 @@ fhandler_base::ready_for_read (int fd, DWORD howlong)
 
       if (WaitForSingleObject (signal_arrived, avail ? 0 : 10) == WAIT_OBJECT_0)
 	{
+	  debug_printf ("interrupted");
 	  set_sig_errno (EINTR);
 	  avail = 0;
 	  break;
@@ -1255,7 +1255,7 @@ start_thread_socket (select_record *me, select_stuff *stuff)
 {
   socketinf *si;
 
-  if ((si = (socketinf *) stuff->device_specific[FHDEVN (FH_SOCKET)]))
+  if ((si = (socketinf *) stuff->device_specific_socket))
     {
       me->h = *si->thread;
       return 1;
@@ -1323,7 +1323,7 @@ start_thread_socket (select_record *me, select_stuff *stuff)
   select_printf ("exitsock %p", si->exitsock);
   WINSOCK_FD_SET ((HANDLE) si->exitsock, &si->readfds);
   WINSOCK_FD_SET ((HANDLE) si->exitsock, &si->exceptfds);
-  stuff->device_specific[FHDEVN (FH_SOCKET)] = (void *) si;
+  stuff->device_specific_socket = (void *) si;
   si->start = &stuff->start;
   select_printf ("stuff_start %p", &stuff->start);
   si->thread = new cygthread (thread_socket, (LPVOID) si, "select_socket");
@@ -1339,7 +1339,7 @@ err:
 void
 socket_cleanup (select_record *, select_stuff *stuff)
 {
-  socketinf *si = (socketinf *) stuff->device_specific[FHDEVN (FH_SOCKET)];
+  socketinf *si = (socketinf *) stuff->device_specific_socket;
   select_printf ("si %p si->thread %p", si, si ? si->thread : NULL);
   if (si && si->thread)
     {
@@ -1367,7 +1367,7 @@ socket_cleanup (select_record *, select_stuff *stuff)
       si->thread->detach ();
       shutdown (si->exitsock, SD_BOTH);
       closesocket (si->exitsock);
-      stuff->device_specific[FHDEVN (FH_SOCKET)] = NULL;
+      stuff->device_specific_socket = NULL;
       delete si;
     }
   select_printf ("returning");
