@@ -1,6 +1,6 @@
 /* spawn.cc
 
-   Copyright 1996, 1997, 1998, 1999, 2000, 2001, 2002, 2003 Red Hat, Inc.
+   Copyright 1996, 1997, 1998, 1999, 2000, 2001, 2002, 2003, 2004 Red Hat, Inc.
 
 This file is part of Cygwin.
 
@@ -21,8 +21,8 @@ details. */
 #include "cygerrno.h"
 #include <sys/cygwin.h>
 #include "security.h"
-#include "fhandler.h"
 #include "path.h"
+#include "fhandler.h"
 #include "dtable.h"
 #include "sigproc.h"
 #include "cygheap.h"
@@ -35,7 +35,7 @@ details. */
 #include "environ.h"
 #include "cygthread.h"
 
-#define LINE_BUF_CHUNK (MAX_PATH * 2)
+#define LINE_BUF_CHUNK (CYG_MAX_PATH * 2)
 
 static suffix_info std_suffixes[] =
 {
@@ -88,7 +88,7 @@ find_exec (const char *name, path_conv& buf, const char *mywinenv,
   const char *suffix = "";
   debug_printf ("find_exec (%s)", name);
   const char *retval = buf;
-  char tmp[MAX_PATH];
+  char tmp[CYG_MAX_PATH];
   const char *posix = (opt & FE_NATIVE) ? NULL : name;
   bool has_slash = strchr (name, '/');
 
@@ -307,7 +307,7 @@ av::unshift (const char *what, int conv)
 
   argv = av;
   memmove (argv + 1, argv, (argc + 1) * sizeof (char *));
-  char buf[MAX_PATH + 1];
+  char buf[CYG_MAX_PATH + 1];
   if (conv)
     {
       cygwin_conv_to_posix_path (what, buf);
@@ -349,9 +349,8 @@ static int __stdcall
 spawn_guts (const char * prog_arg, const char *const *argv,
 	    const char *const envp[], int mode)
 {
-  BOOL rc;
+  bool rc;
   pid_t cygpid;
-  sigframe thisframe (mainthread);
 
   MALLOC_CHECK;
 
@@ -398,12 +397,6 @@ spawn_guts (const char * prog_arg, const char *const *argv,
 
   init_child_info (chtype, &ciresrv, (mode == _P_OVERLAY) ? myself->pid : 1,
 		   subproc_ready);
-  if (!DuplicateHandle (hMainProc, hMainProc, hMainProc, &ciresrv.parent, 0, 1,
-			DUPLICATE_SAME_ACCESS))
-     {
-       system_printf ("couldn't create handle to myself for child, %E");
-       return -1;
-     }
 
   ciresrv.moreinfo = (cygheap_exec_info *) ccalloc (HEAP_1_EXEC, 1, sizeof (cygheap_exec_info));
   ciresrv.moreinfo->old_title = NULL;
@@ -462,7 +455,7 @@ spawn_guts (const char * prog_arg, const char *const *argv,
 
       DWORD done;
 
-      char buf[2 * MAX_PATH + 1];
+      char buf[2 * CYG_MAX_PATH + 1];
       buf[0] = buf[1] = buf[2] = buf[sizeof (buf) - 1] = '\0';
       if (!ReadFile (hnd, buf, sizeof (buf) - 1, &done, 0))
 	{
@@ -587,6 +580,13 @@ spawn_guts (const char * prog_arg, const char *const *argv,
       else
 	one_line.add ("", 1);
       MALLOC_CHECK;
+
+      if (one_line.ix > 32767)
+	{
+	  debug_printf ("Command line too long (>32K), return E2BIG");
+	  set_errno (E2BIG);
+	  return -1;
+	}
     }
 
   char *envblock;
@@ -606,6 +606,8 @@ spawn_guts (const char * prog_arg, const char *const *argv,
 			&ciresrv.moreinfo->myself_pinfo, 0,
 			TRUE, DUPLICATE_SAME_ACCESS))
     ciresrv.moreinfo->myself_pinfo = NULL;
+  else
+    VerifyHandle (ciresrv.moreinfo->myself_pinfo);
 
  skip_arg_parsing:
   PROCESS_INFORMATION pi = {NULL, 0, 0, 0};
@@ -637,8 +639,6 @@ spawn_guts (const char * prog_arg, const char *const *argv,
   syscall_printf ("null_app_name %d (%s, %.9500s)", null_app_name, runpath, one_line.buf);
 
   void *newheap;
-  /* Preallocated buffer for `sec_user' call */
-  char sa_buf[1024];
 
   cygbench ("spawn-guts");
 
@@ -647,21 +647,20 @@ spawn_guts (const char * prog_arg, const char *const *argv,
   /* When ruid != euid we create the new process under the current original
      account and impersonate in child, this way maintaining the different
      effective vs. real ids.
-     FIXME: If ruid != euid and ruid != orig_uid we currently give
+     FIXME: If ruid != euid and ruid != saved_uid we currently give
      up on ruid. The new process will have ruid == euid. */
   if (!cygheap->user.issetuid ()
-      || (cygheap->user.orig_uid == cygheap->user.real_uid
-	  && cygheap->user.orig_gid == cygheap->user.real_gid
+      || (cygheap->user.saved_uid == cygheap->user.real_uid
+	  && cygheap->user.saved_gid == cygheap->user.real_gid
 	  && !cygheap->user.groups.issetgroups ()))
     {
-      PSECURITY_ATTRIBUTES sec_attribs = sec_user_nih (sa_buf);
       ciresrv.moreinfo->envp = build_env (envp, envblock, ciresrv.moreinfo->envc,
 					  real_path.iscygexec ());
       newheap = cygheap_setup_for_child (&ciresrv, cygheap->fdtab.need_fixup_before ());
       rc = CreateProcess (runpath,	/* image name - with full path */
 			  one_line.buf,	/* what was passed to exec */
-			  sec_attribs,	/* process security attrs */
-			  sec_attribs,	/* thread security attrs */
+			  &sec_none_nih,/* process security attrs */
+			  &sec_none_nih,/* thread security attrs */
 			  TRUE,		/* inherit handles from parent */
 			  flags,
 			  envblock,	/* environment */
@@ -671,10 +670,9 @@ spawn_guts (const char * prog_arg, const char *const *argv,
     }
   else
     {
-      PSID sid = cygheap->user.sid ();
-
-      /* Set security attributes with sid */
-      PSECURITY_ATTRIBUTES sec_attribs = sec_user_nih (sa_buf, sid);
+      /* Give access to myself */
+      if (mode == _P_OVERLAY)
+	myself.set_acl();
 
       /* allow the child to interact with our window station/desktop */
       HANDLE hwst, hdsk;
@@ -699,8 +697,8 @@ spawn_guts (const char * prog_arg, const char *const *argv,
       rc = CreateProcessAsUser (cygheap->user.token (),
 		       runpath,		/* image name - with full path */
 		       one_line.buf,	/* what was passed to exec */
-		       sec_attribs,     /* process security attrs */
-		       sec_attribs,     /* thread security attrs */
+		       &sec_none_nih,   /* process security attrs */
+		       &sec_none_nih,   /* thread security attrs */
 		       TRUE,		/* inherit handles from parent */
 		       flags,
 		       envblock,	/* environment */
@@ -709,24 +707,9 @@ spawn_guts (const char * prog_arg, const char *const *argv,
 		       &pi);
     }
 
-  /* FIXME: There is a small race here */
-
-  DWORD res;
-  pthread_cleanup cleanup;
-  pthread_cleanup_push (do_cleanup, (void *) &cleanup);
-  if (mode == _P_SYSTEM)
-    {
-      sigset_t child_block;
-      cleanup.oldint = signal (SIGINT, SIG_IGN);
-      cleanup.oldquit = signal (SIGQUIT, SIG_IGN);
-      sigemptyset (&child_block);
-      sigaddset (&child_block, SIGCHLD);
-      (void) sigprocmask (SIG_BLOCK, &child_block, &cleanup.oldmask);
-    }
-
   /* Restore impersonation. In case of _P_OVERLAY this isn't
      allowed since it would overwrite child data. */
-  if (mode != _P_OVERLAY)
+  if (mode != _P_OVERLAY || !rc)
       cygheap->user.reimpersonate ();
 
   MALLOC_CHECK;
@@ -747,7 +730,22 @@ spawn_guts (const char * prog_arg, const char *const *argv,
       return -1;
     }
 
-  /* Fixup the parent datastructure if needed and resume the child's
+  /* FIXME: There is a small race here */
+
+  int res;
+  pthread_cleanup cleanup;
+  pthread_cleanup_push (do_cleanup, (void *) &cleanup);
+  if (mode == _P_SYSTEM)
+    {
+      sigset_t child_block;
+      cleanup.oldint = signal (SIGINT, SIG_IGN);
+      cleanup.oldquit = signal (SIGQUIT, SIG_IGN);
+      sigemptyset (&child_block);
+      sigaddset (&child_block, SIGCHLD);
+      (void) sigprocmask (SIG_BLOCK, &child_block, &cleanup.oldmask);
+    }
+
+  /* Fixup the parent data structures if needed and resume the child's
      main thread. */
   if (!cygheap->fdtab.need_fixup_before ())
     cygheap_setup_for_child_cleanup (newheap, &ciresrv, 0);
@@ -788,16 +786,25 @@ spawn_guts (const char * prog_arg, const char *const *argv,
     {
       myself->set_has_pgid_children ();
       ProtectHandle (pi.hThread);
-      pinfo child (cygpid, 1);
+      pinfo child (cygpid, PID_IN_USE);
       if (!child)
 	{
-	  set_errno (EAGAIN);
-	  syscall_printf ("-1 = spawnve (), process table full");
-	  return -1;
+	  syscall_printf ("pinfo failed");
+	  if (get_errno () != ENOMEM)
+	    set_errno (EAGAIN);
+	  res = -1;
+	  goto out;
 	}
       child->dwProcessId = pi.dwProcessId;
       child->hProcess = pi.hProcess;
-      child.remember ();
+      if (!child.remember ())
+	{
+	  syscall_printf ("process table full");
+	  set_errno (EAGAIN);
+	  res = -1;
+	  goto out;
+	}
+
       strcpy (child->progname, real_path);
       /* FIXME: This introduces an unreferenced, open handle into the child.
 	 The purpose is to keep the pid shared memory open so that all of
@@ -814,10 +821,10 @@ spawn_guts (const char * prog_arg, const char *const *argv,
 
   sigproc_printf ("spawned windows pid %d", pi.dwProcessId);
 
-  BOOL exited;
+  bool exited;
 
   res = 0;
-  exited = FALSE;
+  exited = false;
   MALLOC_CHECK;
   if (mode == _P_OVERLAY)
     {
@@ -833,7 +840,7 @@ spawn_guts (const char * prog_arg, const char *const *argv,
 	      if (!GetExitCodeProcess (pi.hProcess, &exitcode))
 		exitcode = 1;
 	      res |= exitcode;
-	      exited = TRUE;
+	      exited = true;
 	      break;
 	    case WAIT_OBJECT_0 + 1:
 	      sigproc_printf ("signal arrived");
@@ -883,6 +890,7 @@ spawn_guts (const char * prog_arg, const char *const *argv,
 				    0, FALSE, DUPLICATE_SAME_ACCESS);
 	  sigproc_printf ("%d = DuplicateHandle, oldh %p, newh %p",
 			  rc, oldh, myself->hProcess);
+	  VerifyHandle (myself->hProcess);
 	  if (!rc && my_parent_is_alive ())
 	    {
 	      system_printf ("Reparent failed, parent handle %p, %E", h);
@@ -900,12 +908,12 @@ spawn_guts (const char * prog_arg, const char *const *argv,
     {
     case _P_OVERLAY:
       ForceCloseHandle1 (pi.hProcess, childhProc);
-      proc_terminate ();
       myself->exit (res, 1);
       break;
     case _P_WAIT:
     case _P_SYSTEM:
-      waitpid (cygpid, (int *) &res, 0);
+      if (waitpid (cygpid, (int *) &res, 0) != cygpid)
+	res = -1;
       break;
     case _P_DETACH:
       res = 0;	/* Lose all memory of this child. */
@@ -919,6 +927,7 @@ spawn_guts (const char * prog_arg, const char *const *argv,
       break;
     }
 
+out:
   pthread_cleanup_pop (1);
   return (int) res;
 }
@@ -939,12 +948,14 @@ spawnve (int mode, const char *path, const char *const *argv,
 	 const char *const *envp)
 {
   int ret;
+#ifdef NEWVFORK
   vfork_save *vf = vfork_storage.val ();
 
   if (vf != NULL && (vf->pid < 0) && mode == _P_OVERLAY)
     mode = _P_NOWAIT;
   else
     vf = NULL;
+#endif
 
   syscall_printf ("spawnve (%s, %s, %x)", path, argv[0], envp);
 
@@ -965,14 +976,16 @@ spawnve (int mode, const char *path, const char *const *argv,
     case _P_SYSTEM:
       subproc_init ();
       ret = spawn_guts (path, argv, envp, mode);
+#ifdef NEWVFORK
       if (vf)
 	{
-	  debug_printf ("longjmping due to vfork");
-	  if (ret < 0)
-	    vf->restore_exit (ret);
-	  else
-	    vf->restore_pid (ret);
+	  if (ret > 0)
+	    {
+	      debug_printf ("longjmping due to vfork");
+	      vf->restore_pid (ret);
+	    }
 	}
+#endif
       break;
     default:
       set_errno (EINVAL);
