@@ -16,13 +16,14 @@ details. */
 #include <stdlib.h>
 #include "cygerrno.h"
 #include "security.h"
-#include "fhandler.h"
 #include "path.h"
+#include "fhandler.h"
 #include "dtable.h"
 #include "cygheap.h"
 #include "cygwin/version.h"
 #include "perprocess.h"
 #include "sigproc.h"
+#include "cygtls.h"
 #include <sys/termios.h>
 
 /* tcsendbreak: POSIX 7.2.2.1 */
@@ -119,8 +120,6 @@ tcsetattr (int fd, int a, const struct termios *t)
 
   while (1)
     {
-      sigframe thisframe (mainthread);
-
       res = -1;
       cygheap_fdget cfd (fd);
       if (cfd < 0)
@@ -148,7 +147,7 @@ tcsetattr (int fd, int a, const struct termios *t)
 	  e = get_errno ();
 	  break;
 	case bg_signalled:
-	  if (thisframe.call_signal_handler ())
+	  if (_my_tls.call_signal_handler ())
 	    continue;
 	  res = -1;
 	  /* fall through intentionally */
