@@ -1,7 +1,7 @@
 /* wincap.cc -- figure out on which OS we're running. Set the
 		capability class to the appropriate values.
 
-   Copyright 2001, 2002 Red Hat, Inc.
+   Copyright 2001, 2002, 2003, 2004 Red Hat, Inc.
 
 This file is part of Cygwin.
 
@@ -16,6 +16,7 @@ static NO_COPY wincaps wincap_unknown = {
   chunksize:0x0,
   shared:FILE_SHARE_READ | FILE_SHARE_WRITE,
   is_winnt:false,
+  is_server:false,
   access_denied_on_delete:false,
   has_delete_on_close:false,
   has_page_guard:false,
@@ -50,7 +51,9 @@ static NO_COPY wincaps wincap_unknown = {
   needs_memory_protection:false,
   pty_needs_alloc_console:false,
   has_terminal_services:false,
-  has_switch_to_thread:false
+  has_switch_to_thread:false,
+  cant_debug_dll_entry:false,
+  has_ioctl_storage_get_media_types_ex:false
 };
 
 static NO_COPY wincaps wincap_95 = {
@@ -58,6 +61,7 @@ static NO_COPY wincaps wincap_95 = {
   chunksize:32 * 1024 * 1024,
   shared:FILE_SHARE_READ | FILE_SHARE_WRITE,
   is_winnt:false,
+  is_server:false,
   access_denied_on_delete:true,
   has_delete_on_close:false,
   has_page_guard:false,
@@ -92,7 +96,9 @@ static NO_COPY wincaps wincap_95 = {
   needs_memory_protection:false,
   pty_needs_alloc_console:false,
   has_terminal_services:false,
-  has_switch_to_thread:false
+  has_switch_to_thread:false,
+  cant_debug_dll_entry:true,
+  has_ioctl_storage_get_media_types_ex:false
 };
 
 static NO_COPY wincaps wincap_95osr2 = {
@@ -100,6 +106,7 @@ static NO_COPY wincaps wincap_95osr2 = {
   chunksize:32 * 1024 * 1024,
   shared:FILE_SHARE_READ | FILE_SHARE_WRITE,
   is_winnt:false,
+  is_server:false,
   access_denied_on_delete:true,
   has_delete_on_close:false,
   has_page_guard:false,
@@ -134,7 +141,9 @@ static NO_COPY wincaps wincap_95osr2 = {
   needs_memory_protection:false,
   pty_needs_alloc_console:false,
   has_terminal_services:false,
-  has_switch_to_thread:false
+  has_switch_to_thread:false,
+  cant_debug_dll_entry:true,
+  has_ioctl_storage_get_media_types_ex:false
 };
 
 static NO_COPY wincaps wincap_98 = {
@@ -142,6 +151,7 @@ static NO_COPY wincaps wincap_98 = {
   chunksize:32 * 1024 * 1024,
   shared:FILE_SHARE_READ | FILE_SHARE_WRITE,
   is_winnt:false,
+  is_server:false,
   access_denied_on_delete:true,
   has_delete_on_close:false,
   has_page_guard:false,
@@ -176,7 +186,9 @@ static NO_COPY wincaps wincap_98 = {
   needs_memory_protection:false,
   pty_needs_alloc_console:false,
   has_terminal_services:false,
-  has_switch_to_thread:false
+  has_switch_to_thread:false,
+  cant_debug_dll_entry:true,
+  has_ioctl_storage_get_media_types_ex:false
 };
 
 static NO_COPY wincaps wincap_98se = {
@@ -184,6 +196,7 @@ static NO_COPY wincaps wincap_98se = {
   chunksize:32 * 1024 * 1024,
   shared:FILE_SHARE_READ | FILE_SHARE_WRITE,
   is_winnt:false,
+  is_server:false,
   access_denied_on_delete:true,
   has_delete_on_close:false,
   has_page_guard:false,
@@ -218,7 +231,9 @@ static NO_COPY wincaps wincap_98se = {
   needs_memory_protection:false,
   pty_needs_alloc_console:false,
   has_terminal_services:false,
-  has_switch_to_thread:false
+  has_switch_to_thread:false,
+  cant_debug_dll_entry:true,
+  has_ioctl_storage_get_media_types_ex:false
 };
 
 static NO_COPY wincaps wincap_me = {
@@ -226,6 +241,7 @@ static NO_COPY wincaps wincap_me = {
   chunksize:32 * 1024 * 1024,
   shared:FILE_SHARE_READ | FILE_SHARE_WRITE,
   is_winnt:false,
+  is_server:false,
   access_denied_on_delete:true,
   has_delete_on_close:false,
   has_page_guard:false,
@@ -260,14 +276,17 @@ static NO_COPY wincaps wincap_me = {
   needs_memory_protection:false,
   pty_needs_alloc_console:false,
   has_terminal_services:false,
-  has_switch_to_thread:false
+  has_switch_to_thread:false,
+  cant_debug_dll_entry:true,
+  has_ioctl_storage_get_media_types_ex:false
 };
 
 static NO_COPY wincaps wincap_nt3 = {
-  lock_file_highword:0xffffffff,
+  lock_file_highword:UINT32_MAX,
   chunksize:0,
   shared:FILE_SHARE_READ | FILE_SHARE_WRITE | FILE_SHARE_DELETE,
   is_winnt:true,
+  is_server:false,
   access_denied_on_delete:false,
   has_delete_on_close:true,
   has_page_guard:true,
@@ -302,14 +321,17 @@ static NO_COPY wincaps wincap_nt3 = {
   needs_memory_protection:true,
   pty_needs_alloc_console:true,
   has_terminal_services:false,
-  has_switch_to_thread:false
+  has_switch_to_thread:false,
+  cant_debug_dll_entry:false,
+  has_ioctl_storage_get_media_types_ex:false
 };
 
 static NO_COPY wincaps wincap_nt4 = {
-  lock_file_highword:0xffffffff,
+  lock_file_highword:UINT32_MAX,
   chunksize:0,
   shared:FILE_SHARE_READ | FILE_SHARE_WRITE | FILE_SHARE_DELETE,
   is_winnt:true,
+  is_server:false,
   access_denied_on_delete:false,
   has_delete_on_close:true,
   has_page_guard:true,
@@ -344,14 +366,17 @@ static NO_COPY wincaps wincap_nt4 = {
   needs_memory_protection:true,
   pty_needs_alloc_console:true,
   has_terminal_services:false,
-  has_switch_to_thread:true
+  has_switch_to_thread:true,
+  cant_debug_dll_entry:false,
+  has_ioctl_storage_get_media_types_ex:false
 };
 
 static NO_COPY wincaps wincap_nt4sp4 = {
-  lock_file_highword:0xffffffff,
+  lock_file_highword:UINT32_MAX,
   chunksize:0,
   shared:FILE_SHARE_READ | FILE_SHARE_WRITE | FILE_SHARE_DELETE,
   is_winnt:true,
+  is_server:false,
   access_denied_on_delete:false,
   has_delete_on_close:true,
   has_page_guard:true,
@@ -386,14 +411,17 @@ static NO_COPY wincaps wincap_nt4sp4 = {
   needs_memory_protection:true,
   pty_needs_alloc_console:true,
   has_terminal_services:false,
-  has_switch_to_thread:true
+  has_switch_to_thread:true,
+  cant_debug_dll_entry:false,
+  has_ioctl_storage_get_media_types_ex:false
 };
 
 static NO_COPY wincaps wincap_2000 = {
-  lock_file_highword:0xffffffff,
+  lock_file_highword:UINT32_MAX,
   chunksize:0,
   shared:FILE_SHARE_READ | FILE_SHARE_WRITE | FILE_SHARE_DELETE,
   is_winnt:true,
+  is_server:false,
   access_denied_on_delete:false,
   has_delete_on_close:true,
   has_page_guard:true,
@@ -428,14 +456,17 @@ static NO_COPY wincaps wincap_2000 = {
   needs_memory_protection:true,
   pty_needs_alloc_console:true,
   has_terminal_services:true,
-  has_switch_to_thread:true
+  has_switch_to_thread:true,
+  cant_debug_dll_entry:false,
+  has_ioctl_storage_get_media_types_ex:false
 };
 
 static NO_COPY wincaps wincap_xp = {
-  lock_file_highword:0xffffffff,
+  lock_file_highword:UINT32_MAX,
   chunksize:0,
   shared:FILE_SHARE_READ | FILE_SHARE_WRITE | FILE_SHARE_DELETE,
   is_winnt:true,
+  is_server:false,
   access_denied_on_delete:false,
   has_delete_on_close:true,
   has_page_guard:true,
@@ -470,14 +501,17 @@ static NO_COPY wincaps wincap_xp = {
   needs_memory_protection:true,
   pty_needs_alloc_console:true,
   has_terminal_services:true,
-  has_switch_to_thread:true
+  has_switch_to_thread:true,
+  cant_debug_dll_entry:false,
+  has_ioctl_storage_get_media_types_ex:true
 };
 
 static NO_COPY wincaps wincap_2003 = {
-  lock_file_highword:0xffffffff,
+  lock_file_highword:UINT32_MAX,
   chunksize:0,
   shared:FILE_SHARE_READ | FILE_SHARE_WRITE | FILE_SHARE_DELETE,
   is_winnt:true,
+  is_server:true,
   access_denied_on_delete:false,
   has_delete_on_close:true,
   has_page_guard:true,
@@ -512,7 +546,9 @@ static NO_COPY wincaps wincap_2003 = {
   needs_memory_protection:true,
   pty_needs_alloc_console:true,
   has_terminal_services:true,
-  has_switch_to_thread:true
+  has_switch_to_thread:true,
+  cant_debug_dll_entry:false,
+  has_ioctl_storage_get_media_types_ex:true
 };
 
 wincapc wincap;
@@ -521,13 +557,15 @@ void
 wincapc::init ()
 {
   const char *os;
+  bool has_osversioninfoex = false;
 
   if (caps)
     return;		// already initialized
 
   memset (&version, 0, sizeof version);
-  version.dwOSVersionInfoSize = sizeof version;
-  GetVersionEx (&version);
+  /* Request simple version info first. */
+  version.dwOSVersionInfoSize = sizeof (OSVERSIONINFO);
+  GetVersionEx (reinterpret_cast<LPOSVERSIONINFO>(&version));
 
   switch (version.dwPlatformId)
     {
@@ -543,10 +581,15 @@ wincapc::init ()
 	      if (strcmp (version.szCSDVersion, "Service Pack 4") < 0)
 		caps = &wincap_nt4;
 	      else
-		caps = &wincap_nt4sp4;
+	        {
+		  caps = &wincap_nt4sp4;
+		  if (strcmp (version.szCSDVersion, "Service Pack 6") >= 0)
+		    has_osversioninfoex = true;
+		}
 	      break;
 	    case 5:
 	      os = "NT";
+	      has_osversioninfoex = true;
 	      switch (version.dwMinorVersion)
 	        {
 		  case 0:
@@ -599,6 +642,17 @@ wincapc::init ()
 	caps = &wincap_unknown;
 	break;
     }
+
+  if (has_osversioninfoex)
+    {
+      /* Request extended version to get server info.
+	 Available since NT4 SP6. */
+      version.dwOSVersionInfoSize = sizeof (OSVERSIONINFOEX);
+      GetVersionEx (reinterpret_cast<LPOSVERSIONINFO>(&version));
+      if (version.wProductType != VER_NT_WORKSTATION)
+	((wincaps *)this->caps)->is_server = true;
+    }
+
   __small_sprintf (osnam, "%s-%d.%d", os, version.dwMajorVersion,
 		   version.dwMinorVersion);
 }
