@@ -24,8 +24,6 @@
  *
  */
 
-#ifndef	__STRICT_ANSI__
-
 #ifndef	_TIMEB_H_
 #define	_TIMEB_H_
 
@@ -77,6 +75,3 @@ _CRTIMP void __cdecl	ftime (struct timeb*);
 #endif	/* Not RC_INVOKED */
 
 #endif	/* Not _TIMEB_H_ */
-
-#endif	/* Not __STRICT_ANSI__ */
-
