@@ -1,6 +1,6 @@
 /* cygcheck.cc
 
-   Copyright 1998, 1999, 2000, 2001, 2002, 2003 Red Hat, Inc.
+   Copyright 1998, 1999, 2000, 2001, 2002, 2003, 2004 Red Hat, Inc.
 
    This file is part of Cygwin.
 
@@ -40,7 +40,7 @@ void dump_setup (int, char **, bool);
 void package_find (int, char **);
 void package_list (int, char **);
 
-static const char version[] = "$Revision: 1.38 $";
+static const char version[] = "$Revision: 1.42 $";
 
 static const char *known_env_vars[] = {
   "c_include_path",
@@ -782,15 +782,25 @@ pretty_id (const char *s, char *cygwin, size_t cyglen)
   if (access (id, X_OK))
     {
       fprintf (stderr, "`id' program not found\n");
-      exit (1);
+      return;
     }
 
   FILE *f = popen (id, "rt");
 
   char buf[16384];
+  static char empty[] = "";
+  buf[0] = '\0';
   fgets (buf, sizeof (buf), f);
-  char *uid = strtok (buf, ")") + strlen ("uid=");
-  char *gid = strtok (NULL, ")") + strlen ("gid=") + 1;
+  char *uid = strtok (buf, ")");
+  if (uid)
+    uid += strlen ("uid=");
+  else
+    uid = empty;
+  char *gid = strtok (NULL, ")");
+  if (gid)
+    gid += strlen ("gid=") + 1;
+  else
+    gid = empty;
   char **ng;
   size_t sz = 0;
   for (ng = groups; (*ng = strtok (NULL, ",")); ng++)
@@ -1370,7 +1380,7 @@ print_version ()
   printf ("\
 cygcheck version %.*s\n\
 System Checker for Cygwin\n\
-Copyright 1998, 1999, 2000, 2001, 2002, 2003 Red Hat, Inc.\n\
+Copyright 1998, 1999, 2000, 2001, 2002, 2003, 2004 Red Hat, Inc.\n\
 Compiled on %s\n\
 ", len, v, __DATE__);
 }
@@ -1380,6 +1390,7 @@ main (int argc, char **argv)
 {
   int i;
 
+  (void) putenv("POSIXLY_CORRECT=1");
   while ((i = getopt_long (argc, argv, opts, longopts, NULL)) != EOF)
     switch (i)
       {
