@@ -1,5 +1,5 @@
 /* cygpath.cc -- convert pathnames between Windows and Unix format
-   Copyright 1998, 1999, 2000, 2001, 2002 Red Hat, Inc.
+   Copyright 1998, 1999, 2000, 2001, 2002, 2003, 2004 Red Hat, Inc.
 
 This file is part of Cygwin.
 
@@ -20,8 +20,9 @@ details. */
 #include <sys/fcntl.h>
 #include <sys/cygwin.h>
 #include <ctype.h>
+#include <errno.h>
 
-static const char version[] = "$Revision: 1.29 $";
+static const char version[] = "$Revision: 1.36 $";
 
 static char *prog_name;
 static char *file_arg;
@@ -40,6 +41,7 @@ static struct option long_options[] = {
   {(char *) "ignore", no_argument, NULL, 'i'},
   {(char *) "long-name", no_argument, NULL, 'l'},
   {(char *) "mixed", no_argument, NULL, 'm'},
+  {(char *) "mode", no_argument, NULL, 'M'},
   {(char *) "option", no_argument, NULL, 'o'},
   {(char *) "path", no_argument, NULL, 'p'},
   {(char *) "short-name", no_argument, NULL, 's'},
@@ -56,7 +58,7 @@ static struct option long_options[] = {
   {0, no_argument, 0, 0}
 };
 
-static char options[] = "ac:df:hilmopst:uvwADHPSW";
+static char options[] = "ac:df:hilmMopst:uvwADHPSW";
 
 static void
 usage (FILE * stream, int status)
@@ -70,6 +72,7 @@ Convert Unix and Windows format paths, or output system path information\n\
 Output type options:\n\
   -d, --dos	        print DOS (short) form of NAME (C:\\PROGRA~1\\)\n\
   -m, --mixed           like --windows, but with regular slashes (C:/WINNT)\n\
+  -M, --mode		report on mode of file (binmode or textmode)\n\
   -u, --unix	        (default) print Unix form of NAME (/cygdrive/c/winnt)\n\
   -w, --windows         print Windows form of NAME (C:\\WINNT)\n\
   -t, --type TYPE       print TYPE form: 'dos', 'mixed', 'unix', or 'windows'\n\
@@ -86,9 +89,13 @@ System information:\n\
   -S, --sysdir		output system directory and exit\n\
   -W, --windir		output `Windows' directory and exit\n\
 ", prog_name, prog_name);
-  if (stream == stdout)
+  if (ignore_flag)
+    /* nothing to do */;
+  else if (stream != stdout)
+    fprintf(stream, "Try `%s --help' for more information.\n", prog_name);
+  else
     {
-    fprintf (stream, "\
+      fprintf (stream, "\
 Other options:\n\
   -f, --file FILE       read FILE for input; use - to read from STDIN\n\
   -o, --option          read options from FILE as well (for use with --file)\n\
@@ -98,8 +105,6 @@ Other options:\n\
   -v, --version		output version information and exit\n\
 ");
     }
-  else
-    fprintf(stream, "Try `%s --help' for more information.\n", prog_name);
   exit (ignore_flag ? 0 : status);
 }
 
@@ -391,8 +396,11 @@ dowin (char option)
 
   if (!windows_flag)
     {
-      cygwin_conv_to_posix_path (buf, buf2);
-      buf = buf2;
+      if (cygwin_conv_to_posix_path (buf, buf2))
+	fprintf (stderr, "%s: error converting \"%s\" - %s\n",
+		 prog_name, buf, strerror (errno));
+      else
+	buf = buf2;
     }
   else
     {
@@ -405,12 +413,30 @@ dowin (char option)
   exit (0);
 }
 
+static void
+report_mode (char *filename)
+{
+  switch (cygwin_internal (CW_GET_BINMODE, filename))
+    {
+    case O_BINARY:
+      printf ("%s: binary\n", filename);
+      break;
+    case O_TEXT:
+      printf ("%s: text\n", filename);
+      break;
+    default:
+      fprintf (stderr, "%s: file '%s' - %s\n", prog_name, filename,
+	       strerror (errno));
+      break;
+    }
+}
+
 static void
 doit (char *filename)
 {
   char *buf;
   DWORD len;
-  int retval;
+  int err;
   int (*conv_func) (const char *, char *);
 
   if (!path_flag)
@@ -441,10 +467,12 @@ doit (char *filename)
   if (path_flag)
     {
       if (unix_flag)
-	cygwin_win32_to_posix_path_list (filename, buf);
+	err = cygwin_win32_to_posix_path_list (filename, buf);
       else
 	{
-	  cygwin_posix_to_win32_path_list (filename, buf);
+	  err = cygwin_posix_to_win32_path_list (filename, buf);
+	  if (err)
+	    /* oops */;
 	  if (shortname_flag)
 	    buf = get_short_paths (buf);
 	  if (longname_flag)
@@ -452,6 +480,12 @@ doit (char *filename)
 	  if (mixed_flag)
 	    buf = get_mixed_name (buf);
 	}
+      if (err)
+	{
+	  fprintf (stderr, "%s: error converting \"%s\" - %s\n",
+		   prog_name, filename, strerror (errno));
+	  exit (1);
+	}
     }
   else
     {
@@ -461,13 +495,13 @@ doit (char *filename)
       else
 	conv_func = (absolute_flag ? cygwin_conv_to_full_win32_path :
 		     cygwin_conv_to_win32_path);
-      retval = conv_func (filename, buf);
+      err = conv_func (filename, buf);
       if (mixed_flag)
 	buf = get_mixed_name (buf);
-      if (retval < 0)
+      if (err)
 	{
-	  fprintf (stderr, "%s: error converting \"%s\"\n",
-		   prog_name, filename);
+	  fprintf (stderr, "%s: error converting \"%s\" - %s\n",
+		   prog_name, filename, strerror (errno));
 	  exit (1);
 	}
       if (!unix_flag)
@@ -510,7 +544,7 @@ main (int argc, char **argv)
 {
   int c, o = 0;
   int options_from_file_flag;
-  char *filename;
+  int mode_flag;
 
   prog_name = strrchr (argv[0], '/');
   if (prog_name == NULL)
@@ -530,6 +564,7 @@ main (int argc, char **argv)
   options_from_file_flag = 0;
   allusers_flag = 0;
   output_flag = 0;
+  mode_flag = 0;
   while ((c = getopt_long (argc, argv, options,
 			   long_options, (int *) NULL)) != EOF)
     {
@@ -555,6 +590,10 @@ main (int argc, char **argv)
 	  file_arg = optarg;
 	  break;
 
+	case 'M':
+	  mode_flag = 1;
+	  break;
+
 	case 'o':
 	  options_from_file_flag = 1;
 	  break;
@@ -675,11 +714,14 @@ main (int argc, char **argv)
       if (output_flag)
 	dowin (o);
 
-      if (optind != argc - 1)
+      if (optind > argc - 1)
 	usage (stderr, 1);
 
-      filename = argv[optind];
-      doit (filename);
+      for (int i = optind; argv[i]; i++)
+	if (mode_flag)
+	  report_mode (argv[i]);
+	else
+	  doit (argv[i]);
     }
   else
     {
