@@ -1,6 +1,6 @@
 /* mount.cc
 
-   Copyright 1996, 1997, 1998, 1999, 2000, 2001, 2002 Red Hat, Inc.
+   Copyright 1996, 1997, 1998, 1999, 2000, 2001, 2002, 2003 Red Hat, Inc.
 
 This file is part of Cygwin.
 
@@ -16,6 +16,7 @@ details. */
 #include <sys/cygwin.h>
 #include <stdlib.h>
 #include <getopt.h>
+#include <dirent.h>
 
 #ifdef errno
 #undef errno
@@ -104,6 +105,25 @@ do_mount (const char *dev, const char *where, int flags)
 	}
     }
 
+  if (!force && flags & MOUNT_ENC)
+    {
+      DIR *dd = opendir (dev);
+      if (dd)
+	{
+	  struct dirent *d;
+	  while ((d = readdir (dd)))
+	    {
+	      if (d->d_name[0] != '.')
+		/* fall through */;
+	      else if (d->d_name[1] == '\0'
+		       || (d->d_name[1] == '.' && d->d_name[2] == '\0'))
+		continue;
+	      fprintf (stderr, "%s: error: don't use \"-o managed\" on non-empty directories\n", progname);
+	      exit (1);
+	    }
+	}
+    }
+
   if (mount (dev, where, flags))
     error (where);
 
@@ -120,6 +140,7 @@ static struct option longopts[] =
   {"help", no_argument, NULL, 'h' },
   {"mount-commands", no_argument, NULL, 'm'},
   {"no-executable", no_argument, NULL, 'E'},
+  {"options", required_argument, NULL, 'E'},
   {"show-cygdrive-prefix", no_argument, NULL, 'p'},
   {"system", no_argument, NULL, 's'},
   {"text", no_argument, NULL, 't'},
@@ -128,12 +149,14 @@ static struct option longopts[] =
   {NULL, 0, NULL, 0}
 };
 
-static char opts[] = "bcfhmpstuvxEX";
+static char opts[] = "bcfhmpstuvxEXo:";
 
 static void
 usage (FILE *where = stderr)
 {
   fprintf (where, "Usage: %s [OPTION] [<win32path> <posixpath>]\n\
+Display information about mounted filesystems, or mount a filesystem\n\
+\n\
   -b, --binary     (default)    text files are equivalent to binary files\n\
 				(newline = \\n)\n\
   -c, --change-cygdrive-prefix  change the cygdrive path prefix to <posixpath>\n\
@@ -142,6 +165,7 @@ usage (FILE *where = stderr)
   -h, --help                    output usage information and exit\n\
   -m, --mount-commands          write mount commands to replace user and\n\
 				system mount points and cygdrive prefixes\n\
+  -o, --options X[,X...]	specify mount options\n\
   -p, --show-cygdrive-prefix    show user and/or system cygdrive path prefix\n\
   -s, --system     (default)    add system-wide mount point\n\
   -t, --text                    text files get \\r\\n line endings\n\
@@ -156,6 +180,24 @@ usage (FILE *where = stderr)
   exit (where == stderr ? 1 : 0);
 }
 
+struct opt
+{
+  const char *name;
+  unsigned val;
+  bool clear;
+} oopts[] =
+{
+  {"user", MOUNT_SYSTEM, 1},
+  {"system", MOUNT_SYSTEM, 0},
+  {"binary", MOUNT_BINARY, 0},
+  {"text", MOUNT_BINARY, 1},
+  {"exec", MOUNT_EXEC, 0},
+  {"notexec", MOUNT_NOTEXEC, 0},
+  {"cygexec", MOUNT_CYGWIN_EXEC, 0},
+  {"nosuid", 0, 0},
+  {"managed", MOUNT_ENC, 0}
+};
+
 static void
 print_version ()
 {
@@ -179,12 +221,21 @@ Compiled on %s\n\
 ", progname, len, v, __DATE__);
 }
 
+static char *
+concat3 (char *a, const char *b, const char *c)
+{
+  size_t totlen = strlen (a) + strlen (b) + strlen (c) + 1;
+  a = (char *) realloc (a, totlen);
+  return strcat (strcat (a, b), c);
+}
+
 int
 main (int argc, char **argv)
 {
   int i;
   int flags = MOUNT_BINARY;
   int default_flag = MOUNT_SYSTEM;
+  char *options = strdup ("");
   enum do_what
   {
     nada,
@@ -231,6 +282,12 @@ main (int argc, char **argv)
 	else
 	  usage ();
 	break;
+      case 'o':
+	if (*options)
+	  options = concat3 (options, ",", optarg);
+	else
+	  options = strdup (optarg);
+	break;
       case 'p':
 	if (do_what == nada)
 	  do_what = saw_show_cygdrive_prefix;
@@ -264,6 +321,30 @@ main (int argc, char **argv)
 	usage ();
       }
 
+  while (*options)
+    {
+      char *p = strchr (options, ',');
+      if (p)
+	*p++ = '\0';
+      else
+	p = strchr (options, '\0');
+
+      for (opt *o = oopts; o < (oopts + (sizeof (oopts) / sizeof (oopts[0]))); o++)
+	if (strcmp (options, o->name) == 0)
+	  {
+	    if (o->clear)
+	      flags &= ~o->val;
+	    else
+	      flags |= o->val;
+	    goto gotit;
+	  }
+      fprintf (stderr, "%s: invalid option - '%s'\n", progname, options);
+      exit (1);
+
+    gotit:
+      options = p;
+    }
+
   if (flags & MOUNT_NOTEXEC && flags & (MOUNT_EXEC | MOUNT_CYGWIN_EXEC))
     {
       fprintf (stderr, "%s: invalid combination of executable options\n", progname);
@@ -325,48 +406,54 @@ mount_commands (void)
   char system_flags[MAX_PATH];
 
   // write mount commands for user and system mount points
-  while ((p = getmntent (m)) != NULL) {
+  while ((p = getmntent (m)) != NULL)
     // Only list non-cygdrives
-    if (!strstr (p->mnt_opts, ",noumount")) {
-      strcpy(opts, " -f");
-      if      (p->mnt_type[0] == 'u')
-        strcat (opts, " -u");
-      else if (p->mnt_type[0] == 's')
-        strcat (opts, " -s");
-      if      (p->mnt_opts[0] == 'b')
-        strcat (opts, " -b");
-      else if (p->mnt_opts[0] == 't')
-        strcat (opts, " -t");
-      if (strstr (p->mnt_opts, ",exec"))
-        strcat (opts, " -x");
-      if (strstr (p->mnt_opts, ",noexec"))
-        strcat (opts, " -E");
-      while ((c = strchr (p->mnt_fsname, '\\')) != NULL)
-        *c = '/';
-      printf (format_mnt, opts, p->mnt_fsname, p->mnt_dir);
-    }
-  }
+    if (!strstr (p->mnt_opts, ",noumount"))
+      {
+	strcpy(opts, " -f");
+	if      (p->mnt_type[0] == 'u')
+	  strcat (opts, " -u");
+	else if (p->mnt_type[0] == 's')
+	  strcat (opts, " -s");
+	if      (p->mnt_opts[0] == 'b')
+	  strcat (opts, " -b");
+	else if (p->mnt_opts[0] == 't')
+	  strcat (opts, " -t");
+	if (strstr (p->mnt_opts, ",exec"))
+	  strcat (opts, " -x");
+	if (strstr (p->mnt_opts, ",noexec"))
+	  strcat (opts, " -E");
+	if (strstr (p->mnt_opts, ",managed"))
+	  strcat (opts, " -o managed");
+	while ((c = strchr (p->mnt_fsname, '\\')) != NULL)
+	  *c = '/';
+	printf (format_mnt, opts, p->mnt_fsname, p->mnt_dir);
+      }
   endmntent (m);
 
   // write mount commands for cygdrive prefixes
   cygwin_internal (CW_GET_CYGDRIVE_INFO, user, system, user_flags,
 		   system_flags);
-  if (strlen (user) > 0) {
-    strcpy (opts, "   ");
-    if      (user_flags[0] == 'b')
-      strcat (opts, " -b");
-    else if (user_flags[0] == 't')
-      strcat (opts, " -t");
-    printf (format_cyg, opts, user);
-  }
-  if (strlen (system) > 0) {
-    strcpy (opts, " -s");
-    if      (system_flags[0] == 'b')
-      strcat (opts, " -b");
-    else if (system_flags[0] == 't')
-      strcat (opts, " -t");
-    printf (format_cyg, opts, system);
-  }
+
+  if (strlen (user) > 0)
+    {
+      strcpy (opts, " -u");
+      if (user_flags[0] == 'b')
+	strcat (opts, " -b");
+      else if (user_flags[0] == 't')
+	strcat (opts, " -t");
+      printf (format_cyg, opts, user);
+    }
+
+  if (strlen (system) > 0)
+    {
+      strcpy (opts, " -s");
+      if (system_flags[0] == 'b')
+	strcat (opts, " -b");
+      else if (system_flags[0] == 't')
+	strcat (opts, " -t");
+      printf (format_cyg, opts, system);
+    }
 
   exit(0);
 }
