@@ -1,6 +1,6 @@
 /* strace.cc
 
-   Copyright 2000, 2001, 2002, 2003 Red Hat Inc.
+   Copyright 2000, 2001, 2002, 2003, 2004 Red Hat Inc.
 
    Written by Chris Faylor <cgf@redhat.com>
 
@@ -870,7 +870,7 @@ struct option longopts[] = {
   {NULL, 0, NULL, 0}
 };
 
-static const char *const opts = "b:dhfm:no:p:S:tTuvw";
+static const char *const opts = "+b:dhfm:no:p:S:tTuvw";
 
 static void
 print_version ()
@@ -890,7 +890,7 @@ print_version ()
   printf ("\
 %s (cygwin) %.*s\n\
 System Trace\n\
-Copyright 2000, 2001, 2002, 2003 Red Hat, Inc.\n\
+Copyright 2000, 2001, 2002, 2003, 2004 Red Hat, Inc.\n\
 Compiled on %s\n\
 ", pgm, len, v, __DATE__);
 }
@@ -985,7 +985,7 @@ character #%d.\n", optarg, (int) (endptr - optarg), endptr);
     error (0, "must provide a process id to toggle tracing");
 
   if (!mask)
-    mask = 1;
+    mask = _STRACE_ALL;
 
   if (bufsize)
     setvbuf (ofile, (char *) alloca (bufsize), _IOFBF, bufsize);
